/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* display_stack.c
   Manage a stack to keep track of what view is currently visible */

#include "display_stack.h"
#include "world.h"

#include "tools.h"
#include "shade.h"
#include "text.h"

static DisplayStack *allocate_display_stack(void);
static void free_display_stack_entry(DisplayStack *entry);

/* Allocate space for an empty display stack entry */
DisplayStack *allocate_display_stack(void)
{
  DisplayStack *entry = malloc(sizeof(DisplayStack));

  if(entry == NULL)
  {
    Neprintf("Unable to allocate space for a display stack entry\n");
    return NULL;
  }

  entry->mode = NATOW_DISPLAY_UNKNOWN;
  entry->next = NULL;

  return entry;
}

/* Set the scene/view mode to display */
void set_display_mode(World *world, NatowDisplayMode mode)
{
  NatowDisplayMode old_mode;

  if(!world->display_stack)
  {
    world->display_stack = allocate_display_stack();
    old_mode = NATOW_DISPLAY_UNKNOWN;
  }
  else
    old_mode = world->display_stack->mode;

  /* check for shading */
  if(old_mode != NATOW_DISPLAY_UNKNOWN)
    if(check_for_shade(old_mode, mode))
    {
      if (init_shading(world))
        world->shade_mode = NATOW_SHADE_DOWN;
      else
        world->shade_mode = NATOW_SHADE_NONE;
    }

  if(world->display_stack != NULL)
    world->display_stack->mode = mode;
}

/* Get the scene/view/mode to display */
NatowDisplayMode get_display_mode(World *world)
{
  if(world->display_stack != NULL)
    return world->display_stack->mode;
  else
    return NATOW_DISPLAY_UNKNOWN;
}

/* Push the display mode stack */
void push_display_mode(World *world, NatowDisplayMode mode)
{
  DisplayStack *entry = allocate_display_stack();

  if(entry == NULL)
    return;

  /* check if this new screen requieres a shade */
  check_shade_action(world, mode);

  entry->mode = mode;
  entry->next = world->display_stack;
  world->display_stack = entry;
}

/* Pop the display mode stack */
void pop_display_mode(World *world)
{
  DisplayStack *entry = world->display_stack;

  if(entry == NULL)
  {
    Neprintf("Error: Attempt to pop empty display stack\n");
    return;
  }

  /* check if this new screen requieres a shade */
  check_shade_action(world, NATOW_DISPLAY_UNKNOWN);

  world->display_stack = entry->next;
  free_display_stack_entry(entry);
}

/* Free a display stack entry */
void free_display_stack_entry(DisplayStack *entry)
{
  Nfree(entry);
}

/* Free the natow display stack */
void clear_display_stack(World *world)
{
  DisplayStack *entry = world->display_stack, *tmp;

  while(entry != NULL)
  {
    tmp = entry->next;
    free_display_stack_entry(entry);
    entry = tmp;
  }
}

/* check if changing for this new screen needs to activate shading */
void check_shade_action(World *world, NatowDisplayMode mode)
{
  NatowDisplayMode old_mode, new_mode;

  /* no stack: nothing */
  if (world->display_stack == NULL)
    return;

  /* POP action: */
  if (mode == NATOW_DISPLAY_UNKNOWN)
    {
    /* POP at the end: nothing to do */
    if (world->display_stack->next)
      {
      return;
      }
    new_mode = world->display_stack->next->mode;
    old_mode = world->display_stack->mode;
    }
  else /* PUSH action */
    {
    new_mode = mode;
    old_mode = world->display_stack->mode;
    }

  /* check if going from 'old' to 'new' implies shading */
  if (check_for_shade(old_mode, new_mode))
    {
    if (init_shading(world))
      world->shade_mode = NATOW_SHADE_DOWN;
    else
      world->shade_mode = NATOW_SHADE_NONE;
    }
}

/* returns TRUE in going from 'old' to 'new' implies shading */
int check_for_shade(NatowDisplayMode old, NatowDisplayMode new)
{
  /* just a big enumaretion of cases */

  if ((old == NATOW_DISPLAY_MAIN_MENU)&&(new == NATOW_DISPLAY_GAME_CONFIG))
    return(TRUE);

  if ((old == NATOW_DISPLAY_GAME_CONFIG)&&(new == NATOW_DISPLAY_MAIN_MENU))
    return(TRUE);

  if ((old == NATOW_DISPLAY_GAME_CONFIG)&&(new == NATOW_DISPLAY_PLAYER_CONFIG))
    return(TRUE);

  if ((old == NATOW_DISPLAY_PLAYER_CONFIG)&&(new == NATOW_DISPLAY_PLAYER_SHOP))
    return(TRUE);

  if ((old == NATOW_DISPLAY_PLAYER_SHOP)&&(new == NATOW_DISPLAY_START_ROUND))
    return(TRUE);

  if ((old == NATOW_DISPLAY_START_ROUND)&&(new == NATOW_DISPLAY_ROUND))
    return(TRUE);

  if ((old == NATOW_DISPLAY_ROUND)&&(new == NATOW_DISPLAY_END_ROUND))
    return(TRUE);

  if ((old == NATOW_DISPLAY_END_ROUND)&&(new == NATOW_DISPLAY_END_GAME))
    return(TRUE);

  if ((old == NATOW_DISPLAY_END_GAME)&&(new == NATOW_DISPLAY_MAIN_MENU))
    return(TRUE);

  /*
  if ((old == NATOW_DISPLAY_)&&(new == NATOW_DISPLAY_))
    return(TRUE);
  */

  /* default: no */
  return(FALSE);
}
