/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* explosion.c
   This file contains all the explosion related code */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <GL/gl.h>

#include "explosion.h"

#include "world.h"
#include "game.h"
#include "round.h"
#include "turn.h"

#include "tools.h"
#include "text.h"
#include "vector.h"

#include "map.h"
#include "smoke.h"
#include "player.h"
#include "map_texture.h"
#include "objects.h"
#include "player_stats.h"
#include "player_list.h"
#include "light.h"

static Blast *add_blast(Round *round, float *pos, float radius, float speed);
static void update_blast(Blast *blast, float dt);
static void draw_blast(Blast *blast);
static void draw_sphere(void);
static void calc_map_explosion(Map *map, float *pos, float radius, float force);
static void calc_map_trace(Map *map, float *pos, float radius);
static void calc_player_explosion_damage(Player *shooter, float *pos, float force);
static void shake_cameras(Round *round, float force);
static int free_blast(Blast *blast);

/* Causes an explosion at 'pos', with radius 'radius', and force 'force'*/
void calc_explosion(Player *shooter, float *pos, float radius, float force)
{
  Round *round = shooter->game->round;
  Map *map = shooter->game->map;

  /* First destroy the map */
  calc_map_explosion(map, pos, radius, force);

  /* change the texture appearence */
  calc_map_trace(map, pos, radius);

  /* Damage any players in range */
  calc_player_explosion_damage(shooter, pos, force);

  /* Shake the camera(s) */
  shake_cameras(round, force/2.0);

  /* Add a blast(fireball) */
  add_blast(round, pos, radius, 2.5);
  
  /* add smokes (improve the density calculation) */
  add_smoke(round, pos, radius, force);
}

/* Adds a blast into the world */
static Blast *add_blast(Round *round, float *pos, float radius, float speed)
{
  Blast *new_blast = malloc(sizeof(Blast));

  if(new_blast == NULL)
  {
    Neprintf("Warning: Unable to allocate space for blast");
    return NULL;
  }

  new_blast->world = round->game->world;
  new_blast->game = round->game;
  new_blast->round = round;

  new_blast->pos[0] = pos[0];
  new_blast->pos[1] = pos[1];
  new_blast->pos[2] = pos[2];
  new_blast->age = 0.0;

  new_blast->speed = speed;
  new_blast->radius = radius;
  new_blast->current_radius = 0.0;
  new_blast->last = NULL;
  new_blast->next = NULL;
  /* in future, we will be able make weapons
     with various explosion colors (like paintball mode!) */
  new_blast->color[0] = 0.9;
  new_blast->color[1] = 0.6;
  new_blast->color[2] = 0.1;
  new_blast->color[3] = 1.;

  /* Attach to the front of the list */
  new_blast->next = round->blast_list;
  if(round->blast_list != NULL)
    round->blast_list->last = new_blast;
  round->blast_list = new_blast;

  return new_blast;
}

/* Remove a blast from the list */
void remove_blast(Blast *blast)
{
  if(blast->next != NULL)
    blast->next->last = blast->last;
  if(blast->last != NULL)
    blast->last->next = blast->next;
  else
    blast->round->blast_list = blast->next;

  free_blast(blast);
}

/* Updates all the blasts in the world */
void update_blasts(Round *round, float dt)
{
  Blast *blast_ptr = round->blast_list, *temp_blast;

  while(blast_ptr != NULL)
  {
    /* Blast may be removed so update now */
    temp_blast = blast_ptr->next;

    update_blast(blast_ptr, dt);

    blast_ptr = temp_blast;
  }
}

/* Updates a blast (fireball) */
static void update_blast(Blast *blast, float dt)
{
  /* Undo the last light */
  clear_dynamic_light(blast->world->game->map, blast->pos,
		      100.0 * blast->radius * exp(-5.0 * blast->current_radius/blast->radius));

  blast->age += dt;
  /* increase sphere transparency until .4 */
  blast->color[3] -= dt/2.;
  if (blast->color[3] < .4)
    blast->color[3] = .4;
  blast->current_radius += blast->age * blast->speed;

  /* Some blasts are having a radius of NaN, so this gets rid of them */
  if(blast->age > 5.0)
    remove_blast(blast);

  if(blast->current_radius > blast->radius)
    remove_blast(blast);
  else /* Flash from the explosion */
    add_dynamic_light(blast->world->game->map, blast->pos, blast->color,
		      100.0 * blast->radius * exp(-5.0 * blast->current_radius/blast->radius));
}

/* Draws all the blasts in the world */
void draw_blasts(Round *round)
{
  Blast *blast_ptr = round->blast_list;

  glPushAttrib(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT);
  glDepthMask(GL_TRUE);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  while(blast_ptr)
  {
    draw_blast(blast_ptr);
    blast_ptr = blast_ptr->next;
  }

  glPopAttrib();
}

/* Draws a blast (fireball) */
static void draw_blast(Blast *blast)
{
  glPushMatrix();

  glColor4fv(blast->color);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, blast->color);
  glTranslatef(blast->pos[0], blast->pos[1], blast->pos[2]);
  glScalef(2.0*blast->current_radius, 2.0*blast->current_radius, 2.0*blast->current_radius);
  draw_sphere();

  glPopMatrix();
}

/* Draw the unit sphere (temp) */
static void draw_sphere(void)
{
  GLfloat theta, h;
  GLfloat sin_theta, cos_theta, old_sin_theta = 0.0, old_cos_theta = 1.0;
  GLfloat sin_phi, cos_phi, old_cos_phi;
  GLfloat dtheta = (2.0 * M_PI) / 8.0, dh = 2.0 / 8.0;

  glBegin(GL_TRIANGLES);
  for(theta = dtheta; theta <= 2.0 * M_PI; theta += dtheta)
  {
      sin_theta = sin(theta);
      cos_theta = cos(theta);

      old_cos_phi = 0.0;
      for(h = dh - 1.0; h <= 1.0; h += dh)
      {
	sin_phi = h;
	cos_phi = sqrt(1.0 - h*h);

	glNormal3f(cos_phi * cos_theta, sin_phi, cos_phi * sin_theta);

	glVertex3f(cos_phi * old_cos_theta, h, cos_phi * old_sin_theta);
	glVertex3f(cos_phi * cos_theta, h, cos_phi * sin_theta);
	glVertex3f(old_cos_phi * cos_theta, h - dh, old_cos_phi * sin_theta);

	glVertex3f(cos_phi * old_cos_theta, h, cos_phi * old_sin_theta);
	glVertex3f(old_cos_phi * cos_theta, h - dh, old_cos_phi * sin_theta);
	glVertex3f(old_cos_phi * old_cos_theta, h - dh, old_cos_phi * old_sin_theta);

	old_cos_phi = cos_phi;
      }

      old_sin_theta = sin_theta;
      old_cos_theta = cos_theta;
    }
  glEnd();
}

/* Make a hole in the map */
static void calc_map_explosion(Map *map, float *pos, float radius, float force)
{
  int x0, x1, y0, y1, x, y;
  float h, hm, rx, ry;

  /* The map square that contains the blast */
  x0 = (int)((pos[0] - radius + 0.5f) / 5.0f);
  x1 = (int)((pos[0] + radius + 0.5f) / 5.0f);
  y0 = (int)((pos[2] - radius + 0.5f) / 5.0f);
  y1 = (int)((pos[2] + radius + 0.5f) / 5.0f);
  /* Don't try to blow up what isn't the map */
  /* protect borders in order to get a good
       joint with the extended ground */
  if(x0 < 1)
    x0 = 1;
  if(x1 >= map->size_x-1)
    x1 = map->size_x - 2;
  if(y0 < 1)
    y0 = 1;
  if(y1 >= map->size_y-1)
    y1 = map->size_y - 2;

  /* Go through the above square of verticies and reduce their size if hit by explosion */
  for(x = x0; x <= x1; x++)
    for(y = y0; y <= y1; y++)
    {
      /* Radii along the x and y-axis (map) */
      rx = 5.0f * (float)x - pos[0];
      ry = 5.0f * (float)y - pos[2];
      
      /* manage damages on objects */
      if(map->objects[x][y])
	damage_objects(map->game, map->objects[x][y], pos, radius, force);
      
      h = radius*radius - (rx*rx + ry*ry);
      if(h >= 0) /* If explosion exists at this point */
      {
	/* Map height at current vertex */
	hm = map->pos[x][y][1];
	/* height/depth of explosion at this point */
	h = sqrt(h);

	if(hm > pos[1] - h) /* Explosion isn't completely above ground */
	{
	  if(hm > (pos[1] + h)) /* Explosion has land above and below */
	    map->pos[x][y][1] += -2.0 * h;
	  else /* Explosion has land below only */
	    map->pos[x][y][1] = pos[1] - h;
	}
      }
    }


  /* Re-calculate the normals */
  for(x = x0; x <= x1; x++)
    for(y = y0; y <= y1; y++)
      calc_normal(map, x, y);

  /* Re-shade the hole */
  /* NOTE: this doesn't recalculate the shadows */
  shade_map_area_with_sun(map, x0, y0, x1, y1);
}

/* not finished! */
/* BOB: If you reduce the resolution the lag as it regens the
   map texture is minimised, to have high res textures, we need
   to divide the texture into pieces, or use blast decals (probably
   the best method) */
/* compute effects on the map texture */
static void calc_map_trace(Map *map, float *pos, float radius)
{
  int x0, x1, y0, y1, x, y;
  float h, rx, ry;
  int texx, texy;
  int texx2, texy2;
  float contrib;
  float dark_color[3] = { 0., 0., 0.};
  int pos_0, pos_2;

  /* Disabled as it slows the game down too much */
  return;

  /* The map square that contains the blast */
  x0 = (int)((pos[0] - radius + 0.5f) / 5.0f);
  x1 = (int)((pos[0] + radius + 0.5f) / 5.0f);
  y0 = (int)((pos[2] - radius + 0.5f) / 5.0f);
  y1 = (int)((pos[2] + radius + 0.5f) / 5.0f);

  /* re-compute the radius in texture units */
  radius = (float)( radius / 5.0f ) * map->tex_width / map->size_x;
  /* Don't try to blow up what isn't the map */
  if(x0 < 0)
    x0 = 0;
  if(x1 >= map->size_x)
    x1 = map->size_x - 1;
  if(y0 < 0)
    y0 = 0;
  if(y1 >= map->size_y)
    y1 = map->size_y - 1;

  /* texture coordinates */
  texy  = (int)((float)(x0 + 1) * map->tex_width / map->size_x);
  texy2 = (int)((float)(x1 + 1) * map->tex_width / map->size_x);
  texx  = (int)((float)(y0 + 1) * map->tex_height / map->size_y);
  texx2 = (int)((float)(y1 + 1) * map->tex_height / map->size_y);

  pos_2 = ((float)texy2+texy) / 2.;
  pos_0 = ((float)texx2+texx) / 2.;

  if (texx < 0)
    texx = 0;
  if (texy < 0)
    texy = 0;
  if (texx >= map->tex_height)
    texx = map->tex_height - 1;
  if (texy >= map->tex_width)
    texy = map->tex_width - 1;
  if (texx2 < 0)
    texx = 0;
  if (texy2 < 0)
    texy2 = 0;
  if (texx2 >= map->tex_height)
    texx2 = map->tex_height - 1;
  if (texy2 >= map->tex_width)
    texy2 = map->tex_width - 1;

  /* Go through the above square of texture and draw blast marks */
  
  /* TODO: take in count explosions that are not on the ground */
  for(x = texx; x <= texx2; x++)
    for(y = texy; y <= texy2; y++)
    {
      /* Radii along the x and y-axis (map) */
      rx = (float)x - pos_0;
      ry = (float)y - pos_2;
      h = radius*radius - (rx*rx + ry*ry);
      if(h >= 0) /* If explosion exists at this point */
      {
	h = sqrt(h);
        contrib = (.8*h / radius) + random_float(-0.05, 0.05);
	if (contrib > 0.)
	  {
	  if (contrib >= 1.)
	    {
            map->texture_data[x*map->tex_width*3 + 3*y + 0] = (GLubyte)(dark_color[0]*255);
            map->texture_data[x*map->tex_width*3 + 3*y + 1] = (GLubyte)(dark_color[1]*255);
            map->texture_data[x*map->tex_width*3 + 3*y + 2] = (GLubyte)(dark_color[2]*255);
	    }
	  else
	    {
            map->texture_data[x*map->tex_width*3 + 3*y + 0] = (GLubyte)
              (map->texture_data[x*map->tex_width*3 + 3*y + 0]*(1.-contrib) +
               dark_color[0]*contrib);
            map->texture_data[x*map->tex_width*3 + 3*y + 1] = (GLubyte)
              (map->texture_data[x*map->tex_width*3 + 3*y + 1]*(1.-contrib) +
               dark_color[1]*contrib);
            map->texture_data[x*map->tex_width*3 + 3*y + 2] = (GLubyte)
              (map->texture_data[x*map->tex_width*3 + 3*y + 2]*(1.-contrib) +
               dark_color[2]*contrib);
	    }
	  }
      }
    }

  /* regenerate the map texture */
  regen_map_texture(map);
}

/* Damage the players */
static void calc_player_explosion_damage(Player *shooter, float *pos, float force)
{
  Game *game;
  Round *round;
  PlayerEntry *pentry;
  Player *victim;
  float *victim_pos, *shooter_pos;
  float dx, dy, dz, d2, damage, distance[3], shot_range;
  int killed_victim;
  int winnings;

  game = shooter->game;
  round = game->round;
  shooter_pos = shooter->pos;

  /* Calculate the damage on the other players */
  pentry = round->players->first;
  while(pentry)
  {
    victim = pentry->player;

    victim_pos = victim->pos;

    /* No point hurting dead players */
    if(victim->player_state == PLAYER_ALIVE)
    {
      killed_victim = FALSE;

      dx = pos[0] - victim_pos[0];
      dy = pos[1] - victim_pos[1];
      dz = pos[2] - victim_pos[2];
      d2 = dx*dx + dy*dy + dz*dz;

      /* Amount of damage done by the shot */
      damage = 1000.0 * force / (d2*sqrt(d2));
      if(damage < 1.0)
	damage = 0.0;
      victim->health -= damage;
      /* Actual damage done to player */
      if(victim->health < 0.0)
      {
	killed_victim = TRUE;
	damage += victim->health;
      }

      /* Record this in the players statistics */
      record_damage(shooter, victim, damage);

      /* Vector between player who shot projectile, and target */
      distance[0] = victim_pos[0] - shooter_pos[0];
      distance[1] = victim_pos[1] - shooter_pos[1];
      distance[2] = victim_pos[2] - shooter_pos[2];
      shot_range = sqrt(distance[0]*distance[0] +
			distance[1]*distance[1] +
			distance[2]*distance[2]);


      /* If you've killed the player */
      if(killed_victim)
      {
	Nprintf("\"%s\" is dead! (long live \"%s\")\n", victim->name, victim->name);
	/* Player is now dead */
	victim->player_state = PLAYER_DEAD;

	/* Record this in the players statistics */
	record_kill(shooter, victim);

	/* Give the shooter some credit (or shame) */
	if(victim == shooter)
	{
	  shooter->kills--;
	  shooter->round_kills--;
	  Nprintf("\"%s\" commits suicide!\n", shooter->name);
	}
	else
	{
	  shooter->kills++;
	  shooter->round_kills++;
	}

	/* Replace player with an explosion */
	calc_explosion(victim, victim->pos, 20.0, 100.0);

	/* If this is the current player, cycle to the next available */
	if(victim == round->current_player)
	  player_end_turn(victim);
      }

      /* Calculate the winnings for the shooter (If they
	 did any damage) */
#define DAMAGE_FACTOR 1.0
#define RANGE_FACTOR (1.0 / 50.0)
#define KILL_FACTOR 0.5
      if(damage)
      {
	/* No money for killing yourself! */
	if(victim == shooter)
	  winnings = 0;
	else
	{
	  winnings = (DAMAGE_FACTOR + (RANGE_FACTOR * shot_range)) * damage;
	  /* Take a percentage of the players starting health */
	  if(killed_victim)
	    winnings += KILL_FACTOR * victim->tank->health;

	  Nprintf("\"%s\" does %.2f damage on \"%s\"", shooter->name, damage, victim->name);
	  if(killed_victim)
	    Nprintf_simple(" and kills them");
	  Nprintf_simple(" at a range of %.2fm, awarded $%d\n",
			 shot_range, winnings);
	}

	give_player_winnings(shooter, winnings);
	shooter->round_winnings += winnings;
      }
    }

    pentry = pentry->next;
  }
}
#undef DAMAGE_FACTOR
#undef RANGE_FACTOR
#undef KILL_FACTOR

/* Make the camera(s) shake from the explosion */
static void shake_cameras(Round *round, float force)
{
  Camera *camera = round->camera;

  /* if we just add the force, it can became too large */
  if (camera->shudder > force)
    camera->shudder += force/20.;
  else
    camera->shudder = force;
}

/* Remove all blasts */
int clear_blasts(Round *round)
{
  Blast *blast, *tmp;

  if(!round)
    return FALSE;

  blast = round->blast_list;
  while(blast)
  {
    tmp = blast->next;
    remove_blast(blast);
    blast = tmp;
  }

  return TRUE;
}

/* Free a blast */
static int free_blast(Blast *blast)
{
  if(!blast)
    return FALSE;

  Nfree(blast);

  return TRUE;
}
