/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* game_config.c
   Configure and start a game */

#include <string.h>

#include "game_config.h"
#include "game.h"
#include "world.h"
#include "text.h"
#include "tools.h"
#include "2d.h"
#include "glut_interface.h"
#include "display_stack.h"
#include "texture.h"
#include "player_list.h"
#include "level_list.h"

/* The properties for the game config screen */
static int selected = 0, nplayers = 3, nrounds = 4, starting_cash = 1000;
static int current_level = 0;
GameMode game_mode = TURN_BASED;
static LevelEntry *level = NULL;

static int init_game(Game *game);

/* Set the game config screen to defaults */
void reset_game_config(void)
{
  selected = 0;
  nplayers = 3;
  nrounds = 4;
  starting_cash = 1000;
  current_level = 0;
  game_mode = TURN_BASED;
  level = NULL;
}

/* Draw the game config screen */
int draw_game_config(Game *game)
#define MAX_TEXT_LENGTH 100
{
  char text[MAX_TEXT_LENGTH];

  if(!level)
    level = game->world->levels;

  activate_2d();

  draw_2d_text_center("Game Config", 1, 98, 91, 1, FALSE);

  snprintf(text, MAX_TEXT_LENGTH, "Number of players: %d", nplayers);
  draw_2d_text(text, 1, 85, 0, selected == 0);
  if(cursor_inside_text())
    selected = 0;

  snprintf(text, MAX_TEXT_LENGTH, "Number of rounds: %d", nrounds);
  draw_2d_text(text, 1, 80, 0, selected == 1);
  if(cursor_inside_text())
    selected = 1;

  snprintf(text, MAX_TEXT_LENGTH, "Starting cash: $%d", starting_cash);
  draw_2d_text(text, 1, 75, 0, selected == 2);
  if(cursor_inside_text())
    selected = 2;

  if(level)
  {
    snprintf(text, MAX_TEXT_LENGTH, "Map: %s", level->name);
    draw_2d_text(text, 1, 69, 0, selected == 3);
  }
  else
    draw_2d_text("Map: (no map)", 1, 69, 0, selected == 3);
  if(cursor_inside_text())
    selected = 3;

  snprintf(text, MAX_TEXT_LENGTH, "Game type:");
  draw_2d_text(text, 1, 63, 0, selected == 4);
  if(cursor_inside_text())
    selected = 4;
  switch(game_mode)
    {
    case TURN_BASED:
      snprintf(text, MAX_TEXT_LENGTH, "Turn based");
      break;
    case REAL_TIME:
      snprintf(text, MAX_TEXT_LENGTH, "Real time");
      break;
    case SINGLE_PLAYER:
      snprintf(text, MAX_TEXT_LENGTH, "Single player");
      break;
    default:
      snprintf(text, MAX_TEXT_LENGTH, "???");
      break;
    }
  draw_2d_text(text, 10, 57, 0, selected == 4);
  if(cursor_inside_text())
    selected = 4;
  if(game_mode != TURN_BASED)
    draw_2d_text("(does not work)", 10, 53, -2, TRUE);

  glColor4f(1.0, 1.0, 1.0, 1.0);
  if(level && level->snapshot)
  {
    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D, level->snapshot);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  }
  else
  {
    glColor4f(.5, .5, .5, 1.0);
    glDisable(GL_TEXTURE_2D);
  }

  glBegin(GL_QUADS);

  glTexCoord2f(1.0, 1.0);
  glVertex2f(97.0, 65.0);
  glTexCoord2f(0.0, 1.0);
  glVertex2f(57.0, 65.0);
  glTexCoord2f(0.0, 0.0);
  glVertex2f(57.0, 65.0 - (40.0 * 640.0 / 480.0));
  glTexCoord2f(1.0, 0.0);
  glVertex2f(97.0, 65.0 - (40.0 * 640.0 / 480.0));

  glEnd();
  glDisable(GL_TEXTURE_2D);

  draw_2d_text("Last", 1, 1, 3, selected == 6);
  if(cursor_inside_text())
    selected = 6;

  draw_2d_text_right("Next", 1, 98, 1, 3, selected == 5);
  if(cursor_inside_text())
    selected = 5;

  /*draw_2d_text_center("Up/Down selects", 1, 98, 7, -3, FALSE);
  draw_2d_text_center("+/- Increases/Decreases", 1, 98, 4, -3, FALSE);
  draw_2d_text_center("Press enter when finished", 1, 98, 1, -3, 2);*/

  deactivate_2d();

  return TRUE;
}

/* Deal with the input when in the game config mode */
int manage_game_config_input(Game *game, int key)
{
  World *world;

  if(!game || !game->world)
    return FALSE;

  world = game->world;

  /* Abort this game */
  if(key == KEY_ESC || (key == KEY_MOUSE1 && selected == 6))
  {
    reset_game_config();
    finish_game(game);
    set_display_mode(world, NATOW_DISPLAY_MAIN_MENU);
  }

  else if(key == KEY_UP)
  {
    selected--;
    if(selected < 0)
      selected = 6;
  }

  else if(key == KEY_DOWN)
  {
    selected++;
    if(selected > 6)
      selected = 0;
  }

  else if(key == KEY_PLUS || key == KEY_RIGHT || key == KEY_MOUSE1)
  {
    switch(selected)
    {
    case 0:
      nplayers++;
      break;
    case 1:
      nrounds++;
      break;
    case 2:
      starting_cash += 10;
      break;
    case 3:
      level = level->next;
      if(!level)
	level = world->levels;
      break;
    case 4:
      game_mode++;
      if(game_mode > SINGLE_PLAYER)
        game_mode = TURN_BASED;
      /* temp: REAL_TIME not implemented */
      if(game_mode == REAL_TIME)
        game_mode = SINGLE_PLAYER;
      break;
    case 5:
      init_game(game);
      break;
    case 6:
      break;
    }
  }

  else if(key == KEY_MINUS || key == KEY_LEFT || key == KEY_MOUSE3)
  {
    switch(selected)
    {
    case 0:
      if(nplayers > 2)
	nplayers--;
      break;
    case 1:
      if(nrounds > 1)
	nrounds--;
      break;
    case 2:
      if(starting_cash > 10)
	starting_cash -= 10;
      break;
    case 3:
      level = level->last;
      if(!level)
      {
	level = world->levels;
	while(level->next)
	  level = level->next;
      }
      break;
    case 4:
      /* I do that this way 'cause it seems to be treated as unsigned */
      if (game_mode == TURN_BASED)
        game_mode = SINGLE_PLAYER;
      else
        game_mode--;
      /* temp: REAL_TIME not implemented */
      if (game_mode == REAL_TIME)
        game_mode = TURN_BASED;
      break;
    }
  }

  else if((key == KEY_SPACE) || (key == KEY_RETURN) ||
	  (key == KEY_MOUSE1 && selected == 5))
    init_game(game);

  return TRUE;
}

/* Get the configured game going */
static int init_game(Game *game)
#define MAX_PLAYER_LENGTH 50
{
  World *world;
  Player *player;
  char player_name[MAX_PLAYER_LENGTH];
  int i;

  if(!game || !game->world)
    return FALSE;
  
  world = game->world;

  if(nplayers > 0)
  {
    /* Make that many players, give some starting cash */
    for(i = 0; i < nplayers; i++)
    {
      snprintf(player_name, MAX_PLAYER_LENGTH, "Player %d", i+1);
      player = add_player(world, player_name);
      if(game_mode == SINGLE_PLAYER)
      {
	if (i == 0)
	  set_player_type(player, PLAYER_HUMAN);
	else
	  set_player_type(player, PLAYER_AI_STUPID);
      }
      else if (game_mode == TURN_BASED)
	set_player_type(player, PLAYER_HUMAN);
      else if (game_mode == REAL_TIME)
      {/* not implemented */
      }
    }

    if(!level)
      level = game->world->levels;

    /* Configure the game */
    set_game_mode(game, game_mode);
    set_game_gravity(game, -9.81);
    set_game_level(game, level);
    set_game_driving_time(game, 20.0);
    set_game_rounds(game, nrounds);
    set_game_starting_cash(game, starting_cash);

    /* Add all the availiable players at the moment */
    /* TODO: To be replaced with a network style add/remove players */
    player = world->player_list;
    while(player)
    {
      player->game = game;

      /* Set the player up for this game */
      player->winnings = 0;
      player->temp_cash = 0;
      player->temp_winnings = 0;
      player->percent_transferred = 0.0;
      player->cash_display_timer = 0.0;
      player->kills = 0;
      set_player_cash(player, game->starting_cash);

      /* Add the player to this game */
      append_player(game->players, player);

      player = player->next;
    }

    /* Configure the players */
    set_display_mode(world, NATOW_DISPLAY_PLAYER_CONFIG);

    /* Reset selection */
    reset_game_config();
  }
  else
  {
    Neprintf("Cannot start a game with no players!\n");
    return FALSE;
  }

  return TRUE;
}

static PlayerEntry *end_game_player = NULL;

/* Draw the end game screen */
int draw_end_game(Game *game)
#define MAX_LENGTH 100
#define MAX_VIS 11
{
  PlayerEntry *pentry;
  Player *player;
  char text[MAX_LENGTH];
  int i, selected, position, is_winner, nwinners;

  if(!game)
    return FALSE;

  activate_2d();

  draw_2d_text_center("Game over!", 1, 98, 90, 4, FALSE);

  draw_2d_text("Name", 10, 80, 0, FALSE);
  draw_2d_text("Cash", 55, 80, 0, FALSE);
  draw_2d_text("Kills", 80, 80, 0, FALSE);

  draw_2d_text_center("Press any key to continue", 1, 98, 1, 0, 2);

  /* If no player selected, select the first player in this round */
  if(!end_game_player)
    end_game_player = game->players->first;
  if(!end_game_player)
  {
    deactivate_2d();
    return FALSE;
  }

  /* Count how many winners there are */
  nwinners = 0;
  is_winner = TRUE;
  pentry = game->players->first;
  while(pentry && is_winner)
  {
    if(compare_game_winners(&pentry, &game->players->first) == 0)
      nwinners++;
    else
      is_winner = FALSE;

    pentry = pentry->next;
  }

  /* Find out which number the selected player is */
  pentry = game->players->first;
  selected = 0;
  while((pentry != end_game_player) && (selected < (game->players->nplayers - MAX_VIS)))
  {
    pentry = pentry->next;
    selected++;
  }

  /* Don't select the bottom players, as they're already visible */
  if(selected == (game->players->nplayers - MAX_VIS))
    end_game_player = pentry;

  /* Display MAX_VIS players */
  for(i = selected; i < selected + MAX_VIS && pentry; i++)
  {
    player = pentry->player;

    position = 70 - (i - selected)*5;
    is_winner = i < nwinners;

    draw_2d_text(player->name, 10, position, 0, is_winner);

    snprintf(text, MAX_LENGTH, "$%d", player->cash);
    draw_2d_text(text, 55, position, 0, is_winner);

    snprintf(text, MAX_LENGTH, "%d", player->kills);
    draw_2d_text(text, 80, position, 0, is_winner);

    pentry = pentry->next;
  }

  /* If more players below, show that */
  if(pentry)
    draw_2d_text("...", 10, 70 - MAX_VIS*5, 0, TRUE);

  deactivate_2d();

  return TRUE;
}

/* Deal with the input when in the end game screen */
int manage_end_game_input(Game *game, int key)
{
  if(key == KEY_UP)
  {
    if(end_game_player->last)
      end_game_player = end_game_player->last;
  }
  else if(key == KEY_DOWN)
  {
    if(end_game_player->next)
      end_game_player = end_game_player->next;
  }
  else /* On any other key, end the game */
    finish_game(game);

  return TRUE;
}
