/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* level.c
   loads all the required info for a level (map, textures etc) */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <math.h>

#include "level.h"
#include "game.h"

#include "file.h"
#include "text.h"
#include "tools.h"

#include "map.h"
#include "texture.h"
#include "objects.h"
#include "map_texture.h"
#include "lakes.h"
#include "satellite.h"

#include "glut_interface.h"


/* list of keywords for level description */
/*
name              "string"	name of the map
object            "file x y"	add an object at position (x,y)
map		  "file"	.rgb file for the map (altitude)
max_altitude      float         the maximum altiude for the map
fog_colour	  "r g b"	color for fog
fog_density	  float		density for fog
fog_end  	  float		distance of fog-ending
surface		  "file"	.rgb texture for the map
sky		  "file"	.rgb texture for the sky
lake		  "x0 y0 xm ym"	lake creation (start xy0, end xym)
sea_level	  float		altitude for the sea
sea_variations	  float		altitude variations for the sea
sea_period        float		period of these variations
sea_frequency	  float		frequency of these variations
fog_state	  boolean	if 0, disable fog for the level
gravity		  float		gravity
start_point	  "x y"		add a starting point in map
sun_direction	  "x y z"	direction of the sun
sun_colour	  "r g b"	color of the sun
sun_ambient	  "r g b"	ambient light
sea_colour	  "r g b a"	color for the sea
lake_colour	  "r g b a"	color for the lakes
birds		  int		number of birds in the level
objects		  "name"	add objects 'name' in the level
nb_objects	  int		set the number of objects
rock_tex_file     "file"        filename of rock texture
ice_tex_file      "file"        filename of ice texture
tree_tex_file     "file"        filename of tree texture
grass_tex_file    "file"        filename of grass texture
sand_tex_file     "file"        filename of sand texture
undersea_tex_file "file"        filename of under-sea texture
sun|moon1|moon2   "r ax bx ay by az bz r g b"
			create a satellite with radius r, color
			r,g,b, 
			and with rotation angle of a[xyz]*t+b[xyz]
			for each axis, with t, the current time in sec.
*/

/* Load a level file */
int load_level(Game *game, const char *fname)
{
  FILE *level_file = fopen(fname, "r");
  char *option = NULL, *value = NULL;
  char *map_name = NULL, *map_file = NULL, *map_texture = NULL;
  char *sky_texture = NULL;
  char *fog_colour = NULL;
  char *sun_colour = NULL;
  char *sun_direction = NULL, *sats[3] = {NULL,NULL,NULL};
  char *sun_ambient = NULL;
  char *sea_colour = NULL;
  char *lake_colour = NULL;
  char *rock_tex_fname = NULL, *ice_tex_fname = NULL, *tree_tex_fname = NULL;
  char *grass_tex_fname = NULL, *sand_tex_fname = NULL, *undersea_tex_fname = NULL;
  float gravity=-1.;
  int fog_state=1;  /* default: fog anabled */
  float sea_level = 0.0, has_sea_level = FALSE;
  float sea_variations = 0.0;
  float sea_frequency = 1.0;
  float max_altitude = 51.0;
  int valid_level = TRUE;
  /* YP: lakes management */
  Lake **lks = NULL; /* list of lakes */
  int nbl = 0, i;
  char *lake_param[32] = {NULL}; /* max 32 lakes */
  float fog_density = -1.;
  float fog_end = -1.;
  float fcr, fcg, fcb, fca;
  char *startp[64] = {NULL}; /* make it dynamic */
  int nb_startp=0;
  int birds_number = 0;
  char *objects = NULL;
  int nb_objects = 10;
  Map *map = NULL;

  Nprintf("Loading n.a.t.o.w level \"%s\"\n", fname);

  if(level_file == NULL)
  {
    Neprintf("  Unable to open level file \"%s\"\n", fname);
    return FALSE;
  }

  do
  {
    Nfree(option);
    option = get_file_word(level_file);
    Nfree(value);
    if(option != NULL)
      value = get_file_word(level_file);

    if(option == NULL);
    else if(!strcmp(option, "sun_direction"))
    {
      Nfree(sun_direction);
      sun_direction = strdup(value);
    }

    else if(!strcmp(option, "sun_colour"))
    {
      Nfree(sun_colour);
      sun_colour = strdup(value);
    }

    else if(!strcmp(option, "start_point"))
    {
      if (nb_startp >= 64)
        {
	Neprintf("  Too many starting points in the map (new one ignored).\n");
	}
      else
        startp[nb_startp++] = strdup(value);
    }

    else if(!strcmp(option, "sea_colour"))
    {
      Nfree(sea_colour);
      sea_colour = strdup(value);
    }

    else if(!strcmp(option, "lake_colour"))
    {
      Nfree(lake_colour);
      lake_colour= strdup(value);
    }

    else if(!strcmp(option, "object"))
    {/* to do */
    }

    else if(!strcmp(option, "sun_ambient"))
    {
      Nfree(sun_ambient);
      sun_ambient = strdup(value);
    }

    else if(!strcmp(option, "fog_state"))
    {
      fog_state = atoi(value);
    }

    else if(!strcmp(option, "birds"))
    {
      birds_number = atoi(value);
    }

    else if(!strcmp(option, "gravity"))
    {
      gravity = atof(value);
    }

    else if(!strcmp(option, "name"))
    {
      Nfree(map_name);
      map_name = strdup(value);
    }

    else if(!strcmp(option, "map"))
    {
      Nfree(map_file);
      map_file = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "max_altitude"))
    {
      max_altitude = atof(value);
    }

    else if(!strcmp(option, "fog_colour"))
    {
      Nfree(fog_colour);
      fog_colour = strdup(value);
    }

    else if(!strcmp(option, "sun"))
    {
      Nfree(sats[0]);
      sats[0] = strdup(value);
    }

    else if(!strcmp(option, "moon1"))
    {
      
      Nfree(sats[1]);
      sats[1] = strdup(value);
    }

    else if(!strcmp(option, "moon2"))
    {
      
      Nfree(sats[2]);
      sats[2] = strdup(value);
    }

    else if(!strcmp(option, "fog_density"))
    {
      fog_density = atof(value);
    }

    else if(!strcmp(option, "fog_end"))
    {
      fog_end = atof(value);
    }

    else if(!strcmp(option, "surface"))
    {
      Nfree(map_texture);
      map_texture = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "sky"))
    {
      Nfree(sky_texture);
      sky_texture = make_absolute_fname(fname, value);
    }

    /* Get lakes values */
    else if(!strcmp(option, "lake"))
    {
      /* Add a new lake entry */
      lake_param[nbl] = strdup(value);
      nbl++;
    }

    else if(!strcmp(option, "sea_level"))
    {
      sea_level = atof(value);
      has_sea_level = TRUE;
    }

    else if(!strcmp(option, "sea_variations"))
    {
      sea_variations = atof(value);
    }

    else if(!strcmp(option, "sea_period"))
    {
      sea_frequency = 1.0 / atof(value);
    }

    else if(!strcmp(option, "sea_frequency"))
    {
      sea_frequency = atof(value);
    }

    else if(!strcmp(option, "objects"))
    {
      Nfree(objects);
      objects = strdup(value);
    }

    else if(!strcmp(option, "nb_objects"))
    {
      nb_objects = atoi(value);
    }

    else if(!strcmp(option, "rock_texture"))
    {
      Nfree(rock_tex_fname);
      rock_tex_fname = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "ice_texture"))
    {
      Nfree(ice_tex_fname);
      ice_tex_fname = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "tree_texture"))
    {
      Nfree(tree_tex_fname);
      tree_tex_fname = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "grass_texture"))
    {
      Nfree(grass_tex_fname);
      grass_tex_fname = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "sand_texture"))
    {
      Nfree(sand_tex_fname);
      sand_tex_fname = make_absolute_fname(fname, value);
    }

    else if(!strcmp(option, "undersea_texture"))
    {
      Nfree(undersea_tex_fname);
      undersea_tex_fname = make_absolute_fname(fname, value);
    }
  } while(option != NULL);

  fclose(level_file);

  /* Load the altitude map */
  if(map_file == NULL)
  {
    Neprintf("  No map file specified in level \"%s\"\n", fname);
    valid_level = FALSE;
  }
  else
  {
    Nprintf("  Setting a maximum altitude of %fm\n", max_altitude);

    Nprintf("  Loading map altitude file: \"%s\"\n", map_file);
    map = game->map = load_map(map_file, max_altitude);

    if(!map)
      Neprintf("    Error loading map!\n");
    else
    {
      map->game = game;

      if(map_name)
      {
	Nprintf("  Setting map name: \"%s\"\n", map_name);
	Nfree(map->name);
	map->name = strdup(map_name);
      }
    }
  }

  /* Load the map texture */
  if(map_texture)
  {
    if(!map_file)
      Neprintf("  Unable to load map texture file, as there is no map!\n", map_texture);
    else
    {
      Nprintf("  Loading map texture file: \"%s\"\n", map_texture);
      map->texture = load_gl_texture(map_texture, FALSE);
      map->tex_scale_x = 1.0/(map->size_x);
      map->tex_scale_y = 1.0/(map->size_y);
    }
  }

  /* Load the sky texture */
  if(sky_texture)
  {
    Nprintf("  Loading sky texture file: \"%s\"\n", sky_texture);
    map->sky_texture = load_gl_texture(sky_texture, FALSE);
  }
  init_sky(game);

  /* Set the sea level */
  if(has_sea_level)
  {
    Nprintf("  Setting world sea level: %fm\n", sea_level);
    game->sea_level      = sea_level;
    game->sea_variations = sea_variations;
    game->sea_frequency  = sea_frequency;
  }

  /* Treating lakes (if has some) */
  if((lks = (Lake**)malloc(sizeof(Lake*)*(nbl+1))) == NULL)
    Neprintf("  Not enough memory. Lakes are deactivated.\n");
  else
  {
    lks[nbl] = NULL;  /* end of list */
    for(i=0; i<nbl; i++)
    {
      Nprintf("  Creating lake '%s'\n", lake_param[i]);
      lks[i] = create_lake(map, lake_param[i]);
    }
  }
  map->lakes = lks;
  init_lakes(game, map->lakes);

  if(fog_density > 0.0)
  {
    Nprintf("  Setting fog density to %f\n", fog_density);
    map->fog_density = fog_density;
  }

  if(fog_end > 0.0)
  {
    Nprintf("  Setting fog-ending distance to %f\n", fog_end);
    map->fog_end = fog_end;
  }

  if(fog_colour != NULL)
  {
    Nprintf("  Setting fog color: '%s'\n", fog_colour);
    sscanf(fog_colour, "%f%f%f", &fcr, &fcg, &fcb);
    map->fog_colour[0] = fcr;
    map->fog_colour[0] = fcg;
    map->fog_colour[0] = fcb;
  }

  if(!fog_state)
    Nprintf("  Switching off fog for level\n");
  else
    Nprintf("  Switching on fog for level\n");
  game->world->is_level_fog = fog_state;

  if(gravity > 0.0)
  {
    Nprintf("  Setting gravity to %f\n", gravity);
    game->gravity = gravity;
  }

  if(sun_direction != NULL)
  {
    Nprintf("  Setting sun direction: '%s'\n", sun_direction);
    sscanf(sun_direction, "%f%f%f", &fcr, &fcg, &fcb);
    map->sun_dir[0] = fcr;
    map->sun_dir[1] = fcg;
    map->sun_dir[2] = fcb;
    map->sun_dir[3] = 0.0;
  }

  if(sun_colour != NULL)
  {
    Nprintf("  Setting sun color: '%s'\n", sun_colour);
    sscanf(sun_colour, "%f%f%f", &fcr, &fcg, &fcb);
    map->sun_diffuse[0] = fcr;
    map->sun_diffuse[1] = fcg;
    map->sun_diffuse[2] = fcb;
    map->sun_diffuse[3] = 1.0;
  }

  if(sun_ambient != NULL)
  {
    Nprintf("  Setting ambient color: '%s'\n", sun_ambient);
    sscanf(sun_ambient, "%f%f%f", &fcr, &fcg, &fcb);
    map->sun_ambient[0] = fcr;
    map->sun_ambient[1] = fcg;
    map->sun_ambient[2] = fcb;
    map->sun_ambient[3] = 1.0;
  }

  if(sea_colour != NULL)
  {
    Nprintf("  Setting sea color: '%s'\n", sea_colour);
    sscanf(sea_colour, "%f%f%f%f", &fcr, &fcg, &fcb, &fca);
    map->sea_colour[0] = fcr;
    map->sea_colour[1] = fcg;
    map->sea_colour[2] = fcb;
    map->sea_colour[3] = fca;
  }

  if(lake_colour != NULL)
  {
    Nprintf("  Setting lakes color: '%s'\n", lake_colour);
    sscanf(lake_colour, "%f%f%f%f", &fcr, &fcg, &fcb, &fca);
    map->lake_colour[0] = fcr;
    map->lake_colour[1] = fcg;
    map->lake_colour[2] = fcb;
    map->lake_colour[3] = fca;
  }

  if(birds_number > 0)
  {
    Nprintf("  Setting birds number to %d.\n", birds_number);
    map->birds_number = birds_number;
  }

  /* The maps rock texture */
  if(rock_tex_fname != NULL)
  {
    Nfree(map->rock_tex_fname);
    map->rock_tex_fname = strdup(rock_tex_fname);
  }

  /* The maps ice texture */
  if(ice_tex_fname != NULL)
  {
    Nfree(map->ice_tex_fname);
    map->ice_tex_fname = strdup(ice_tex_fname);
  }

  /* The maps tree texture */
  if(tree_tex_fname != NULL)
  {
    Nfree(map->tree_tex_fname);
    map->tree_tex_fname = strdup(tree_tex_fname);
  }

  /* The maps grass texture */
  if(grass_tex_fname != NULL)
  {
    Nfree(map->grass_tex_fname);
    map->grass_tex_fname = strdup(grass_tex_fname);
  }

  /* The maps sand texture */
  if(sand_tex_fname != NULL)
  {
    Nfree(map->sand_tex_fname);
    map->sand_tex_fname = strdup(sand_tex_fname);
  }

  /* The maps under-sea texture */
  if(undersea_tex_fname != NULL)
  {
    Nfree(map->undersea_tex_fname);
    map->undersea_tex_fname = strdup(undersea_tex_fname);
  }

  /* Generate the map's texture if map texture wasn't explicitly specified */
  if(map_texture == NULL && map != NULL)
  {
    Nprintf("  Generating map texture...\n");
    generate_map_texture(map);
    Nprintf("  Finished generating map texture.\n");
  }

  /* create satellites */
  if (sats[0] != NULL)
    {
      game->world->sun = create_satellite(game->world, sats[0]);
    }
  if (sats[1] != NULL)
    {
      game->world->moon1 = create_satellite(game->world, sats[1]);
    }
  if (sats[2] != NULL)
    {
      game->world->moon2 = create_satellite(game->world, sats[2]);
    }

  /* Adding starting points */
  if (nb_startp > 0)
    map->startingpoints = malloc(sizeof(StartingPoint*)*(nb_startp+1));
  if (map->startingpoints != NULL)
    {
    Nprintf("  Adding %d starting points.\n", nb_startp);
    map->startingpoints[nb_startp] = NULL;
    for(i=0; i<nb_startp; i++)
      {
      map->startingpoints[i] = malloc(sizeof(StartingPoint));
      if (map->startingpoints[i] == NULL)
        {
	Neprintf("    Not enough memory. Starting points are truncated.\n");
	break;
	}
      sscanf(startp[i], "%f%f",
             &(map->startingpoints[i]->x),
	     &(map->startingpoints[i]->y));
      }
    }
  else
  if (nb_startp > 0)
    {
    Neprintf("Not enough memory to create starting points. Ignored.\n");
    }

  /* adding objects */
  if ((objects != NULL)&&(nb_objects > 0))
    {/* create the objects in the list */
    Object *obj = NULL;
    float x, y;

    Nprintf("Adding %d object(s) '%s'\n", nb_objects, objects);
    for(i=0; i<nb_objects; i++)
      {
      obj = load_object(game, objects);
      if (obj == NULL)
        {
	Nprintf("  cannot load/create object 'objects'\n");
	break;
	}
      add_object(game, obj);
      /* set object position (on the ground) */
      do
        {
        x = random_integer(0, map->size_x - 1)*map->x_resolution;
	y = random_integer(0, map->size_y - 1)*map->y_resolution;
	}
      while(get_height(game->map, x, y) <= game->sea_level);
      add_object_in_map(game, obj,
           (int)(x/map->x_resolution), (int)(y/map->y_resolution));
      obj->pos[0] = x;
      obj->pos[1] = get_height(game->map, x, y);
      obj->pos[2] = y;
      }
    }

  /* Free all those strings */
  Nfree(map_name);
  Nfree(map_file);
  Nfree(map_texture);
  Nfree(sky_texture);
  Nfree(lake_colour);
  Nfree(rock_tex_fname);
  Nfree(ice_tex_fname);
  Nfree(tree_tex_fname);
  Nfree(grass_tex_fname);
  Nfree(sand_tex_fname);
  Nfree(undersea_tex_fname);
  Nfree(sats[0]);
  Nfree(sats[1]);
  Nfree(sats[2]);
  /* add other char*... */

  Nprintf("Level loaded.\n");

  return valid_level;
}

/* Make the display lists for the lakes */
void init_lakes(Game *game, Lake **lakes)
{
  World *world;
  Map *map;
  int i, j, nlakes;
  GLfloat lake_colour[4]={.3, .4, 1., .5};

  if(!game)
    return;

  world = game->world;
  map = game->map;

  Nprintf("  Initializing lake(s)...\n");
  fflush(stdout);
  if(lakes == NULL || map == NULL)
    return;

  /* Count how many lakes there are */
  for(nlakes = 0; lakes[nlakes] != NULL; nlakes++);

  if(nlakes == 0)
  {
    Nprintf("  Lake(s) initialized\n");
    return;
  }

  /* Generate nlakes consecutive display lists */
  world->lists.lakes = glGenLists(nlakes);

  if(world->lists.lakes == 0)
  {
    Neprintf("    Unable to generate %d display lists for lake(s), aborting\n", nlakes);
    return;
  }

  Nprintf("    Using display lists %d-%d for lake(s)\n",
	  world->lists.lakes, world->lists.lakes + nlakes - 1);

  for(i = 0; i < nlakes; i++)
  {
    glNewList(world->lists.lakes+i, GL_COMPILE);

    glPushAttrib(GL_COLOR_BUFFER_BIT);
    glColor4fv(map->lake_colour);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, lake_colour);
    glMaterialf(GL_FRONT, GL_SHININESS, 128.0f);
    glDisable(GL_TEXTURE_2D);
    glNormal3f(0., 1., 0.);

    glBegin(GL_POLYGON);
    for(j=0; j<lakes[i]->nb_c; j++)
      glVertex3f(lakes[i]->cx[j]*map->x_resolution,
		 lakes[i]->altitude,
		 lakes[i]->cy[j]*map->y_resolution);

    glEnd();

    glPopAttrib();

    glEndList();
  }

  Nprintf(" Lakes initialized\n");
}

/* Make the display list for the sky */
#define nMAX(a,b)  ((a)>(b)?(a):(b))
void init_sky(Game *game)
{
  Map *map;
  float angle;
  float lng;

  if(!game)
    return;

  map = game->map;

  Nprintf("  Initializing sky...\n");
  fflush(stdout);

  lng = nMAX(map->height, map->width);

  /* Create a display list for the sky */
  game->world->lists.sky = glGenLists(1);
  Nprintf("    Using display list %d for sky\n", game->world->lists.sky);

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  glNewList(game->world->lists.sky, GL_COMPILE);

  glColor4f(1.0, 1.0, 1.0, 1.0);
  glBegin(GL_TRIANGLES);

  for(angle=0; angle<2.*M_PI; angle += 2.*M_PI/16.)
  {
    glNormal3f(-cos((double)angle), -0.05, -sin((double)angle));
    glTexCoord2f(.5 + cos((double)angle+2.*M_PI/16.)*.5, .5 + sin((double)angle+2.*M_PI/16.)*.5);
    glVertex3f(map->width/2. + cos((double)angle+2.*M_PI/16.)*lng*4.,
               0.,
	       map->height/2. + sin((double)angle+2.*M_PI/16.)*lng*4.);
    glTexCoord2f(.5, .5);
    glVertex3f(map->width/2., lng*4., map->height/2.);
    glTexCoord2f(.5 + cos((double)angle)*.5, .5 + sin((double)angle)*.5);
    glVertex3f(map->width/2. + cos((double)angle)*lng*4.,
               0.,
	       map->height/2. + sin((double)angle)*lng*4.);
  }

  glEnd();

  glEndList();

  glPopAttrib();

  Nprintf("  Sky initialized\n");
}
#undef nMAX

/* free the sky list */
void free_sky(Game *game)
{/* to be done */
}
