/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* level_list.c
   Manage a list of levels and their snapshots etc */

#include <stdlib.h>
#include <string.h>

#include "level_list.h"
#include "text.h"
#include "tools.h"
#include "file.h"
#include "world.h"

static LevelEntry *allocate_level_entry(void);
static int set_level_entry_name(LevelEntry *entry, const char *name);
static int set_level_ss_name(LevelEntry *entry, const char *name);
static int set_level_file_name(LevelEntry *entry, const char *name);
static LevelEntry *add_level_entry(World *world);
static int free_level_entry(LevelEntry *entry);

/* Allocate space for a level entry */
static LevelEntry *allocate_level_entry(void)
{
  LevelEntry *entry;

  entry = malloc(sizeof(LevelEntry));
  if(!entry)
  {
    Neprintf("Error: Unable to allocate space for level entry\n");
    return NULL;
  }

  entry->name = NULL;
  entry->snapshot = 0;
  entry->snapshot_name = NULL;
  entry->file_name = NULL;

  entry->last = NULL;
  entry->next = NULL;

  return entry;
}

/* Set the name of the level in the level entry */
static int set_level_entry_name(LevelEntry *entry, const char *name)
{
  if(!entry)
    return FALSE;

  Nfree(entry->name);

  if(!name)
    entry->name = strdup("");
  else
    entry->name = strdup(name);

  return TRUE;
}

/* Set the name of the snapshot in the level entry */
static int set_level_ss_name(LevelEntry *entry, const char *name)
{
  if(!entry)
    return FALSE;

  Nfree(entry->snapshot_name);

  if(!name)
    entry->snapshot_name = strdup("");
  else
    entry->snapshot_name = strdup(name);

  return TRUE;
}

/* Set the name of the level file in the level entry */
static int set_level_file_name(LevelEntry *entry, const char *name)
{
  if(!entry)
    return FALSE;

  Nfree(entry->file_name);

  if(!name)
    entry->file_name = strdup("");
  else
    entry->file_name = strdup(name);

  return TRUE;
}

/* Add a level entry */
static LevelEntry *add_level_entry(World *world)
{
  LevelEntry *entry, *entry_ptr;

  if(!world)
    return NULL;

  entry = allocate_level_entry();
  if(!entry)
    return NULL;

  if(!world->levels)
    world->levels = entry;
  else
  {
    /* Append this entry */
    entry_ptr = world->levels;
    while(entry_ptr->next)
      entry_ptr = entry_ptr->next;
    entry_ptr->next = entry;
    entry->last = entry_ptr;
  }

  return entry;
}

/* Load a list of availiable levels */
int load_level_list(World *world, const char *fname)
{
  FILE *f;
  LevelEntry *entry = NULL;
  int nlevels = 0, in_begin = FALSE;
  char *keyw = NULL, *w = NULL;

  Nprintf("Loading level list...\n");

  /* open the file */
  if((f = fopen(fname, "r")) == NULL)
  {
    Neprintf("  Cannot open levels list file '%s'.\n", fname);
    return FALSE;
  }

  while((keyw = get_file_word(f)) != NULL)
  {
    if(!in_begin)
    {
      if(strcmp(keyw, "begin") == 0)
      {
	in_begin = TRUE;
	nlevels++;
	entry = add_level_entry(world);
      }
      else
      {
        Neprintf("  Error: '%s' found in levels file, 'begin' expected.\n");
        Nfree(keyw);
        fclose(f);
        return FALSE;
      }
    }
    else
    {
      if(strcmp(keyw, "end") == 0)
	in_begin = FALSE;
      else
      {
        if(strcmp(keyw, "level") == 0)
	{
	  w = get_file_word(f);
	  set_level_file_name(entry, make_absolute_fname(fname, w));
	  Nfree(w);
	}
        else if(strcmp(keyw, "snapshot") == 0)
	{
	  w = get_file_word(f);
	  set_level_ss_name(entry, make_absolute_fname(fname, w));
	  Nfree(w);
	}
        else if(strcmp(keyw, "name") == 0)
	  set_level_entry_name(entry, get_file_word(f));
        else
        {
	  Neprintf("  Error: unknown option '%s'.\n", keyw);
          Nfree(keyw);
          fclose(f);

          return FALSE;
	}
      }
      Nfree(keyw);
    }
  }

  if(nlevels == 0)
  {
    Neprintf("  No levels found in level list!\n");
    fclose(f);

    return FALSE;
  }
  else
    Nprintf("  %d level(s) found.\n", nlevels);

  /* Create the snapshots */
  entry = world->levels;
  while(entry)
  {
    if(!entry->snapshot_name)
      Neprintf("  No snapshot for level '%s'.\n", entry->name);
    else
    {
      entry->snapshot = load_gl_texture(entry->snapshot_name, FALSE);
      if(!entry->snapshot)
        Neprintf("  Warning: cannot read snapshot '%s'.\n", entry->snapshot_name);
    }

    entry = entry->next;
  }

  fclose(f);
  return(TRUE);
}

/* Remove all the availiable levels */
int clear_level_list(World *world)
{
  LevelEntry *entry, *tmp;

  if(!world)
    return FALSE;

  entry = world->levels;
  while(entry)
  {
    tmp = entry->next;
    free_level_entry(entry);
    entry = tmp;
  }

  world->levels = NULL;

  return TRUE;
}

/* Free a level entry */
static int free_level_entry(LevelEntry *entry)
{
  if(!entry)
    return FALSE;

  Nfree(entry->name);
  glDeleteTextures(1, &entry->snapshot);
  Nfree(entry->snapshot_name);
  Nfree(entry->file_name);

  return TRUE;
}
