/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include <GL/gl.h>

#include "main.h"
#include "world.h"

#include "tools.h"
#include "text.h"
#include "cursor.h"

#include "glut_interface.h"

/* Things to init */
#include "keymap.h"
#include "config.h"
#include "videoc.h"
#include "start.h"
#include "2d.h"
#include "about.h"
#include "draw_map.h"
#include "bird.h"
#include "tanks.h"
#include "plasma.h"

/* Describe the program */
void usage(char *binary_name)
{
  printf("\n"
         "n.a.t.o.w. version " NATOW_VERSION_STRING "\n"
	 "\n"
	 "Description:\n"
	 "  n.a.t.o.w. is a 3D tank game.\n"
	 "Authors:\n"
	 "  * neutronium (Robert Ancell)\n"
	 "  * extropy    (Joel Pitt)\n"
	 "  * hexasoft   (Yannick Perret)\n"
	 "Usage:\n"
	 "  %s [options]\n"
	 "Options are:\n"
	 "  --help       : This message\n"
	 "  --version    : n.a.t.o.w version\n"
	 "  --quiet      : Do not print error/info messages\n"
	 "  --low-detail : Start n.a.t.o.w with minimal detail\n"
	 "  --hi-detail  : Start n.a.t.o.w with maximum detail\n"
	 "\n"
	 "Send bugs to natow-devel@lists.sourceforge.net\n"
	 "\n", binary_name);
}

/* Parse the arguments from the command line */
void parse_arguments(World *world, int argc, char **argv)
{
  int i;

  /* parsing options */
  for(i = 1; i < argc; i++)
  {
    if(argv[i][0] == '-') /* options */
    {
      if((strcmp(argv[i], "-quiet") == 0) ||
	 (strcmp(argv[i], "--quiet") == 0))
	set_quiet_mode(TRUE);
      else if(((strcmp(argv[i], "-debug") == 0) ||
	       (strcmp(argv[i], "--debug") == 0)) &&
	      (i+1 < argc))
	set_debug_level(atoi(argv[i++]));
      else if ((strcmp(argv[i], "--version") == 0)||
	       (strcmp(argv[i], "-version") == 0)||
	       (strcmp(argv[i], "--v") == 0)||
	       (strcmp(argv[i], "-v") == 0))
      {
	printf(NATOW_VERSION_STRING "\n");
	exit(0);
      }
      else if ((strcmp(argv[i], "--help") == 0)||
	       (strcmp(argv[i], "-help") == 0)||
	       (strcmp(argv[i], "--h") == 0)||
	       (strcmp(argv[i], "-h") == 0))
      {
	usage(argv[0]);
	exit(0);
      }
      else if (!strcmp(argv[i], "--low-detail")||
	       !strcmp(argv[i], "--low_detail")||
	       !strcmp(argv[i], "--ld")||
	       !strcmp(argv[i], "-L"))
      {
	world->low_detail = TRUE;
      }
      else if (!strcmp(argv[i], "--hi-detail")||
	       !strcmp(argv[i], "--hi_detail")||
	       !strcmp(argv[i], "--hd")||
	       !strcmp(argv[i], "-H"))
      {
	world->hi_detail = TRUE;
      }
      else
	Neprintf("Unknown option '%s'. Ignored.\n", argv[i]);
    }
    else  /* nothing here now */
      Neprintf("Unknown parameter '%s'. Ignored.\n", argv[i]);
  }
}

/* Set up n.a.t.o.w */
void init_natow(World *world)
{
  /* Init random generator */
  srand((int)time(NULL));

  /* Because glut wont pass any extra args to display(), mouse() etc */
  set_glut_world(world);

  /* Set up openGL */
  init_gl(world);
  
  /* Initialize things that are needed right away */
  init_start(world);
  init_2d(world);
  init_draw_map(world);
  init_birds(world);
  init_cursor(world);
  init_about(world);
  init_plasma(world);

  /* TODO: Move these to somewhere more appropriated (and above ones) */
  /* Load the levels */
  load_level_list(world, "levels/list.levs");

  /* Load the tanks */
  load_tanks(world, "tanks");

  /* Load the munitons */
  world->munition_list = load_munition_list("munitions");

  /* Make availiable (for firing by units) the wanted ones */
  load_usable_munitions("usable", world);

  /* load the tank models (temp) */
  generate_tank_lists(world);

  /* Make GLUT use display etc */
  update_glut_functions(world);
}

/* Set up openGL */
void init_gl(World *world)
{
  glClearColor(1.0, 1.0, 1.0, 1.0);
  if(world->is_smooth)
    glShadeModel(GL_SMOOTH);
  else
    glShadeModel(GL_FLAT);

  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  if(world->is_light)
    glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
}

/* The very beginning */
int main(int argc, char **argv)
{
  World *world;

  /* And there was light */
  world = new_world();

  parse_arguments(world, argc, argv);

  /* Init the NATOW-config system */
  init_configdir(world);

  /* Load the video-config */
  if(!read_videoc(world))
  {
    Nprintf("  Trying to create a default video-config file...\n");
    if(!write_videoc(world))
      Neprintf("  Cannot create default video-config file!\n");
  }

  /* If specified on command line, start of in ultra-low detail */
  if(world->low_detail)
    set_low_detail(world);

  /* If specified on command line, start of in ultra-hi detail */
  if(world->hi_detail)
    set_hi_detail(world);

  /* Set GLUT up */
  init_glut(&argc, argv);

  /* Load the key-config */
  if(!read_keylist(world))
  {
    Nprintf("  No file. Creating a default key-map file...\n");
    if(!save_keylist(world))
      Neprintf("    Error creating key-map file!\n");
  }

  make_glut_window(world->video_size_x, world->video_size_y);
  if(world->fullscreen)
    fullscreen_glut();

  init_natow(world);

  /* Start GLUT */
  start_glut();

  return 0;
}
