/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* map.c
   This file contains the code to interface with the map */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "map.h"
#include "world.h"
#include "vector.h"
#include "image.h"
#include "tools.h"
#include "text.h"
#include "light.h"

static Map *allocate_map(const char *name);
static int build_arrays(Map *map);
static int free_arrays(Map *map);
static float *average_4vectors(float *v1, float *v2, float *v3, float *v4);
static int update_coord_array(Map *map);

/* Allocate space for a map */
static Map *allocate_map(const char *name)
{
  Map *map;

  map = malloc(sizeof(Map));
  if(!map)
  {
    Neprintf("    Unable to allocate space for map\n");
    return NULL;
  }

  map->game = NULL;

  if(!name)
    map->name = strdup("");
  else
    map->name = strdup(name);

  map->x_resolution = 5.0;
  map->y_resolution = 5.0;
  map->size_x = 0;
  map->size_y = 0;

  map->startingpoints = NULL;

  map->width = 0.0;
  map->height = 0.0;
  map->max_altitude = 51.0;

  map->tex_width = 0;
  map->tex_height = 0;
  map->texture_data = NULL;
  map->texture = 0;
  map->tex_scale_x = 0.01 * map->x_resolution; /* 0.01 makes the texture */
  map->tex_scale_y = 0.01 * map->y_resolution; /* repeat every 10m */

  _make_vector(-1.0, 1.0, 1.0, map->sun_dir);
  normalize_vector(map->sun_dir);
  map->sun_dir[3] = 0.0;
  map->sun_diffuse[0] = 0.7;
  map->sun_diffuse[1] = 0.7;
  map->sun_diffuse[2] = 0.6;
  map->sun_diffuse[3] = 1.0;
  map->sun_ambient[0] = 0.2;
  map->sun_ambient[1] = 0.2;
  map->sun_ambient[2] = 0.2;
  map->sun_ambient[3] = 1.0;

  map->rock_tex_fname = NULL;
  map->ice_tex_fname = NULL;
  map->tree_tex_fname = NULL;
  map->grass_tex_fname = NULL;
  map->sand_tex_fname = NULL;
  map->undersea_tex_fname = NULL;
  map->rock_texture = NULL;
  map->ice_texture = NULL;
  map->tree_texture = NULL;
  map->grass_texture = NULL;
  map->sand_texture = NULL;
  map->undersea_texture = NULL;

  map->sky_texture = 0;

  map->pos = NULL;
  map->norm = NULL;

  map->objects = NULL;
  map->lakes = NULL;

  map->lake_colour[0] = 0.3;
  map->lake_colour[1] = 0.4;
  map->lake_colour[2] = 1.0;
  map->lake_colour[3] = 0.25;

  map->sea_colour[0] = 0.0;
  map->sea_colour[1] = 0.1;
  map->sea_colour[2] = 1.0;
  map->sea_colour[3] = 0.25;

  map->static_light = NULL;
  map->dynamic_light = NULL;

  map->birds_number = 0;

  /* default fog attribute */
  map->fog_colour[0] = 1.0;
  map->fog_colour[1] = 1.0;
  map->fog_colour[2] = 1.0;
  map->fog_colour[3] = 1.0;
  map->fog_density = 0.005;
  map->fog_end     = 250.;

  return map;
}

/* Build the altitude/normal/lighting etc arrays */
static int build_arrays(Map *map)
{
  int w, h;
  float *tmp = NULL;

  if(!map)
    return FALSE;

  /* Free any existing arrays */
  free_arrays(map);

  /* Map objects */
  map->objects = malloc(sizeof(Object **) * map->size_x);
  for(w = 0; w < map->size_x; w++)
  {
    map->objects[w] = malloc(sizeof(Object *) * map->size_y);
    for(h = 0; h < map->size_y; h++)
      map->objects[w][h] = NULL;
  }

  /* Map co-ordinates */
  map->pos = malloc(sizeof(float **) * map->size_x);
  for(w = 0; w < map->size_x; w++)
  {
    map->pos[w] = malloc(sizeof(float *) * map->size_y);
    tmp = malloc(sizeof(float) * 3 * map->size_y);
    for(h = 0; h < map->size_y; h++)
      map->pos[w][h] = &(tmp[h*3]);
  }

  /* Map normals */
  map->norm = malloc(sizeof(float **) * map->size_x);
  for(w = 0; w < map->size_x; w++)
  {
    map->norm[w] = malloc(sizeof(float *) * map->size_y);
    tmp = malloc(sizeof(float) * 3* map->size_y);
    for(h = 0; h < map->size_y; h++)
      map->norm[w][h] = &(tmp[h*3]);
  }

  /* Static lighting */
  map->static_light = malloc(sizeof(float **) * map->size_x);
  for(w = 0; w < map->size_x; w++)
  {
    map->static_light[w] = malloc(sizeof(float *) * map->size_y);
    tmp = malloc(sizeof(float)*3*map->size_y);
    for(h = 0; h < map->size_y; h++)
      map->static_light[w][h] = &(tmp[h*3]);
  }

  /* Dynamic lighting */
  map->dynamic_light = malloc(sizeof(float **) * map->size_x);
  for(w = 0; w < map->size_x; w++)
  {
    map->dynamic_light[w] = malloc(sizeof(float *) * map->size_y);
    tmp = malloc(sizeof(float)*3*map->size_y);
    for(h = 0; h < map->size_y; h++)
      map->dynamic_light[w][h] = &(tmp[h*3]);
  }

  /* Precalculated dynamic lighting */
  init_lighting(map);

  if(!map->pos || !map->norm || !map->static_light || !map->dynamic_light)
  {
    Neprintf("    Error allocating space for map components\n");
    return FALSE;
  }

  return TRUE;
}

/* Free the altitude/normal/lighting etc arrays */
static int free_arrays(Map *map)
{
  unsigned int w;

  /* Map objects */
  if(map->objects)
    for(w = 0; w < map->size_x; w++)
      Nfree(map->objects[w]);
  Nfree(map->objects);
  map->objects = NULL;

  /* Map co-ordinates */
  if(map->pos)
    for(w = 0; w < map->size_x; w++)
    {
      Nfree(map->pos[w][0]);
      Nfree(map->pos[w]);
    }
  Nfree(map->pos);
  map->pos = NULL;

  /* Map normals */
  if(map->norm)
    for(w = 0; w < map->size_x; w++)
    {
      Nfree(map->norm[w][0]);
      Nfree(map->norm[w]);
    }
  Nfree(map->norm);
  map->norm = NULL;

  /* Static lighting */
  if(map->static_light)
    for(w = 0; w < map->size_x; w++)
    {
      Nfree(map->static_light[w][0]);
      Nfree(map->static_light[w]);
    }
  Nfree(map->static_light);
  map->static_light = NULL;

  /* Dynamic lighting */
  if(map->dynamic_light)
    for(w = 0; w < map->size_x; w++)
    {
      Nfree(map->dynamic_light[w][0]);
      Nfree(map->dynamic_light[w]);
    }
  Nfree(map->dynamic_light);
  map->dynamic_light = NULL;

  return TRUE;
}

/* Free map (and associated) arrays */
int free_map(Map *map)
{
  if(!map)
    return FALSE;

  Nfree(map->name);

  free_arrays(map);

  return TRUE;
}

/* Loads a map file */
Map *load_map(char *fname, float max_altitude)
{
  IMAGE *elevImg;
  Map *map;
  unsigned int w, h, pad;
  float scale;

  map = allocate_map("");
  if(!map)
    return NULL;

  elevImg = ImageLoad(fname);
  if(!elevImg)
  {
    Neprintf("    Error opening file:%s\n", fname);
    return NULL;
  }

  map->size_x = elevImg->sizeX;
  map->size_y = elevImg->sizeY;
  /* Map edges must be a multiple of 4 + 1 (to make the LOD easier) */
  map->size_x -= (map->size_x - 1)% 4;
  map->size_y -= (map->size_y - 1)% 4;
  map->width = map->x_resolution * (float)(map->size_x - 1);
  map->height = map->y_resolution * (float)(map->size_y - 1);

  /* Build the arrays for the map size */
  build_arrays(map);

  /* Copy the data into the map */
  map->max_altitude = max_altitude;
  scale = max_altitude / 255.0;
  pad = 0;
  for(h = 0; h < map->size_y; h++)
  {
    for(w = 0; w < map->size_x; w++)
      map->pos[w][map->size_y - h - 1][1] =
	scale * elevImg->data[(w + elevImg->sizeX *  h) * elevImg->sizeZ + pad];
    pad += (elevImg->sizeX * elevImg->sizeZ) % 4;
  }
  Nprintf("    Map Loaded (%dx%d)\n", map->size_x, map->size_y);

  /* Free the image */
  Nfree(elevImg->data);
  Nfree(elevImg);

  /* force borders to be at altitude 0 */
  for(w = 0; w < map->size_x; w++)
  {
    map->pos[w][0][1] = 0.0;
    map->pos[w][map->size_y-1][1] = 0.0;
  }
  for(h = 0; h < map->size_y; h++)
  {
    map->pos[0][h][1] = 0.0;
    map->pos[map->size_x-1][h][1] = 0.0;
  }

  /* Calculate the map normals */
  calc_normals(map);

  /* Update the co-ordinate array */
  update_coord_array(map);

  return map;
}

/* Randomizes a map */
/* TODO: Make much more complicated */
void randomize_map(Map *map)
{
  int w, h;

  for(w = 0; w < map->size_x; w++)
    for(h = 0; h < map->size_y; h++)
      map->pos[w][h][1] = 50.0 * rand()/(RAND_MAX + 1.0);
}

/* Calculates the normals for the whole map */
void calc_normals(Map *map)
{
  int x, y;

  /* Calculate the normals for the main part of the map */
  for(x = 0; x < map->size_x; x++)
    for(y = 0; y < map->size_y; y++)
      calc_normal(map, x, y);

  Nprintf("    Normals calculated\n");
}

/* Calculates the map normal at the co-ordinates (x, y) */
void calc_normal(Map *map, int x, int y)
{
  float h, m1[3], m2[3], m3[3], m4[3], *v;

  h = map->pos[x][y][1];

  if(x == (map->size_x - 1))
  {
    m1[0] = 0.0;
    m1[1] = 0.0;
    m1[2] = 0.0;
  }
  else
  {
    m1[0] = h - map->pos[x + 1][y][1];
    m1[1] = 1.0;
    m1[2] = 0.0;
  }

  if(x == 0)
  {
    m2[0] = 0.0;
    m2[1] = 0.0;
    m2[2] = 0.0;
  }
  else
  {
    m2[0] = map->pos[x - 1][y][1] - h;
    m2[1] = 1.0;
    m2[2] = 0.0;
  }

  if(y == (map->size_y - 1))
  {
    m3[0] = 0.0;
    m3[1] = 0.0;
    m3[2] = 0.0;
  }
  else
  {
    m3[0] = 0.0;
    m3[1] = 1.0;
    m3[2] = h - map->pos[x][y + 1][1];
  }

  if(y == 0)
  {
    m4[0] = 0.0;
    m4[1] = 0.0;
    m4[2] = 0.0;
  }
  else
  {
    m4[0] = 0.0;
    m4[1] = 1.0;
    m4[2] = map->pos[x][y - 1][1] - h;
  }

  v = average_4vectors(m1, m2, m3, m4);
  map->norm[x][y][0] = v[0];
  map->norm[x][y][1] = v[1];
  map->norm[x][y][2] = v[2];
}

/* Gets the normalized average of four vectors */
static float *average_4vectors(float *v1, float *v2, float *v3, float *v4)
{
  static float v[3];
  float va[3], vb[3];

  normalize_vector(v1);
  normalize_vector(v2);
  normalize_vector(v3);
  normalize_vector(v4);

  _vector_add(v1, v2, va);
  normalize_vector(va);
  _vector_add(v3, v4, vb);
  normalize_vector(vb);
  _vector_add(va, vb, v);
  normalize_vector(v);

  return v;
}

/* Update the map co-ordinate array, i.e. scale the positions by the
   resolutions */
static int update_coord_array(Map *map)
{
  int x, y;

  if(!map)
    return FALSE;

  for(x = 0; x < map->size_x; x++)
    for(y = 0; y < map->size_y; y++)
    {
      map->pos[x][y][0] = (float)x * map->x_resolution;
      map->pos[x][y][1] = map->pos[x][y][1];
      map->pos[x][y][2] = (float)y * map->y_resolution;
    }

  return TRUE;
}

/* Gets the intersection when there's a large distance between the two points */
/*float *get_map_intersection_long(Map *map, float *posa, float *posb, float error)
{
  int width, height, xres, yres;
  float pos0[3], pos1[3];
  float p0[3], p1[3];
  float dx, dy, dz;
  float xf, yf;
  int x, y, x0, x1, y0, y1;
  float ha, hb, h0, h1, h2, h3;
  float incx, incy, incz;*/

  /* The dimensions of the map */
  /*width = map->size_x;
  height = map->size_y;
  xres = map->x_resolution;
  yres = map->y_resolution;*/

  /* Copy these vectors */
  /*_vector_dup(posa, pos0);
  _vector_dup(posb, pos1);*/

  /* The change in distance between the two points */
  /*dx = pos1[0] - pos0[0];
  dy = pos1[1] - pos0[1];
  dz = pos1[2] - pos0[2];*/

  /* Get the map co-ordinates of these points */
  /*x0 = (int)posa[0];
  x1 = (int)(posb[0] + 0.5);
  y0 = (int)posa[2];
  y1 = (int)(posb[2] + 0.5);

  _vector_dup(pos0, p0);
  _vector_dup(pos1, p1);*/

  /* Follow the line in sections of 1x1 squares */
  /*if(dx > dz)
  {*/
    /* The amount to increase in each direction for one step */
    /*incx = 1.0;
    incy = dy / dx;
    incz = dz / dx;

    for(x = x0; x < x1; x++)
    {*/
      /* Step along the line one unit */
      /*p1[0] += incx;
      p1[1] += incy;
      p1[2] += incz;*/

      /* Low map grid to actual y direction */
      /*ya = (int)p0[2];
      yb = (int)(p1[2] + 0.5);*/

      /* The heights of the map in this cube */
      /*h0 = map->pos[x][y][1];
      h1 = map->pos[x + 1][y][1];
      h2 = map->pos[x][y + 1][1];
      h3 = map->pos[x + 1][y + 1][1];*/

      /* Find the minimum height */
      /*h = ha < hb ? ha : hb;*/

      /* If below one of the map heights, there may be a collision */
      /*if(h < h0 || h < h1 || h < h2 || h < h3)
      {
	p = get_map_intersection(map, p0, p1, error);*/

	/* Collision occured */
	/*if(p)
	  return _vector_dup(p, pos);
      }*/

      /* Copy this vector to be the last vector */
      /*p0[0] = p1[0];
      p0[1] = p1[1];
      p0[2] = p1[2];
    }
  }
  else
  {
    for(y = y0; y < y1; y++)
    {
    }
  }*/

  /* Clip to the map */
  /*if(x0 < 0)
    x0 = 0;
  else if(x0 > width)
    x0 = width;
  if(y0 < 0)
    y0 = 0;
  else if(y0 > height)
    y0 = height;*/

  /* Work along the map co-ordinates checking for collision */
  /*for(x = x0; x < x1; x++)
  {
    for(y = y0; y < y1; y++)
    {
    }
  }
}*/

/* Gets the intersection when there's a large distance between the two points */
float *get_map_intersection_long(Map *map, float *posa, float *posb, float error)
{
  static float pos[3];
  float pos0[3], pos1[3], pos2[3], old_pos2[3], t[3], chord[3], length, *p = NULL;
  float u, step, min_step = 5.0;

  _vector_dup(posa, pos0);
  _vector_dup(posb, pos1);
  _vector_sub(pos0, pos1, chord);

  /*printf("(%f, %f, %f)->(%f, %f, %f)\n",
	 pos0[0], pos0[1], pos0[2],
	 pos1[0], pos1[1], pos1[2]);
  printf("(%f, %f, %f)\n",
	 chord[0], chord[1], chord[2]);*/

  /* !!note: edge of map clipping disabled -- doesn't appear to be working,
     locks up the collision algorithm */

  /* Clip the final point to within boundaries */
  /*if(pos1[0] < -1000.0)
  {
    _vector_scale((pos0[0] + 1000.0) / chord[0], chord, t);
    _vector_sub(pos0, t, pos1);
    _vector_sub(pos0, pos1, chord);
  }
  else if(pos1[0] > 1000.0 + map->width)
  {
    _vector_scale(-(map->width + 1000.0 - pos0[0]) / chord[0], chord, t);
    _vector_sub(pos0, t, pos1);
    _vector_sub(pos0, pos1, chord);
  }

  if(pos1[2] < -1000.0)
  {
    _vector_scale((pos0[2] + 1000.0) / chord[2], chord, t);
    _pos1 = vector_sub(pos0, t, pos1);
    _chord = vector_sub(pos0, pos1, chord);
  }
  else if(pos1[2] > 1000.0 + map->height)
  {
    _vector_scale(-(map->height + 1000.0 - pos0[0]) / chord[2], chord, t);
    _vector_sub(pos0, t, pos1);
    _vector_sub(pos0, pos1, chord);
  }*/

  /* Make the final point above 0 altitude */
  if(pos1[1] < -1000.0)
  {
    _vector_scale((pos0[1] + 1000.0) / chord[1], chord, t);
    _vector_sub(pos0, t, pos1);
    _vector_sub(pos0, pos1, chord);
  }

  /* Clip the initial point to within (close to) the boundaries */
  /*if(pos0[0] < -1000.0)
  {
    _vector_scale((pos1[0] + 1000.0) / chord[0], chord, t);
    _vector_add(pos1, t, pos0);
    _vector_sub(pos0, pos1, chord);
  }
  else if(pos0[0] > 1000.0 + map->width)
  {
    _vector_scale((map->width + 1000.0 - pos1[0]) / chord[0], chord, t);
    _vector_add(pos1, t, pos0);
    _vector_sub(pos0, pos1, chord);
  }

  if(pos0[2] < -1000.0)
  {
    _vector_scale((pos1[2] + 1000.0) / chord[2], chord, t);
    _vector_add(pos1, t, pos0);
    _vector_sub(pos0, pos1, chord);
  }
  else if(pos0[2] > 1000.0 + map->height)
  {
    _vector_scale((map->height + 1000.0 - pos1[0]) / chord[2], chord, t);
    _vector_add(pos1, t, pos0);
    _vector_sub(pos0, pos1, chord);
  }*/

  /* Make the initial point no greater than the highest altitude */
  if(pos0[1] > map->max_altitude)
  {
    _vector_scale((map->max_altitude - pos1[1]) / chord[1], chord, t);
    _vector_add(pos1, t, pos0);
    _vector_sub(pos0, pos1, chord);
  }

  /* TEMP: For debugging puropses */
  /*printf("(%f, %f, %f)->(%f, %f, %f)\n",
	 pos0[0], pos0[1], pos0[2],
	 pos1[0], pos1[1], pos1[2]);*/
  /*glVertex3f(pos0[0], pos0[1], pos0[2]);
  glVertex3f(pos1[0], pos1[1], pos1[2]);*/

  /* Get the length of the chord */
  length = vector_2norm(chord);

  /* Divide the ray into smaller pieces and check them until
     collision */
  step = min_step / length;
  _vector_dup(pos0, old_pos2);
  for(u = step; !p && (u < 1.0); u += step)
  {
    _vector_scale(u, chord, t);
    _vector_sub(pos0, t, pos2);
    p = get_map_intersection(map, old_pos2, pos2, error);

    _vector_dup(pos2, old_pos2);
  }

  if(p == NULL)
    return NULL;
  else
    return _vector_dup(p, pos);
}

/* Returns the point of intersection of a line and the map */
float *get_map_intersection(Map *map, float *posa, float *posb, float error)
{
  /* The position of the collision */
  static float pos[3];
  float h0, h1, h2, pos0[3], pos1[3], d[3], length;

  /* If no error specified use default */
  if(error < 0.0)
    error = 0.01;

  /* Copy the start and end point */
  pos0[0] = posa[0];
  pos0[1] = posa[1];
  pos0[2] = posa[2];
  pos1[0] = posb[0];
  pos1[1] = posb[1];
  pos1[2] = posb[2];

  /* The map altitude at each point */
  h0 = get_height(map, pos0[0], pos0[2]);
  h1 = get_height(map, pos1[0], pos1[2]);

  /*printf("%f %f\n", h1, pos1[1]);*/

  /* If the final position isn't underground, or the initial position is */
  if(h1 < pos1[1] || h0 > pos0[1])
    return NULL;

  /* If final point inside boundaries */
  if((pos1[0] >= 0.0)
     && (pos1[0] < map->width)
     && (pos1[2] >= 0.0)
     && (pos1[2] < map->height))
  {
    /* Get the location of the collision */
    do {
      pos[0] = 0.5f * (pos0[0] + pos1[0]);
      pos[1] = 0.5f * (pos0[1] + pos1[1]);
      pos[2] = 0.5f * (pos0[2] + pos1[2]);

      h2 = get_height(map, pos[0], pos[2]);
      /*printf("%f %f\n", pos2[1], h2);
      printf("%f %f %f\n%f %f %f\n%f %f %f\n",
	     pos0[0], pos0[1], pos0[2],
	     pos1[0], pos1[1], pos1[2],
	     pos2[0], pos2[1], pos2[2]);*/

      /* Sub-divide the region */
      if((pos[1] - h2) < 0.0f)
      {
	pos1[0] = pos[0];
	pos1[1] = pos[1];
	pos1[2] = pos[2];
      } else {
	pos0[0] = pos[0];
	pos0[1] = pos[1];
	pos0[2] = pos[2];
      }

      /* Note: this seems unnecessary, but problems sometimes without it */
      _vector_sub(pos0, pos1, d);
      length = vector_infnorm(d);
    } while((length > error) && ((float)fabs(pos[1] - h2) > error));
  }
  else
    return NULL;

  return pos;
}

/* Gets the height of the map at an arbitrary co-ordinate (x, y) */
float get_height(Map *map, float x, float y)
{
  float z, z0, z1, z2, xr, yr;
  int x0, y0;

  /* One map sector:    */
  /* p is at (x, y)     */
  /*                    */
  /*     xr             */
  /*    <->             */
  /* z0 +----+ z1  ^ yr */
  /*    |   /|     |    */
  /*    | p/ |    \|/   */
  /*    | /  |     `    */
  /*    |/   |          */
  /* z2 +----+ z0'      */
  /*    x0   x1         */

  /* Check bounds */
  if((x >= map->width) || (x < 0.0) || (y >= map->height) || (y < 0.0))
    return 0.0;

  /* Scale the distance down to the resolution of the map */
  x /= map->x_resolution;
  y /= map->y_resolution;

  x0 = (int) x;
  y0 = (int) y;
  xr = x - (float)x0;
  yr = y - (float)y0;
  z1 = map->pos[x0 + 1][y0][1];
  z2 = map->pos[x0][y0 + 1][1];

  if(xr + yr < 1) /* upper triangle */
  {
    z0 = map->pos[x0][y0][1];
    z = xr * (z1 - z0) + yr * (z2 - z0) + z0;
  }
  else /* lower triangle */
  {
    z0 = map->pos[x0 + 1][y0 + 1][1];
    z = (1.0 - xr) * (z2 - z0) + (1.0 - yr) * (z1 - z0) + z0;
  }

  return z;
}

/* Gets a angular normal at an arbitrary co-ordinate (x, y),
   (Value is used internally -- do not free) */
float *get_normal(Map *map, float x, float y)
{
  float dx[3], dz[3];
  static float n[3];
  int xr = (int)(x / map->x_resolution), yr = (int)(y / map->y_resolution);
  float h;

  /* One map sector: */
  /*                 */
  /* (xr,yr)    >+x  */
  /*  h +----+       */
  /*    |   /|       */
  /*    |  / |       */
  /*    | /  |       */
  /* +z |/   |       */
  /* \/ +----+ h'    */
  /*    (xr+1,yr+1)  */

  /* Check bounds */
  if((x >= map->width) || (x < 0.0) || (y >= map->height) || (y < 0.0))
  {
    /* It's flat everywhere else */
    n[0] = 0.0;
    n[1] = 1.0;
    n[2] = 0.0;
    return n;
  }

  if((x - xr) * (y - yr) < 1) /* upper triangle */
  {
    h = map->pos[xr][yr][1];

    dx[0] = map->x_resolution;
    dx[1] = map->pos[xr + 1][yr][1] - h;
    dx[2] = 0.0f;
    dz[0] = 0.0f;
    dz[1] = map->pos[xr][yr + 1][1] - h;
    dz[2] = map->y_resolution;
    normalize_vector(dx);
    normalize_vector(dz);
    _vector_cross(dz, dx, n);
  } else { /* lower triangle */
    h = map->pos[xr + 1][yr + 1][1];

    dx[0] = map->x_resolution;
    dx[1] = h - map->pos[xr][yr + 1][1];
    dx[2] = 0.0f;
    dz[0] = 0.0f;
    dz[1] = h - map->pos[xr + 1][yr][1];
    dz[2] = map->y_resolution;
    normalize_vector(dx);
    normalize_vector(dz);
    _vector_cross(dz, dx, n);
  }

  normalize_vector(n);
  return n;
}

/* Gets a smoothed normal at an arbitrary co-ordinate (x, y)
   (Value is used internally -- do not free) */
float *get_smooth_normal(Map *map, float x, float y)
{
  static float v[3];
  float *n1, *n2, *n3, *n4;
  int xr = (int)(x / map->x_resolution), yr = (int)(y / map->y_resolution);
  float dx, dy, xd, yd;

  /* One map sector: */
  /*                 */
  /* (xr,yr)    >+x  */
  /*  h +----+       */
  /*    |   /|       */
  /*    |  / |       */
  /*    | /  |       */
  /* +z |/   |       */
  /* \/ +----+ h'    */
  /*    (xr+1,yr+1)  */
  /*                 */
  /* n1 +----+ n2    */
  /*    |   /|       */
  /*    |  / |       */
  /*    | /  |       */
  /*    |/   |       */
  /* n3 +----+ n4    */

  /* Check bounds */
  if((x >= map->width) || (x < 0.0) || (y >= map->height) || (y < 0.0))
    return NULL;

  n1 = map->norm[xr][yr];
  n2 = map->norm[xr+1][yr];
  n3 = map->norm[xr][yr+1];
  n4 = map->norm[xr+1][yr+1];

  dx = x / map->x_resolution - (float)xr;
  dy = y / map->y_resolution - (float)yr;
  xd = 1.0 - dx;
  yd = 1.0 - dy;

  _make_vector(n1[0]*xd*yd + n2[0]*dx*yd + n3[0]*xd*dy + n4[0]*dx*dy,
	       n1[1]*xd*yd + n2[1]*dx*yd + n3[1]*xd*dy + n4[1]*dx*dy,
	       n1[2]*xd*yd + n2[2]*dx*yd + n3[2]*xd*dy + n4[2]*dx*dy, v);

  normalize_vector(v);
  return v;
}

/* TODO: improve this (might not be needed any more though) */
/* Gets the gradient of the map at (x,y) */
float *get_gradient(Map *map, float x, float y)
{
  float *v = malloc(sizeof(float) * 3), h;
  int xr, yr;

  x /= map->x_resolution;
  y /= map->y_resolution;
  xr = (int) x;
  yr = (int) y;

  if((x - xr) * (y - yr) > 1) /* upper triangle */
  {
    h = map->pos[xr + 1][yr + 1][1];

    v[0] = h - map->pos[xr][yr + 1][1];
    v[1] = h - map->pos[xr + 1][yr][1];
    v[2] = h - map->pos[xr + 1][yr][1];
  } 
  else
  {
    h = map->pos[xr][yr][1]; /* !!guess? */

    v[0] = map->pos[xr + 1][yr][1] - h;
    v[1] = map->pos[xr + 1][yr][1] - h;
    v[2] = map->pos[xr][yr + 1][1] - h;
  }

  normalize_vector(v);

  return v;
}

/* NOTE: the following two algo. need to be changed to calculate angles
 * with the tracks actually on the ground */

/* Get the rotation required in the x-axis to put an object normal to
 * the ground */
float get_rotx(Map *map, float x, float y)
{
  float *n = get_smooth_normal(map, x, y);
  float theta = (180.0f / M_PI) * atan(n[2] / n[1]);

  return theta;
}

/* Same as above but in the z-axis */
float get_rotz(Map *map, float x, float y)
{
  float *n = get_smooth_normal(map, x, y);
  float theta = -(180.0f / M_PI) * atan(n[0] / sqrt(n[1]*n[1] + n[2]*n[2]));

  return theta;
}

/* Test whether a point is in the bounds of the map or not,
   returning TRUE if in bounds, FALSE otherwise */
int in_bounds(Map *map, float x, float y)
{
  if(x < 0.0 || y < 0.0 || x >= map->width || y >= map->height)
    return FALSE;
  else
    return TRUE;
}

/* Same as above, but for a 3D vector */
int in_boundsv(Map *map, float *pos)
{
  if(pos[0] < 0.0 || pos[2] < 0.0 || pos[0] >= map->width || pos[2] >= map->height)
    return FALSE;
  else
    return TRUE;
}
