/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* munition.c
   This file contains the code for munitions --
   a chain of explosions/events etc that makes up a weapon */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <ctype.h>

#include "munition.h"
#include "world.h"
#include "file.h"
#include "text.h"
#include "tools.h"

static Munition *allocate_munition(void);
static Munition *new_munition(Munition *munition_list,
			      char *name,
			      MunitionType type,
			      float critical_time);
static MunitionAction *allocate_munition_action(void);
static MunitionAction *munition_add_action(MunitionAction *action,
					   MunitionActionType type,
					   ActionParameter *params);
static void add_munition_spawn_action(Munition *munition,
				      MunitionActionType type,
				      ActionParameter *params);
static void add_munition_timer_action(Munition *munition,
				      MunitionActionType type,
				      ActionParameter *params);
static void add_munition_apex_action(Munition *munition,
				     MunitionActionType type,
				     ActionParameter *params);
static void add_munition_collision_action(Munition *munition,
					  MunitionActionType type,
					  ActionParameter *params);
static int get_action_nparams(MunitionActionType type);
static MunitionActionType get_action_type(char *action_name);
static void set_action_param(MunitionActionType type, ActionParameter *params,
			     int index, char *word);
static MunitionAction *load_munition_actions(FILE *mun_file);
static Munition *load_munition(FILE *mun_file);
static int free_munition(Munition *munition);
static int free_munition_actions(MunitionAction *action_list);
static int free_munition_action(MunitionAction *action);

/* Allocate space for a new munition */
static Munition *allocate_munition(void)
{
  Munition *munition = malloc(sizeof(Munition));

  if(munition == NULL)
  {
    Neprintf("  Unable to allocate space for munition\n");
    return NULL;
  }

  munition->name = NULL;
  munition->type = MUNITION_NONE;
  munition->cost = 0;

  munition->usable = FALSE;

  munition->critical_time = 0.0;

  munition->spawn_actions = NULL;
  munition->timer_actions = NULL;
  munition->apex_actions = NULL;
  munition->collision_actions = NULL;

  munition->next = NULL;
  munition->last = NULL;

  return munition;
}

/* Makes a new munition */
static Munition *new_munition(Munition *munition_list,
			      char *name,
			      MunitionType type,
			      float critical_time)
{
  Munition *munition = allocate_munition();

  if(munition == NULL)
    return NULL;

  /* Set some constants */
  munition->name = strdup(name);
  munition->type = type;
  munition->critical_time = critical_time;

  /* If not the first entry... */
  if(munition_list != NULL)
  {
    /* ...add the munition onto the end of the list */
    while(munition_list->next != NULL)
      munition_list = munition_list->next;
    munition_list->next = munition;

    /*munition->next = munition_list;*/
    munition->last = munition_list;
  }

  return munition;
}

/* Get the munition named 'name' */
Munition *find_munition(Munition *munition_list, char *name)
{
  Munition *mun_ptr = munition_list;

  /* Search through the list */
  do
  {
    if(!strcmp(mun_ptr->name, name))
      return mun_ptr;

    mun_ptr = mun_ptr->next;
  } while(mun_ptr != munition_list);

  /* Failed to find one */
  return NULL;
}

/* Aloocates space for a new action */
static MunitionAction *allocate_munition_action(void)
{
  MunitionAction *new_action = malloc(sizeof(MunitionAction));

  if(new_action == NULL)
  {
    Neprintf("  Unable to allocate space for munition action\n");
    return NULL;
  }

  /* Set defaults */
  new_action->type = MUNITION_UNKNOWN_ACTION;
  new_action->parameters = NULL;

  new_action->next = NULL;

  return new_action;
}

/* Adds a new action */
static MunitionAction *munition_add_action(MunitionAction *action,
					   MunitionActionType type,
					   ActionParameter *params)
{
  MunitionAction *new_action = allocate_munition_action();
  int n_params, i;

  /* Set the type */
  new_action->type = type;

  /* Number of parameters used by each action */
  n_params = get_action_nparams(type);

  /* Allocate space for n_params parameters if necessary */
  if(n_params)
    new_action->parameters = malloc(sizeof(ActionParameter) * n_params);
  else
    new_action->parameters = NULL;

  /* Copy the parameters over */
  for(i = 0; i < n_params; i++)
    new_action->parameters[i] = params[i];

  /* If the first entry in the list, not able to start list so return action */
  if(action == NULL)
    return new_action;

  /* Otherwise attach to the end of the list */
  while(action->next != NULL)
    action = action->next;
  action->next = new_action;

  /* Return it anyway */
  return new_action;
}

/* Adds a fire action -- activated on in order when this munition is fired */
static void add_munition_spawn_action(Munition *munition,
				      MunitionActionType type,
				      ActionParameter *params)
{
  if(munition->spawn_actions == NULL)
    munition->spawn_actions = munition_add_action(munition->spawn_actions,
						  type, params);
  else
    munition_add_action(munition->spawn_actions, type, params);
}

/* Adds a time action -- activated when the munition hits its critical time */
static void add_munition_timer_action(Munition *munition,
				      MunitionActionType type,
				      ActionParameter *params)
{
  if(munition->timer_actions == NULL)
    munition->timer_actions = munition_add_action(munition->timer_actions,
						  type, params);
  else
    munition_add_action(munition->timer_actions, type, params);
}

/* Adds a apex action -- activated when the munition starts decreasing in height */
static void add_munition_apex_action(Munition *munition,
				     MunitionActionType type,
				     ActionParameter *params)
{
  if(munition->apex_actions == NULL)
    munition->apex_actions = munition_add_action(munition->apex_actions,
						    type, params);
  else
    munition_add_action(munition->apex_actions, type, params);
}

/* Adds a collision action -- activated when the projectile collides with something */
static void add_munition_collision_action(Munition *munition,
					  MunitionActionType type,
					  ActionParameter *params)
{
  if(munition->collision_actions == NULL)
    munition->collision_actions = munition_add_action(munition->collision_actions,
						      type, params);
  else
    munition_add_action(munition->collision_actions, type, params);
}

/* Get the number of parameters required by an action */
static int get_action_nparams(MunitionActionType type)
{
  switch(type)
  {
  case MUNITION_UNKNOWN_ACTION:
  case MUNITION_SPAWN_NEW_CLONE:
  case MUNITION_DELETE:
  case MUNITION_CLIP_TO_COLLISION:
  case MUNITION_SET_STATE_RISING:
  case MUNITION_SET_STATE_FALLING:
  case MUNITION_DROP_PARACHUTE:
  case MUNITION_RESET_TIMER:
    return 0;
    break;
  case MUNITION_CHANGE_TYPE:
  case MUNITION_SPAWN_NEW:
  case MUNITION_SPAWN_NEW_CLONE_RANDOM:
  case MUNITION_BOUNCE:
  case MUNITION_SPAWN_NEW_CLONE_MULTIPLE:
  case MUNITION_SPAWN_NEW_CLONE_SPREAD:
  case MUNITION_SET_THRUST:
  case MUNITION_SET_DRAG:
  case MUNITION_DEPLOY_PARACHUTE:
  case MUNITION_SET_TIMER:
    return 1;
    break;
  case MUNITION_SPAWN_NEW_MULTIPLE:
  case MUNITION_SPAWN_NEW_RANDOM:
  case MUNITION_SPAWN_NEW_CLONE_RANDOM_MULTIPLE:
  case MUNITION_SPAWN_NEW_SPREAD:
  case MUNITION_EXPLODE:
    return 2;
    break;
  case MUNITION_SPAWN_NEW_RANDOM_MULTIPLE:
    return 3;
    break;
  }

  /* Shouldn't get here */
  return -1;
}

/* Returns the action type described by a string */
static MunitionActionType get_action_type(char *action_name)
{
  char action_names[][50] = {"CHANGE_TYPE", "SPAWN_NEW", "SPAWN_NEW_MULTIPLE",
			     "SPAWN_NEW_RANDOM", "SPAWN_NEW_RANDOM_MULTIPLE",
			     "SPAWN_NEW_SPREAD", "BOUNCE", "EXPLODE", "DELETE",
			     "SPAWN_NEW_CLONE", "SPAWN_NEW_CLONE_MULTIPLE",
			     "SPAWN_NEW_CLONE_RANDOM", "SPAWN_NEW_CLONE_SPREAD",
			     "SPAWN_NEW_CLONE_RANDOM_MULTIPLE", "CLIP_TO_COLLISION",
			     "SET_THRUST", "SET_DRAG", "SET_STATE_RISING",
			     "SET_STATE_FALLING", "DEPLOY_PARACHUTE", "DROP_PARACHUTE",
			     "SET_TIMER", "RESET_TIMER"};
  char *name;
  int i, n_names = 23;

  /* Make an upper-case version of the name */
  name = strdup(action_name);
  for(i = 0; i < strlen(name); i++)
    name[i] = toupper(name[i]);

  for(i = 0; i < n_names; i++)
    if(!strcmp(name, action_names[i]))
      switch(i)
      {
      case 0: /* CHANGE_TYPE */
	Nfree(name);
	return MUNITION_CHANGE_TYPE;
	break;
      case 1: /* SPAWN_NEW */
	Nfree(name);
	return MUNITION_SPAWN_NEW;
	break;
      case 2: /* SPAWN_NEW_MULTIPLE */
	Nfree(name);
	return MUNITION_SPAWN_NEW_MULTIPLE;
	break;
      case 3: /* SPAWN_NEW_RANDOM */
	Nfree(name);
	return MUNITION_SPAWN_NEW_RANDOM;
	break;
      case 4: /* SPAWN_NEW_RANDOM_MULTIPLE */
	Nfree(name);
	return MUNITION_SPAWN_NEW_RANDOM_MULTIPLE;
	break;
      case 5: /* SPAWN_NEW_SPREAD */
	Nfree(name);
	return MUNITION_SPAWN_NEW_SPREAD;
	break;
      case 6: /* BOUNCE */
	Nfree(name);
	return MUNITION_BOUNCE;
	break;
      case 7: /* EXPLODE */
	Nfree(name);
	return MUNITION_EXPLODE;
	break;
      case 8: /* DELETE */
	Nfree(name);
	return MUNITION_DELETE;
	break;
      case 9: /* SPAWN_NEW_CLONE */
	Nfree(name);
	return MUNITION_SPAWN_NEW_CLONE;
	break;
      case 10: /* SPAWN_NEW_CLONE_MULTIPLE */
	Nfree(name);
	return MUNITION_SPAWN_NEW_CLONE_MULTIPLE;
	break;
      case 11: /* SPAWN_NEW_CLONE_RANDOM */
	Nfree(name);
	return MUNITION_SPAWN_NEW_CLONE_RANDOM;
	break;
      case 12: /* SPAWN_NEW_CLONE_SPREAD */
	Nfree(name);
	return MUNITION_SPAWN_NEW_CLONE_SPREAD;
	break;
      case 13: /* SPAWN_NEW_CLONE_RANDOM_MULTIPLE */
	Nfree(name);
	return MUNITION_SPAWN_NEW_CLONE_RANDOM_MULTIPLE;
	break;
      case 14: /* CLIP_TO_COLLISION */
	Nfree(name);
	return MUNITION_CLIP_TO_COLLISION;
	break;
      case 15: /* SET_THRUST */
	Nfree(name);
	return MUNITION_SET_THRUST;
	break;
      case 16: /* SET_DRAG */
	Nfree(name);
	return MUNITION_SET_DRAG;
	break;
      case 17: /* SET_STATE_RISING */
	Nfree(name);
	return MUNITION_SET_STATE_RISING;
	break;
      case 18: /* SET_STATE_FALLING */
	Nfree(name);
	return MUNITION_SET_STATE_FALLING;
	break;
      case 19: /* DEPLOY_PARACHUTE */
	Nfree(name);
	return MUNITION_DEPLOY_PARACHUTE;
	break;
      case 20: /* DROP_PARACHUTE */
	Nfree(name);
	return MUNITION_DROP_PARACHUTE;
	break;
      case 21: /* SET_TIMER */
	Nfree(name);
	return MUNITION_SET_TIMER;
	break;
      case 22: /* RESET_TIMER */
	Nfree(name);
	return MUNITION_RESET_TIMER;
	break;
      }

  return MUNITION_UNKNOWN_ACTION;
}

/* Work out what type of param it is, int or float and set it appropriately */
static void set_action_param(MunitionActionType type, ActionParameter *params,
			     int index, char *word)
{
  ActionParameterType param_type = 0;

  switch(type)
  {
  case MUNITION_UNKNOWN_ACTION:
  case MUNITION_SPAWN_NEW_CLONE:
  case MUNITION_DELETE:
  case MUNITION_CLIP_TO_COLLISION:
  case MUNITION_SET_STATE_RISING:
  case MUNITION_SET_STATE_FALLING:
  case MUNITION_DROP_PARACHUTE:
  case MUNITION_RESET_TIMER:
    return;
    break;
  case MUNITION_CHANGE_TYPE: /* (char *name) */
  case MUNITION_SPAWN_NEW: /* (char *name) */
    param_type = PARAM_STRING;
    break;
  case MUNITION_SPAWN_NEW_MULTIPLE: /* (char *name, int num) */
  case MUNITION_SPAWN_NEW_CLONE_MULTIPLE: /* (char *name, int num) */
    if(index == 0)
      param_type = PARAM_STRING;
    else
      param_type = PARAM_INT;
    break;
  case MUNITION_SPAWN_NEW_RANDOM: /* (char *name, float rand_dist) */
    if(index == 0)
      param_type = PARAM_STRING;
    else
      param_type = PARAM_FLOAT;
    break;
    break;
  case MUNITION_SPAWN_NEW_CLONE_RANDOM_MULTIPLE: /* (int num, float rand_dist) */
    if(index == 0)
      param_type = PARAM_INT;
    else
      param_type = PARAM_FLOAT;
    break;
  case MUNITION_SPAWN_NEW_RANDOM_MULTIPLE: /* (char *name, int num, float rand_dist) */
  case MUNITION_SPAWN_NEW_SPREAD: /* (char *name, int num, float spread_angle) */
    switch(index)
    {
    case 0:
      param_type = PARAM_STRING;
      break;
    case 1:
      param_type = PARAM_INT;
      break;
    case 2:
      param_type = PARAM_FLOAT;
      break;
    }
    break;
  case MUNITION_BOUNCE: /* (float bounce_efficiency) */
  case MUNITION_EXPLODE: /* (float radius, float force) */
  case MUNITION_SPAWN_NEW_CLONE_RANDOM: /* (float rand_dist) */
  case MUNITION_SPAWN_NEW_CLONE_SPREAD: /* (float spread_angle) */
  case MUNITION_SET_THRUST: /* (float thrust) */
  case MUNITION_SET_DRAG: /* (float drag) */
  case MUNITION_DEPLOY_PARACHUTE: /* (float drag) */
  case MUNITION_SET_TIMER: /* (float time) */
    param_type = PARAM_FLOAT;
    break;
  }

  switch(param_type)
  {
  case PARAM_INT:
    params[index].int_param = atoi(word);
    break;
  case PARAM_FLOAT:
    params[index].float_param = atof(word);
    break;
  case PARAM_STRING:
    params[index].string_param = strdup(word);
    break;
  }
}

/* Load a list of munition actions from a file */
static MunitionAction *load_munition_actions(FILE *mun_file)
{
  MunitionAction *action_list = NULL, **action_list_ptr = &action_list;
  MunitionActionType type;
  char *word;
  ActionParameter *params = NULL;
  int i, nparams;

  /* Get the type of the first action */
  word = get_file_word(mun_file);

  do
  {
    /* Get the type of the action */
    type = get_action_type(word);
    if(type == MUNITION_UNKNOWN_ACTION)
    {
      Neprintf("  Warning: Unknown action %s\n", word);
      do
      {
	Nfree(word);
	word = get_file_word(mun_file);
      } while(word && strcmp(word, "end")
	      && (get_action_type(word) == MUNITION_UNKNOWN_ACTION));
      /* If run out of file, or end of list */
      if(!(word && strcmp(word, "end")))
      {
	Nfree(word);
	return action_list;
      }
    }
    Nfree(word);

    /* Copy the required number of params */
    nparams = get_action_nparams(type);
    /* If has params, copy them */
    if(nparams)
    {
      params = malloc(sizeof(ActionParameter) * nparams);
    
      for(i = 0; i < nparams; i++)
      {
	word = get_file_word(mun_file);
	set_action_param(type, params, i, word);
	Nfree(word);
      }
    }

    /* Make the action */
    *action_list_ptr = allocate_munition_action();
    (*action_list_ptr)->type = type;
    if(nparams)
      (*action_list_ptr)->parameters = params;

    /* Get the type of the next action */
    word = get_file_word(mun_file);

    /* Move on to the next action */
    action_list_ptr = &(*action_list_ptr)->next;
  } while(strcmp(word, "end"));

  return action_list;
}

/* Loads a munition entry from a file */
static Munition *load_munition(FILE *mun_file)
{
  Munition *munition;
  char *option = NULL, *value = NULL;
  char options[][50] = {"name", "type", "cost", "critical_time", "begin", "end"};
  char begin_options[][50] = {"spawn_actions", "timer_actions", "apex_actions",
			      "collision_actions"};
  int i, j, n_options = 6, n_begin_options = 4;

  /* Make an empty munition */
  munition = allocate_munition();

  while(1)
  {
    Nfree(option);
    Nfree(value);
    option = get_file_word(mun_file);
    /* Reached the end of the munition */
    if(!strcmp(option, "end"))
       return munition;
    value = get_file_word(mun_file);

    if(!option || !value)
    {
      Nfree(option);
      Nfree(value);
      Neprintf("  Warning: Unexpected EOF in munition\n");
      return NULL;
    }

    /* Find the index that matches the words */
    for(i = 0; (i < n_options) && strcmp(option, options[i]); i++);

    switch(i)
    {
    case 0: /* name */
      Nfree(munition->name);
      munition->name = strdup(value);
      break;
    case 1: /* type */
      if(!strcmp(value, "SHELL"))
	munition->type = MUNITION_SHELL;
      else if(!strcmp(value, "MISSILE"))
	munition->type = MUNITION_MISSILE;
      else if(!strcmp(value, "BOMBLET"))
	munition->type = MUNITION_BOMBLET;
      else
      {
	fprintf(stderr, "Warning: Unknown munition type \"%s\"\n", value);
	munition->type = MUNITION_NONE;
      }
      break;
    case 2: /* cost */
      munition->cost = atoi(value);
      break;
    case 3: /* critical_time */
      munition->critical_time = atof(value);
      break;
    case 4: /* begin */
      for(j = 0; (j < n_begin_options) && strcmp(value, begin_options[j]); j++);

      if(j == n_begin_options)
      {
	Neprintf("  Warning: Unknown `begin' option %s\n", value);
	/* Find the end that went with this begin */
	do
	{
	  Nfree(value);
	  value = get_file_word(mun_file);
	} while(value && strcmp(value, "end"));
      }
      else
      {
	switch(j)
	{
	case 0: /* spawn_actions */
	  munition->spawn_actions = load_munition_actions(mun_file);
	  break;
	case 1: /* timer_actions */
	  munition->timer_actions = load_munition_actions(mun_file);
	  break;
	case 2: /* apex_actions */
	  munition->apex_actions = load_munition_actions(mun_file);
	  break;
	case 3: /* collision_actions */
	  munition->collision_actions = load_munition_actions(mun_file);
	  break;
	}
      }
      break;
    default: /* Didn't find a match */
      Neprintf("  Warning: Unknown option \"%s\" in munitions file\n", option);
      break;
    }
  }

  return NULL;
}

/* Loads the munitions from a file */
Munition *load_munition_list(char *file_name)
{
  Munition *munitions = NULL, **mun_ptr = &munitions, *munition;
  FILE *munition_file;
  char *option;
  int nmunitions = 0;

  Nprintf("Loading munition list...\n");

  munition_file = fopen(file_name, "r");
  if(munition_file == NULL)
  {
    Neprintf("  %s not found trying /usr/local/share/natow/munitions\n", file_name);
    munition_file = fopen("/usr/local/share/natow/munitions", "r");
    if(munition_file == NULL)
    {
      Neprintf("  /usr/local/share/natow/munitions not found, trying ./munitions\n");
      munition_file = fopen("./munitions", "r");
    }
  }

  if(munition_file == NULL)
  {
    Neprintf("Error opening munitions file...\n");
    return NULL;
  }

  /* Make the NULL weapon */
  munitions = allocate_munition();
  munitions->name = strdup("No weapon");
  munitions->usable = TRUE;

  while(1)
  {
    /* Get the next available option */
    option = get_file_word(munition_file);

    /* If no more options */
    if(option == NULL)
    {
      fclose(munition_file);
      Nprintf("  %d munition(s) found\n", nmunitions);
      Nprintf("Munition list loaded\n");

      /* Make the list link back on itself */
      (*mun_ptr)->next = munitions;
      munitions->last = *mun_ptr;

      return munitions;
    }

    if(!strcmp(option, "munition"))
    {
      munition = load_munition(munition_file);
      (*mun_ptr)->next = munition;
      munition->last = *mun_ptr;

      mun_ptr = &(*mun_ptr)->next;
      nmunitions++;
    }
    else
      Neprintf("  Warning: Unknown option \"%s\" in munition file \"%s\".\n", option, file_name);
  }
}

/* Loads a list of usable munitions, and enables those munitions */
void load_usable_munitions(char *file_name, World *world)
{
  FILE *usable_file;
  Munition *munition, *munition_list = world->munition_list;
  char *name;
  int nusable = 0;

  Nprintf("Selecting usable munitions...\n");

  usable_file = fopen(file_name, "r");
  if(usable_file == NULL)
  {
    Neprintf("  %s not found trying /usr/local/share/natow/usable\n", file_name);
    usable_file = fopen("/usr/local/share/natow/usable", "r");
    if(usable_file == NULL)
    {
      Neprintf("  /usr/local/share/natow/usable not found, trying ./usable\n");
      usable_file = fopen("./usable", "r");
    }
  }

  if(usable_file == NULL)
  {
    Neprintf("Error opening `usable munitions' file...\n");
    return;
  }

  /* Check off the usable ones as they're loaded */
  while((name = get_file_word(usable_file)) != NULL)
  {
    munition = find_munition(munition_list, name);

    if(munition == NULL)
      Neprintf("  Warning: munition %s doesn't exist", name);
    else
    {
      munition->usable = TRUE;
      nusable++;
    }

    Nfree(name);
  }

  Nprintf("  %d usable munition(s) activated\n", nusable);

  Nprintf("Usable munitions selected\n");
}

/* Free the munition list */
int clear_munition_list(World *world)
{
  Munition *munition, *tmp;

  if(!world)
    return FALSE;

  munition = world->munition_list;
  while(munition)
  {
    /* List loops back on itself */
    if(munition == world->munition_list)
      break;

    tmp = munition->next;
    free_munition(munition);
    munition = tmp;
  }

  return TRUE;
}

/* Free a munition */
static int free_munition(Munition *munition)
{
  if(!munition)
    return FALSE;

  Nfree(munition->name);
  free_munition_actions(munition->spawn_actions);
  free_munition_actions(munition->timer_actions);
  free_munition_actions(munition->apex_actions);
  free_munition_actions(munition->collision_actions);
  Nfree(munition);

  return TRUE;
}

/* Free a list of munition actions */
static int free_munition_actions(MunitionAction *action_list)
{
  MunitionAction *action, *tmp;

  action = action_list;
  while(action)
  {
    tmp = action->next;
    free_munition_action(action);
    action = tmp;
  }

  return TRUE;
}

/* Free a munition action */
static int free_munition_action(MunitionAction *action)
{
  if(!action)
    return FALSE;

  Nfree(action->parameters);
  Nfree(action);

  return TRUE;
}
