/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* objects.c
   Manage generic objects in the world */

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>

#include "objects.h"
#include "world.h"
#include "game.h"
#include "tools.h"
#include "map.h"
#include "text.h"
#include "file.h"

/* load an object file */
Object *load_object(Game *game, char *fname)
{
  FILE *f = fopen(fname, "r");
  char *option = NULL, *value = NULL;
  Object *object = allocate_object(game, OBJECT_UNKNOWN);
  char *model = NULL, *model_dead = NULL, *name = NULL;
  float mass = -1., health = -1.;
  int floatable = -1, movable = -1;
  int component = -1, component_dead = -1;
  float time_dead = 0.;
  int type = 0;

  if (object == NULL)
    {
    Neprintf("Cannot allocate object.\n");
    return(NULL);
    }

  if (f == NULL)
    {
    Neprintf("Cannot open object file '%s'\n", fname);
    return(NULL);
    }

  do
  {
    Nfree(option);
    option = get_file_word(f);
    Nfree(value);
    if(option != NULL)
      value = get_file_word(f);

    if(option == NULL);
    else if (strcmp("model", option) == 0)
      {
      Nfree(model);
      model = strdup(value);
      }

    else if (strcmp("model_dead", option) == 0)
      {
      Nfree(model_dead);
      model_dead = strdup(value);
      }

    else if (strcmp("time_dead", option) == 0)
      {
      time_dead = atof(value);
      }

    else if (strcmp("comp", option) == 0)
      {
      component = atoi(value);
      }

    else if (strcmp("comp_dead", option) == 0)
      {
      component_dead = atoi(value);
      }

    else if (strcmp("name", option) == 0)
      {
      Nfree(name);
      name = strdup(value);
      }
      
    else if (strcmp("type", option) == 0)
      {
      type = atoi(value);
      }
      
    else if (strcmp("health", option) == 0)
      {
      health = atof(value);
      }
      
    else if (strcmp("mass", option) == 0)
      {
      mass = atof(value);
      }
      
    else if (strcmp("floatable", option) == 0)
      {
      floatable = atoi(value);
      }

    else if (strcmp("movable", option) == 0)
      {
      movable = atoi(value);
      }

    else
      {
      Neprintf("  Unknown option '%s'. Ignored.\n", option);
      }
    }
  while(option != NULL);
  fclose(f);

  /* check options */
  if (model == NULL)
    {
    Neprintf("  No model given in object '%s'.\n", fname);
    delete_object(object);
    return(NULL);
    }

  /* get values */
  object->model = glassLoadObject(model);
  if (object->model == NULL)
    {
    Neprintf("  Cannot read model '%s' in object '%s'.\n", model, fname);
    delete_object(object);
    return(NULL);
    }
  if (model_dead != NULL)
    object->model_dead = glassLoadObject(model_dead);
  if (name != NULL)
    object->name = strdup(name);
  if (mass > 0.)
    object->mass = mass;
  if (health > 0.)
    object->health = health;
  if (movable >= 0)
    object->movable = movable;
  if (floatable >= 0)
    object->floatable = floatable;
  object->type = type;
  object->component = component;
  object->component_dead = component_dead;
  object->time_dead = time_dead;

  return(object);
}

/* allocate a new object */
Object *allocate_object(Game *game, ObjectType type)
{
  Object *obj = malloc(sizeof(Object));

  if (obj == NULL)
    {
    Neprintf("Cannot allocate new object.\n");
    return(NULL);
    }

  /* init the object */
  obj->type = type;
  obj->state = OBJECT_ALIVE;
  obj->name = NULL;
  obj->model = NULL;
  obj->model_dead = NULL;
  obj->health = 0.;
  obj->time_dead = 1.;
  obj->mass = 0.;
  obj->movable = FALSE;
  obj->floatable = FALSE;
  obj->pos[0] = 0.;
  obj->pos[1] = 0.;
  obj->pos[2] = 0.;
  obj->vel[0] = 0.;
  obj->vel[1] = 0.;
  obj->vel[2] = 0.;
  obj->acc[0] = 0.;
  obj->acc[1] = game->gravity;
  obj->acc[2] = 0.;

  obj->next  = NULL;
  obj->next2 = NULL;
  obj->last  = NULL;

  return(obj);
}

/* delete an object */
void delete_object(Object *obj)
{
  if (obj == NULL)
    return;
  Nfree(obj->name);
  if (obj->model != NULL)
    glassFreeObject(obj->model);
  if (obj->model_dead != NULL)
    glassFreeObject(obj->model);
  Nfree(obj);
}

/* add an object in the list */
void add_object(Game *game, Object *obj)
{
  if (obj == NULL)
    return;

  obj->next = game->object_list;
  if (game->object_list != NULL)
    game->object_list->last = obj;
  obj->last = NULL;
  game->object_list = obj;

  game->nobjects++;
}

/* remove an object from the list (Note: it freed it) */
void remove_object(Game *game, Object *obj)
{
  if (obj->next != NULL)
    obj->next->last = obj->last;
  if (obj->last != NULL)
    obj->last->next = obj->next;
  else /* 1st of the list */
    game->object_list = obj->next;

  delete_object(obj);

  game->nobjects--;
}

/* set object mass */
void set_object_mass(Object *object, float mass)
{
  if (object == NULL)
    return;
  object->mass = mass;
}

/* get object mass */
float get_object_mass(Object *object)
{
  if (object == NULL)
    return(0.);
  return(object->mass);
}

/* set object health */
void set_object_health(Object *object, float health)
{
  if (object == NULL)
    return;
  object->health = health;
}

/* get object health */
float get_object_health(Object *object)
{
  if (object == NULL)
    return(0.);
  return(object->health);
}

/* set object 'movable' */
void set_object_movable(Object *object, int movable)
{
  if (object == NULL)
    return;
  object->movable = movable;
}

/* get object 'movable' */
int get_object_movable(Object *object)
{
  if (object == NULL)
    return(0.);
  return(object->movable);
}

/* set object name */
void set_object_name(Object *object, char *name)
{
  if (object == NULL)
    return;
  Nfree(object->name);
  object->name = strdup(name);
}

/* get object name */
char *get_object_name(Object *object)
{
  if (object == NULL)
    return(NULL);
  return(object->name);
}

/* set object position (x,y; height is deduced from map) */
void set_object_position(Game *game, Object *object, float x, float y)
{
  if (object == NULL)
    return;
  
  object->pos[0] = x;
  object->pos[2] = y;
  object->pos[1] = get_height(game->map, x, y);
}

/* get object position */
void get_object_position(Object *object, float *pos)
{
  if (object == NULL)
    return;

  pos[0] = object->pos[0];
  pos[1] = object->pos[1];
  pos[2] = object->pos[2];
}


/* update all objects */
void update_objects(Game *game, float dt)
{
  Object *obj = game->object_list;

  while(obj != NULL)
    {
    update_object(game, obj, dt);
    obj = obj->next;
    }
}

/* update an object */
void update_object(Game *game, Object *object, float dt)
{
  float tmp;

  /* if dead, update its state */
  if (object->state == OBJECT_DEAD)
    {
    object->time_dead -= dt;
    if (object->time_dead <= 0.)
      object->state = OBJECT_REMOVED;
    /* need to remove the object from the list */
    return;
    }
  /* if object dont touch the ground, update */
  tmp = get_height(game->map, object->pos[0], object->pos[2]);
  if (object->pos[1] > tmp)
    {
    object->vel[0] += dt * object->acc[0];
    object->vel[1] += dt * object->acc[1];
    object->vel[2] += dt * object->acc[2];
    object->pos[0] += dt * (object->vel[0] + 0.5f);
    object->pos[1] += dt * (object->vel[1] + 0.5f);
    object->pos[2] += dt * (object->vel[2] + 0.5f);
    /* touch the ground: stop it */
    if (object->pos[1] <= tmp)
      {
      object->vel[0] = 0.;
      object->vel[1] = 0.;
      object->vel[2] = 0.;
      object->pos[1] = tmp;
      }
    }
}

/* give damages to objects in the given list */
void damage_objects(Game *game, Object *objects, float *pos, float radius, float force)
{
  Object *obj = objects;
  float damage, dx, dy, dz, d2;

  if (objects == NULL)
    return;

  do
    {
    if ((obj->state == OBJECT_ALIVE)||(obj->state == OBJECT_DAMAGED))
      {
      dx = pos[0] - obj->pos[0];
      dy = pos[1] - obj->pos[1];
      dz = pos[2] - obj->pos[2];
      d2 = dx*dx + dy*dy + dz*dz;
      /* Amount of damage done by the shot */
      damage = 1000.0 * force / (d2*sqrt(d2));
      if(damage < 1.0)
	damage = 0.0;
      obj->health -= damage;
      if (obj->health <= 0.)
        obj->state = OBJECT_DEAD;
      }
    obj = obj->next2;
    }
  while(obj != NULL);
}

/* add object at given coords */
void add_object_in_map(Game *game, Object *object, int x, int y)
{
  object->next2 = game->map->objects[x][y];
  game->map->objects[x][y] = object;
}


/* draw all objects */
void draw_objects(Game *game)
{
  Object *obj = game->object_list;

  while(obj)
  {
    draw_object(game, obj);
    obj = obj->next;
  }
}

/* draw an object */
void draw_object(Game *game, Object *object)
{
  if ((object == NULL)||(object->model == NULL))
    return;

  glPushMatrix();
  /* Translate to where the object is on the map */
  glTranslatef(object->pos[0], object->pos[1], object->pos[2]);

  /* alive or dead */
  if (object->state == OBJECT_DEAD)
    {
    if (object->model_dead != NULL)
      {
      if (object->component_dead >= 0)
	glassDrawComponent(object->model_dead, object->component_dead);
      else
	glassDrawObject(object->model_dead);
      }
    }
  else
  if (object->state == OBJECT_ALIVE)
    {
    if (object->component >= 0)
      glassDrawComponent(object->model, object->component);
    else
      glassDrawObject(object->model);
    }

  glPopMatrix();
}
