/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* player_config.c
   This file contains all the player config related code */

#include "stdlib.h"
#include "string.h"

#include "player_config.h"
#include "tools.h"
#include "world.h"
#include "game.h"
#include "player.h"
#include "player_list.h"
#include "2d.h"
#include "glut_interface.h"

static int selected = 0;
static float wheel = 0.0, target_wheel = 0.0, spin = 0.0, last_time = 0.0;
static int name_selection = FALSE;
static float theta = 0.0;

static int calc_target_wheel(Player *player);

/* Reset back to defaults */
int reset_player_config(void)
{
  selected = 0;
  wheel = 0.0;
  target_wheel = 0.0;
  spin = 0.0;
  last_time = 0.0;
  name_selection = FALSE;
  theta = 0.0;

  return TRUE;
}

static PlayerEntry *config_player = NULL;

/* Draw a screen to configure a player */
int draw_player_config(Game *game)
#define MAX_LINE_LENGTH 100
{
  World *world;
  Player *player;
  Tank *tank;
  char text[MAX_LINE_LENGTH], *type_text;
  int i = 0, tank_no = 0;
  float dt, dwheel, radius;

  if(!game || !game->world)
    return FALSE;

  world = game->world;
  theta = 360.0 / world->ntanks;

  /* Make radius so there's always 10m between tanks */
  radius = 10.0 / ((M_PI / 180.0) * theta);

  /* If haven't updated for ages, dt will be huge, so estimate */
  if(world->total_elapsed_time - last_time > 1.0)
    last_time = world->total_elapsed_time - 0.001;

  dt = world->total_elapsed_time - last_time;

  /* If no edited player, edit the first player in the game */
  if(!config_player)
  {
    config_player = game->players->first;
    return FALSE;
  }
  player = config_player->player;

  /* Find out which tank they're using */
  for(tank = world->tank_list; tank != player->tank; tank_no++)
    if((tank = tank->next) == NULL)
      tank = player->tank;

  glPushMatrix();

  dwheel = target_wheel - wheel;
  wheel += dwheel * (1.0 - exp(-5.0 * dt));
  spin  = world->total_elapsed_time * 20.0;

  glTranslatef(0.0, -3.0, -radius - 10.0);

  tank = world->tank_list;
  for(i = 0; i < world->ntanks; i++)
  {
    glPushMatrix();

    glRotatef(wheel + i*theta, 0.0, 1.0, 0.0);
    glTranslatef(0.0, 0.0, radius);
    glRotatef(spin, 0.0, 1.0, 0.0);
    draw_tank(tank);

    glPopMatrix();

    tank = tank->next;
    if(!tank)
      tank = world->tank_list;
  }

  glPopMatrix();

  last_time = world->total_elapsed_time;

  activate_2d();

  snprintf(text, MAX_LINE_LENGTH, "Name: \"%s\"", player->name);
  draw_2d_text(text, 1, 95, 0, selected == 0);
  if(cursor_inside_text())
    selected = 0;

  /*snprintf(text, MAX_LINE_LENGTH, "Cash: $%d", player->cash);
  draw_2d_text(text, 1, 90, 0, 0);*/

  switch(player->player_type)
  {
  case PLAYER_HUMAN:
    type_text = strdup("Human");
    break;
  case PLAYER_AI_STUPID:
    type_text = strdup("Stupid AI");
    break;
  case PLAYER_AI_MEDIUM:
    type_text = strdup("OK AI");
    break;
  case PLAYER_AI_GOOD:
    type_text = strdup("Good AI");
    break;
  case PLAYER_AI_GOD:
    type_text = strdup("Super AI");
    break;
  default:
    type_text = strdup("Unknown");
    break;
  }
  snprintf(text, MAX_LINE_LENGTH, "Player type: %s", type_text);
  Nfree(type_text);
  draw_2d_text(text, 1, 90, 0, selected == 1);
  if(cursor_inside_text())
    selected = 1;

  draw_2d_text_center("< Tank >", 1, 98, 75, 0, selected == 2);
  if(cursor_inside_text())
    selected = 2;

  draw_2d_text("Last", 1, 1, 3, selected == 4);
  if(cursor_inside_text())
    selected = 4;

  draw_2d_text_right("Next", 1, 98, 1, 3, selected == 3);
  if(cursor_inside_text())
    selected = 3;

  draw_2d_text_center("Up/Down selects", 1, 98, 4, -3, FALSE);
  draw_2d_text_center("Press enter when finished", 1, 98, 1, -3, TRUE);

  /* The players tank */
  tank = player->tank;
  if(!tank)
  {
    deactivate_2d();
    return FALSE;
  }

  draw_2d_text(tank->name, 1, 70, -5, FALSE);
  snprintf(text, MAX_LINE_LENGTH, "Mass:\t\t\t%.2fkg", tank->mass);
  draw_2d_text(text, 1, 67.5, -5, FALSE);
  snprintf(text, MAX_LINE_LENGTH, "Armour:\t\t%.2f", tank->health);
  draw_2d_text(text, 1, 65, -5, FALSE);
  snprintf(text, MAX_LINE_LENGTH, "Top acceleration:\t%.2fms^-2", tank->acc);
  draw_2d_text(text, 1, 62.5, -5, FALSE);
  snprintf(text, MAX_LINE_LENGTH, "Top speed:\t\t%.2fms^-1", tank->speed);
  draw_2d_text(text, 1, 60, -5, FALSE);
  draw_2d_text(tank->comment, 1, 57.5, -5, FALSE);

  deactivate_2d();

  return TRUE;
}

/* Calculate target_wheel based on what tank is selected */
static int calc_target_wheel(Player *player)
{
  World *world;
  Tank *tank;
  int i;

  if(!player || !player->world)
    return FALSE;

  world = player->world;

  /* Find what tank number they have */
  tank = world->tank_list;
  for(i = 0; tank && tank != player->tank; i++)
    tank = tank->next;

  /* Head in the nearest direction */
  if(i > world->ntanks / 2)
    i -= world->ntanks;

  /* Caclulate the angle to point at this tank */
  target_wheel = -i * theta;

  return TRUE;
}

/* Manage the input for the player config view */
void manage_player_config_input(Game *game, int key)
{
  World *world;
  Player *player;
  PlayerType new_type, old_type;
  GameMode mode;
  char *temp, *temp2;

  if(!config_player)
    return;
  player = config_player->player;

  world = player->world;
  mode = game->mode;

  /* Select the previous option */
  if(key == KEY_UP)
  {
    selected--;
    if(selected < 0)
      selected = 0;
  }

  /* Select the next option */
  else if(key == KEY_DOWN)
  {
    selected++;
    if(selected > 4)
      selected = 4;
  }

  switch(selected)
  {
  case 0: /* name */
    if(key == KEY_RETURN)
      break;
    else if(key == KEY_BACKSPACE)
    {
      if(strlen(player->name) > 0)
      {
        temp = player->name;
        player->name[strlen(temp)-1] = '\0';
        temp2 = strdup(player->name);
        Nfree(temp);
        player->name = temp2;
      }
    }
    else if((key >= KEY_SPACE) && (key <= KEY_LAST)) /* valid keys */
    {
      temp = player->name;
      temp2 = malloc(sizeof(char)*(strlen(temp)+2));
      sprintf(temp2, "%s%c", temp, (char)key);
      player->name = temp2;
      Nfree(temp);
    }
    return;
    break;
  case 1: /* player type */
    new_type = player->player_type;
    old_type = new_type;

    if(key == KEY_LEFT || key == KEY_MOUSE3)
    {
      /* Select the previous type */
      if(player->player_type != PLAYER_HUMAN)
	new_type = player->player_type - 1;

      /* In single player mode only one human player */
      if(mode == SINGLE_PLAYER)
      {
	/* All other players apart from first must be AI's */
        if(config_player != game->players->first &&
	   new_type == PLAYER_HUMAN)
	  new_type = PLAYER_AI_STUPID;
      }

      set_player_type(player, new_type);
    }
    else if(key == KEY_RIGHT || key == KEY_MOUSE1)
    {
      /* Select the next type */
      if(player->player_type != PLAYER_AI_GOD)
	new_type = player->player_type + 1;

      /* In single player mode only one human player */
      if(mode == SINGLE_PLAYER && config_player == game->players->first)
	new_type = PLAYER_HUMAN;

      set_player_type(player, new_type);
    }

    /* AI controled: choose name and tank */
    if(new_type != old_type &&
       player->player_type != PLAYER_HUMAN &&
       player->player_type != PLAYER_HUMAN_NET)
    {
      config_ai_player(player);
      calc_target_wheel(player);
    }
    break;
  case 2: /* tank select */
    if(key == KEY_LEFT || key == KEY_MOUSE3)
    {
      /* Select the next tank */
      player->tank = player->tank->next;

      if(!player->tank)
	player->tank = world->tank_list;

      target_wheel -= theta;
    }
    else if(key == KEY_RIGHT || key == KEY_MOUSE1)
    {
      /* Select the previous tank */
      player->tank = player->tank->last;

      if(player->tank == NULL)
      {
	player->tank = world->tank_list;
	while(player->tank->next != NULL)
	  player->tank = player->tank->next;
      }

      target_wheel += theta;
    }
    break;
  case 4: /* last */
    reset_player_config();
    config_player = config_player->last;
    if(config_player)
      calc_target_wheel(config_player->player);
    else
    {
      /* Remove these players from the game */
      clear_players(game->players);
      /* Remove these players */
      remove_players(world);
      set_display_mode(world, NATOW_DISPLAY_GAME_CONFIG);
      return;
    }
    break;
  default:
    break;
  }

  /* Finished editing this player */
  if(key == KEY_SPACE || key == KEY_RETURN || (key == KEY_MOUSE1 && selected == 3))
  {
    /* Reset for next player */
    reset_player_config();

    /* Go to next player */
    config_player = config_player->next;

    if(config_player)
    {
      /* If single player mode, set name and tank */
      if(mode == SINGLE_PLAYER)
      {
	config_ai_player(config_player->player);
	calc_target_wheel(config_player->player);
      }
    }

    /* If run out of players, start buying for the first one */
    if(!config_player)
      set_display_mode(world, NATOW_DISPLAY_PLAYER_SHOP);
  }
}

static PlayerEntry *shop_player = NULL;
static int selected_button;

/* Draw a menu to buy weapons etc from */
int draw_player_shop(Game *game)
#define MAX_LINE_LENGTH 100
#define MAX_VIS 6
{
  Player *player;
  Weapon *weapon = NULL;
  char text[MAX_LINE_LENGTH];
  int i, selected, index, offset, position;
  int old_button;

  if(!game)
    return FALSE;

  /* If no edited player, edit the first non-AI player in the game */
  if(!shop_player)
  {
    shop_player = game->players->first;

    /* Skip all the AI players (treated by ai_buy_weapons()) */
    while(shop_player && shop_player->player->player_type != PLAYER_HUMAN)
    {
      ai_buy_weapons(shop_player->player);
      shop_player = shop_player->next;
    }

    /* If run out of players, finished */
    if(!shop_player)
    {
      start_game(game);
      return TRUE;
    }
  }

  player = shop_player->player;

  activate_2d();

  snprintf(text, MAX_LINE_LENGTH, "Name: \"%s\"", player->name);
  draw_2d_text(text, 1, 95, 0, 0);

  snprintf(text, MAX_LINE_LENGTH, "Cash: $%d", player->cash);
  draw_2d_text(text, 1, 90, 0, 0);

  /* Find out which number the selected weapon is */
  weapon = player->weapon_list;
  selected = 0;
  while(weapon != player->selected_weapon)
  {
    weapon = weapon->next;
    selected++;
  }

  if(selected < MAX_VIS)
    offset = MAX_VIS - selected;
  else
    offset = 0;

  draw_2d_text("Weapons", 1, 85, 0, 0);

  /* Go to the first visible weapon */
  weapon = player->weapon_list;
  for(i = 0; i < selected - MAX_VIS; i++)
    weapon = weapon->next;

  /* Display 2*MAX_VIS + 1 weapons */
  for(i = selected - MAX_VIS; i <= selected + MAX_VIS && weapon != NULL; i++)
  {
    index = i + offset;
    position = 80 - (i - selected + MAX_VIS)*5;

    draw_2d_text(weapon->munition->name, 6, position, 0, index == selected);
    /*if(cursor_inside_text())
      selected = index;*/

    snprintf(text, MAX_LINE_LENGTH, "%d", weapon->number);
    draw_2d_text(text, 55, position, 0, index == selected);

    snprintf(text, MAX_LINE_LENGTH, "$%d", weapon->munition->cost);
    draw_2d_text(text, 65, position, 0, index == selected);

    weapon = weapon->next;
  }

  /* If more weapons below, show that */
  if(weapon != NULL)
    draw_2d_text("...", 6, 80 -(2*MAX_VIS + 1)*5, 0, TRUE);

  old_button = selected_button;
  selected_button = 0;
  /*draw_2d_text("Last", 1, 1, 3, old_button == 1);
  if(cursor_inside_text())
    selected_button = 1;*/

  draw_2d_text_right("Next", 1, 98, 1, 3, old_button == 2);
  if(cursor_inside_text())
    selected_button = 2;

  draw_2d_text_center("Up/Down selects", 1, 98, 7, -3, FALSE);
  draw_2d_text_center("+/- Buys/Sells", 1, 98, 4, -3, FALSE);
  draw_2d_text_center("Press enter when finished", 1, 98, 1, -3, TRUE);

  deactivate_2d();

  /* Select the weapon the mouse now hovers over */
  /* TODO */
  /*weapon = player->weapon_list;
  for(i = 0; i < selected; i++)
    weapon = weapon->next;
  player->selected_weapon = weapon;*/

  return TRUE;
}

/* Manage the input when the players buy weapons etc */
int manage_player_shop_input(Game *game, int key)
{
  Player *player;
  char *selected_weapon;

  if(!game || !shop_player)
    return FALSE;

  player = shop_player->player;

  selected_weapon = player->selected_weapon->munition->name;

  switch(key)
  {
  case KEY_UP:
    /* Select the previous weapon */
    if(player->selected_weapon->last)
      player->selected_weapon = player->selected_weapon->last;
    break;
  case KEY_DOWN:
    /* Select the next weapon */
    if(player->selected_weapon->next)
      player->selected_weapon = player->selected_weapon->next;
    break;
  case KEY_PLUS:
  case KEY_RIGHT:
    /* Buy one of the current weapon */
    buy_weapon(player, selected_weapon, 1);
    break;
  case KEY_MINUS:
  case KEY_LEFT:
    /* Sell one of the current weapon */
    sell_weapon(player, selected_weapon, 1);
    break;
  case KEY_MOUSE1:
    if(!selected_button)
      break;
  case KEY_SPACE:
  case KEY_RETURN:
    /* Finished editing this player */
    shop_player = shop_player->next;

    /* If run out of players */
    if(!shop_player)
    {
      start_game(game);
      return TRUE;
    }

    /* Skip over the next few AI players */
    while(shop_player->player->player_type != PLAYER_HUMAN)
    {
      ai_buy_weapons(shop_player->player);
      shop_player = shop_player->next;
      /* If run out of players */
      if(!shop_player)
      {
	start_game(game);
        return TRUE;
      }
    }
    break;
  }

  return TRUE;
}
