/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *                     Joel Pitt      <extropy@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* player_list.c
   Manage lists of players for games and rounds etc */

#include <stdlib.h>

#include "player_list.h"
#include "text.h"
#include "tools.h"

static PlayerList *allocate_plist(void);
static PlayerEntry *allocate_pentry(void);
static PlayerEntry *copy_entry(PlayerEntry *entry);
static int free_pentry(PlayerEntry *pentry);

/* Allocate space for a player list */
static PlayerList *allocate_plist(void)
{
  PlayerList *plist;

  plist = malloc(sizeof(PlayerList));
  if(!plist)
  {
    Neprintf("Error: Unable to allocate space for a player list\n");
    return NULL;
  }

  plist->nplayers = 0;
  plist->first = NULL;
  plist->last = NULL;

  return plist;
}

/* Make a new player list */
PlayerList *new_player_list(void)
{
  PlayerList *plist;

  plist = allocate_plist();
  if(!plist)
    return NULL;

  return plist;
}

/* Allocate space for a player entry */
static PlayerEntry *allocate_pentry(void)
{
  PlayerEntry *pentry;

  pentry = malloc(sizeof(PlayerEntry));
  if(!pentry)
  {
    Neprintf("Error: Unable to allocate space for a player entry\n");
    return NULL;
  }

  pentry->player = NULL;
  pentry->last = NULL;
  pentry->next = NULL;

  return pentry;
}

/* Add a player to the front of the player list */
int prepend_player(PlayerList *list, Player *player)
{
  PlayerEntry *entry;

  if(!list || !player)
    return FALSE;

  entry = allocate_pentry();
  if(!entry)
    return FALSE;

  entry->player = player;
  entry->next = list->first;
  if(entry->next)
    entry->next->last = entry;
  else
    list->last = entry;
  list->first = entry;

  list->nplayers++;

  return TRUE;
}

/* Add a player to the end of the player list */
int append_player(PlayerList *list, Player *player)
{
  PlayerEntry *entry;

  if(!list || !player)
    return FALSE;

  entry = allocate_pentry();
  if(!entry)
    return FALSE;

  entry->player = player;
  entry->last = list->last;
  if(entry->last)
    entry->last->next = entry;
  else
    list->first = entry;
  list->last = entry;

  list->nplayers++;

  return TRUE;
}

/* Sort the player list using built-in qsort */
/* NOTE: If anyone feels good at sorting functions, feel free
   to write a better method */
int sort_player_list(PlayerList *list,
		     int (*cmp)(PlayerEntry **entry1, PlayerEntry **entry2))
#define MAX_PLAYERS 100
{
  PlayerEntry *entry_array[MAX_PLAYERS], *pentry;
  int i;

  if(!list || !cmp)
    return FALSE;

  if(!list->nplayers)
  {
    Neprintf("Error: Unable to sort an empty player list!\n");
    return FALSE;
  }

  if(list->nplayers > 100)
  {
    Neprintf("Error: Unable to sort a list of greater than 100 players "
	     "(list contains %d players)\n", list->nplayers);
    return FALSE;
  }

  /* Make an array of the player entries */
  pentry = list->first;
  for(i = 0; i < list->nplayers; pentry = pentry->next)
    entry_array[i++] = pentry;

  /* Sort this array based on the given compare function */
  qsort(entry_array, list->nplayers, sizeof(entry_array[0]),
	(int (*)(const void *, const void *))cmp);

  /* Reorder the list in this way */
  pentry = list->first = entry_array[0];
  pentry->last = NULL;
  pentry->next = NULL;
  for(i = 1; i < list->nplayers; i++)
  {
    pentry->next = entry_array[i];
    pentry->next->last = pentry;
    pentry->next->next = NULL;

    pentry = pentry->next;
  }

  return TRUE;
}

/* Randomize the order of the players in the list */
int randomize_player_list(PlayerList *list)
{
  PlayerEntry *entry, *entries[100];
  int i, new_order[100], nplayers, number;

  if(!list || !list->nplayers)
    return FALSE;

  Nprintf("Randomizing players...\n");

  if((nplayers = list->nplayers) > 100)
  {
    Neprintf("Cannot randomize a list of players longer than 100\n");
    return FALSE;
  }

  /* Copy the entries into an array */
  entry = list->first;
  for(i = 0; entry && i < nplayers; i++)
  {
    entries[i] = entry;
    entry = entry->next;
  }

  /* An error */
  if(i != nplayers)
  {
    Neprintf("Player list says there should be only %d players, but list"
	     "appears to contain %d players (%s:%d)\n", nplayers, i, __FILE__, __LINE__);
    return FALSE;
  }

  /* Set the entries as unordered */
  for(i = 0; i < nplayers; i++)
    new_order[i] = -1;

  /* Make a random order */
  for(i = 0; i < nplayers; i++)
  {
    do
    {
      number = random_integer(0, nplayers - 1);
    } while(new_order[number] != -1);

    new_order[number] = i;
  }

  /* Reconnect the players in that order */
  entry = list->first = entries[new_order[0]];
  entry->last = NULL;
  i = 1;
  while(entry && i < nplayers)
  {
    entry->next = entries[new_order[i]];
    entry->next->last = entry;
    entry->next->next = NULL;

    entry = entry->next;
    i++;
  }

  return TRUE;
}

/* Copy a player list */
PlayerList *copy_player_list(PlayerList *plist)
{
  PlayerList *list_copy;
  PlayerEntry *entry, *src_entry;

  if(!plist)
    return NULL;

  list_copy = allocate_plist();
  if(!list_copy)
    return NULL;

  /* Copy the first entry */
  list_copy->first = copy_entry(plist->first);
  if(list_copy->first)
    list_copy->nplayers = 1;

  /* Copy the second onwards entries */
  entry = list_copy->first;
  src_entry = plist->first->next;
  while(entry && src_entry)
  {
    entry->next = copy_entry(src_entry);
    entry->next->last = entry;

    list_copy->last = entry->next;

    entry = entry->next;
    src_entry = src_entry->next;
    list_copy->nplayers++;
  }

  return list_copy;
}

/* Copy a player entry */
static PlayerEntry *copy_entry(PlayerEntry *entry)
{
  PlayerEntry *entry_copy;

  if(!entry)
    return NULL;

  entry_copy = allocate_pentry();
  if(!entry_copy)
    return NULL;

  entry_copy->player = entry->player;

  return entry_copy;
}

/* Remove all the players from the player list */
int clear_players(PlayerList *list)
{
  PlayerEntry *entry, *tmp;

  if(!list)
    return FALSE;

  entry = list->first;
  while(entry)
  {
    tmp = entry->next;
    free_pentry(entry);
    entry = tmp;
  }

  list->first = NULL;
  list->last = NULL;
  list->nplayers = 0;

  return TRUE;
}

/* Free a player list */
int free_player_list(PlayerList *plist)
{
  PlayerEntry *entry, *tmp;

  if(!plist)
    return FALSE;

  entry = plist->first;
  while(entry)
  {
    tmp = entry->next;
    free_pentry(entry);
    entry = tmp;
  }

  Nfree(plist);

  return TRUE;
}

/* Free a player entry */
static int free_pentry(PlayerEntry *pentry)
{
  if(!pentry)
    return FALSE;

  Nfree(pentry);

  return TRUE;
}
