/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* round.c
   This file contains the functions to play rounds of combat */

#include <stdio.h>
#include <stdlib.h>

#include "round.h"

#include "world.h"
#include "game.h"
#include "turn.h"

#include "tools.h"
#include "text.h"

#include "camera.h"
#include "player.h"
#include "player_list.h"
#include "objects.h"
#include "projectile.h"
#include "trails.h"
#include "explosion.h"
#include "smoke.h"
#include "draw.h"
#include "2d.h"
#include "ai.h"
#include "map.h"
#include "bird.h"
#include "display_stack.h"
#include "draw_map.h"
#include "light.h"
#include "glut_interface.h"

static Round *allocate_round(void);
static void setup_round_players(Round *round);
static int randomize_player_locations(Round *round);
static int compare_round_winners(PlayerEntry **entry1, PlayerEntry **entry2);

/* Allocate space for a new round */
static Round *allocate_round(void)
{
  Round *round;

  round = malloc(sizeof(Round));
  if(!round)
  {
    Neprintf("Error: Unable to allocate space for round\n");
    return NULL;
  }

  round->world = NULL;
  round->map = NULL;

  round->game = NULL;

  round->turn = 0;
  round->max_turns = 10000;

  round->driving_time = 0.0;

  round->camera = NULL;

  round->players = NULL;
  round->current_player_entry = NULL;
  round->current_player = NULL;

  round->projectile_list = NULL;
  round->trail_list = NULL;
  round->blast_list = NULL;
  round->bird_list = NULL;
  round->smoke_list = NULL;

  return round;
}

/* Start a new round */
int start_round(Game *game)
{
  Round *round;

  if(!game)
    return FALSE;

  if(game->round)
  {
    Neprintf("Error: Trying to start round, but round already started!\n");
    return FALSE;
  }

  /* Check that enough things are present to start a round */
  if(!game->map)
  {
    Neprintf("Unable to start round, no map present\n");
    return FALSE;
  }
  if(!game->players)
  {
    Neprintf("Unable to start round, no players present\n");
    return FALSE;
  }

  /* That was the last round */
  if(game->round_num >= game->max_rounds)
  {
    end_game(game);
    return FALSE;
  }

  game->round_num++;
  Nprintf("Beginning round %d\n", game->round_num);

  activate_2d();
  draw_2d_text_center("Loading...", 1, 98, 50, 8, 2);
  deactivate_2d();
  swap_buffers(); /* display now */

  /* Calculate the lighting from the sun */
  calc_map_lighting(game->map);

  round = game->round = allocate_round();
  if(!round)
    return FALSE;

  round->world = game->world;
  round->map = game->map;
  round->game = game;

  round->turn = 0;
  round->max_turns = 10000;

  round->driving_time = game->driving_time;

  /* Add a camera in */
  add_camera(round);
  round->camera->dist = 100.0;
  round->camera->fixed_alt = FALSE;

  /* Use the same players as in the game */
  round->players = copy_player_list(game->players);

  /* Clear out the projectile/trails/blasts (just in case there are some around) */
  clear_projectiles(round);
  clear_trails(round);
  clear_blasts(round);
  clear_smoke(round);
  clear_birds(round);

  /* Set the players up */
  setup_round_players(round);

  /* Create the birds */
  create_birds(round, game->map->birds_number);

  /* Start the first turn */
  start_turn(round);

  /* Display the round */
  set_display_mode(round->world, NATOW_DISPLAY_ROUND);

  return TRUE;
}

/* Draw a screen to start a round */
int draw_start_round(Game *game)
#define MAX_LENGTH 100
{
  char text[MAX_LENGTH];

  activate_2d();

  snprintf(text, MAX_LENGTH, "Round %d", game->round_num+1);
  draw_2d_text_center(text, 1, 98, 95, 0, FALSE);
  draw_2d_text_center("Ready to go!", 1, 98, 90, 0, FALSE);
  draw_2d_text_center("Press enter to continue", 1, 98, 6, 2, TRUE);
  draw_2d_text_center("Escape to end game", 1, 98, 1, -1, FALSE);

  deactivate_2d();

  return TRUE;
}

/* Manage the input while in the start round screen */
int manage_start_round_input(Game *game, int key)
{
  if(key == ' ' || key == KEY_RETURN || key == KEY_MOUSE1)
    start_round(game);

  else if(key == KEY_ESC || key == KEY_MOUSE3)
    end_game(game);

  return TRUE;
}

/* Setup the players for a round */
void setup_round_players(Round *round)
{
  Player *player;
  PlayerEntry *pentry;
  int i;

  /* Resurrect all the players,
     set their initial health, select the null weapon,
     reset their kills for this round, winnings, costs,
     reset their physics etc, etc */
  pentry = round->players->first;
  while(pentry)
  {
    player = pentry->player;

    /* Ressurect */
    player->player_state = PLAYER_ALIVE;
    player->health = player->tank->health;

    player->round_cost = 0;
    player->round_winnings = 0;

    /* Select their first weapon (the numm weapon) */
    player->selected_weapon = player->weapon_list;

    player->round_kills = 0;

    for(i = 0; i < 3; i++)
    {
      player->pos[i] = 0.0;
      player->dir[i] = 0.0;
      player->vel[i] = 0.0;
      player->acc[i] = 0.0;
      player->rot_dir[i] = 0.0;
    }
    /* Point in a valid direction */
    player->dir[2] = 1.0;
    /* To stop error when converting vel[] to dir[] */
    player->vel[1] = -0.01;

    player->rot_dir[0] = 1.0;
    player->rot_angle = 0.0;

    /* Always in the worlds gravity */
    player->acc[1] = round->game->gravity;

    /* If single player, engine is always active */
    if(round->game->mode == SINGLE_PLAYER)
      player->engine_on = TRUE;
    else
      player->engine_on = FALSE;

    player->has_driven = FALSE;
    player->brakes = FALSE;
    player->speed = 0.0;
    player->accel = 0.0;

    player->tank_state = IN_AIR;
    player->heading = 0.0;
    player->rot_dir[0] = 0.0;
    player->rot_dir[1] = 1.0;
    player->rot_dir[2] = 0.0;
    player->rot_angle = 0.0;

    player->turret.heading = 0.0;
    player->turret.pitch = -30.0;

    player->muzzle_vel = 25.0;
    player->thrust = 30.0;

    player->underwater_time = 0.0;

    pentry = pentry->next;
  }

  /* Drop the players onto the map */
  randomize_player_locations(round);
}

/* Put the players in random locations on the map */
static int randomize_player_locations(Round *round)
{
  Player *player;
  PlayerEntry *pentry;
  Map *map;
  float width, height;
  int eff_sp, nb_sp = 0, i;

  if(!round)
    return FALSE;

  map = round->map;
  width = map->width;
  height = map->height;

  /* number of starting positions */
  if(!map->startingpoints)
    nb_sp = 0;
  else
    while(map->startingpoints[nb_sp])
    {
      map->startingpoints[nb_sp]->selected = FALSE;
      nb_sp++;
    }
  eff_sp = nb_sp;

  /* Go through each player, randomizing positions */
  pentry = round->players->first;
  while(pentry)
  {
    player = pentry->player;

    player->pos[1] = map->max_altitude + 5.0;
    if(eff_sp > 0) /* Starting positions still available */
    {
      /* Choose an availiable starting point */
      do
      {
	i = random_integer(0, nb_sp - 1);
      } while(map->startingpoints[i]->selected);

      map->startingpoints[i]->selected = TRUE;
      player->pos[0] = map->startingpoints[i]->x;
      player->pos[2] = map->startingpoints[i]->y;
      eff_sp--; /* Remove this starting point */
    }
    else /* Do a completely random starting point */
    {
      player->pos[0] = (float)rand() / (float)RAND_MAX * width;
      player->pos[2] = (float)rand() / (float)RAND_MAX * height;
    }

    /* Prevent players to be underwater at beginning */
    /* Note: invalid starting points are not a problem as they will
             not be accepted here, so we will choose a new one, but
             the corresponding starting point is removed from the list */
    if(get_height(map, player->pos[0], player->pos[2]) >= round->game->sea_level)
      pentry = pentry->next;
  }

  return TRUE;
}

/* Update everything in the round */
void update_round(Round *round, float dt)
{
  /* Update the ai for the current player */
  /* (ai.c) */
  update_ai_player_round(round->current_player, dt);

  /* Move, shake etc the camera(s) */
  /* (camera.c) */
  update_cameras(round, dt);

  /* Move, etc the players tanks */
  /* (player.c) */
  update_players(round, dt);

  /* Move, explode, etc the projectiles */
  /* (projectile.c) */
  update_projectiles(round, dt);

  /* Fade away the smoke trails */
  /* (trails.h) */
  update_trails(round, dt);

  /* Expand the explosions */
  /* (explosion.c) */
  update_blasts(round, dt);

  /* expand smokes */
  /* (explosion.c) */
  update_smokes(round, dt);

  /* update birds */
  /* (bird.c) */
  update_birds(round, dt);

  /* update objects */
  /* (objects.c) */
  update_objects(round->game, dt);

  /* Test to see if the current turn is over */
  test_turn_end(round);

  /* Test to see if the round is over */
  test_round_end(round);
}

/* Draw the round */
void draw_round(Round *round)
{
  World *world;
  Game *game;
  Camera *camera;

  world = round->world;
  game = round->game;
  camera = round->camera;

  /* If fog is used */
  if(world->is_fog)
  {
    if(camera->underwater)
    {
      glEnable(GL_FOG);
      glFogf(GL_FOG_DENSITY, 0.01);
      glFogfv(GL_FOG_COLOR, game->map->sea_colour);
    }
    else if(world->is_level_fog)
    {
      glEnable(GL_FOG);
      glFogf(GL_FOG_DENSITY, game->map->fog_density);
      glFogfv(GL_FOG_COLOR, game->map->fog_colour);
    }
  }
  else
    glDisable(GL_FOG);

  /* Clear the buffers */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* Reset the transformation matrix */
  glLoadIdentity();

  /* Rotate camera */
  glRotatef(-camera->rot[2], 0.0f, 0.0f, 1.0f);
  glRotatef(-camera->rot[1], 1.0f, 0.0f, 0.0f);
  glRotatef(-camera->rot[0], 0.0f, 1.0f, 0.0f);

  /* Translate camera */
  glTranslatef(-camera->pos[0], -camera->pos[1], -camera->pos[2]);

  /* Shake the camera */
  glTranslatef(camera->shudder_pos[0], camera->shudder_pos[1], camera->shudder_pos[2]);

  /* Translate light */
  glLightfv(GL_LIGHT0, GL_POSITION, game->map->sun_dir);

  /* FIXME: Add a light onto the camera? */
  /*{
    float col[3] = {1.0, 1.0, 1.0};
    clear_dynamic_light(game->round->map, camera->pos, 1000.0);
    add_dynamic_light(game->round->map, camera->pos, col, 25.0);
  }*/

  /* Draw the map (ground/water/sky etc) */
  /* (draw_map.c) */
  draw_map(game);

  /* Draw players */
  /* (player.c) */
  draw_players(round);

  /* Draw the projectiles */
  /* (projectile.c) */
  draw_projectiles(round);

  /* Draw the projectile trails */
  /* (trails.c) */
  draw_trails(round);

  /* Draw the explosion blasts */
  /* (explosion.c) */
  draw_blasts(round);

  /* Draw smoke */
  draw_smokes(round);

  /* Draw birds */
  draw_birds(round);

  /* Draw the water */
  draw_water(game);

  /* draw objects */
  draw_objects(game);

  /* Draw a pincusion */
  if(round->world->is_pin)
    draw_pincushion(round);

  /* Draw the grid */
  if(round->world->is_grid)
    draw_grid(round);
}

/* See if the round is over, i.e. someone has won (or draw) */
void test_round_end(Round *round)
{
  World *world;
  Player *player, *alive_player = NULL;
  PlayerEntry *pentry;
  int n_alive_players = 0;
  Weapon *weapon;
  int weapons_left = FALSE;

  world = round->world;

  /* If explosions are still occuring, or projeciles are yet to explode, not the end */
  if(round->projectile_list || round->blast_list || round->trail_list)
    return;

  /* Go through the players, see who is alive,
     and if any weapons remain */
  pentry = round->players->first;
  while(pentry)
  {
    player = pentry->player;

    if(player->player_state == PLAYER_ALIVE)
    {
      alive_player = player;
      n_alive_players++;

      /* See if they have any weapons left */
      if(!weapons_left)
      {
	weapon = player->weapon_list;
	while(weapon)
        {
	  if(weapon->number > 0)
	  {
	    weapons_left = TRUE;
	    break;
	  }
	  weapon = weapon->next;
	}
      }
    }

    pentry = pentry->next;
  }

  /* Everyone's dead! */
  if(n_alive_players == 0)
  {
    Nprintf("Everyone's dead!\n");
    end_round(round);
    return;
  }

  /* No more weapons! */
  if(!weapons_left)
  {
    Nprintf("No more weapons in the round!\n");
    end_round(round);
    return;
  }

  /* One player left -- the winner! */
  if(n_alive_players == 1)
  {
    Nprintf("\"%s\" wins!\n", alive_player->name);
    end_round(round);
    return;
  }
}

PlayerEntry *end_round_player = NULL;

/* Draw a screen for the end of a round */
int draw_end_round(Round *round)
#define MAX_LENGTH 100
#define MAX_VIS 11
{
  Player *player;
  PlayerEntry *pentry;
  char text[MAX_LENGTH];
  int i, selected, position, is_winner, nwinners;

  if(!round)
    return FALSE;

  activate_2d();

  snprintf(text, MAX_LENGTH, "Round %d over!", round->game->round_num);
  draw_2d_text_center(text, 1, 98, 95, 0, FALSE);

  draw_2d_text("Name", 2, 80, 0, FALSE);
  draw_2d_text("Winnings", 35, 80, 0, FALSE);
  draw_2d_text("Cost", 65, 80, 0, FALSE);
  draw_2d_text("Kills", 85, 80, 0, FALSE);

  draw_2d_text_center("Press enter to continue", 1, 98, 6, 2, TRUE);
  draw_2d_text_center("Escape to end game", 1, 98, 1, -1, FALSE);

  /* If no player selected, select the first player in this round */
  if(!end_round_player)
    end_round_player = round->players->first;
  if(!end_round_player)
  {
    deactivate_2d();
    return FALSE;
  }

  /* Count how many winners there are */
  nwinners = 0;
  is_winner = TRUE;
  pentry = round->players->first;
  while(pentry && is_winner)
  {
    if(compare_round_winners(&pentry, &round->players->first) == 0)
      nwinners++;
    else
      is_winner = FALSE;

    pentry = pentry->next;
  }

  /* Find out which number the selected player is */
  pentry = round->players->first;
  selected = 0;
  while((pentry != end_round_player) && (selected < (round->players->nplayers - MAX_VIS)))
  {
    pentry = pentry->next;
    selected++;
  }

  /* Don't select the bottom players, as they're already visible */
  if(selected == (round->players->nplayers - MAX_VIS))
     end_round_player = pentry;

  /* Display MAX_VIS players */
  for(i = selected; i < selected + MAX_VIS && pentry; i++)
  {
    player = pentry->player;

    position = 70 - (i - selected)*5;
    is_winner = i < nwinners;

    draw_2d_text(player->name, 2, position, 0, is_winner);

    snprintf(text, MAX_LENGTH, "$%d/%d", player->round_winnings, player->cash);
    draw_2d_text(text, 35, position, 0, is_winner);

    snprintf(text, MAX_LENGTH, "$%d", player->round_cost);
    draw_2d_text(text, 65, position, 0, is_winner);

    snprintf(text, MAX_LENGTH, "%d/%d", player->round_kills, player->kills);
    draw_2d_text(text, 85, position, 0, is_winner);

    pentry = pentry->next;
  }

  /* If more players below, show that */
  if(pentry)
    draw_2d_text("...", 2, 70 - MAX_VIS*5, 0, TRUE);

  deactivate_2d();

  return TRUE;
}

/* Manage the input when in the end round screen */
int manage_end_round_input(Round *round, int key)
{
  if(key == ' ' || key == KEY_RETURN || key == KEY_MOUSE1)
    finish_round(round);
  else if(key == KEY_UP)
  {
    if(end_round_player->last)
      end_round_player = end_round_player->last;
  }
  else if(key == KEY_DOWN)
  {
    if(end_round_player->next)
      end_round_player = end_round_player->next;
  }
  else if(key == KEY_ESC || key == KEY_MOUSE3)
    end_game(round->game);

  return TRUE;
}

/* End a round */
int end_round(Round *round)
{
  PlayerEntry *pentry;

  if(!round)
    return FALSE;

  /* Clear out the projectile/trails/blasts/smoke */
  clear_projectiles(round);
  clear_trails(round);
  clear_blasts(round);
  clear_smoke(round);

  /* Bank any remaining cash floating around */
  pentry = round->players->first;
  while(pentry)
  {
    bank_cash(pentry->player);
    pentry = pentry->next;
  }

  /* Caclulate round winners */
  Nprintf("Caclulating round %d winners\n", round->game->round_num);
  sort_player_list(round->players, compare_round_winners);

  /* Display a winner screen */
  set_display_mode(round->world, NATOW_DISPLAY_END_ROUND);

  return TRUE;
}

/* Comparison function to compare entries for sorting */
static int compare_round_winners(PlayerEntry **entry1, PlayerEntry **entry2)
{
  Player *player1, *player2;
  int t;

  player1 = (*entry1)->player;
  player2 = (*entry2)->player;

  /* Winner is the one with the most kills */
  if((t = player2->round_kills - player1->round_kills))
    return t;

  /* Winner is the one with the most money won */
  if((t = (player2->round_winnings - player2->round_cost) -
      (player1->round_winnings - player1->round_cost)))
    return t;

  /* Winner is the one with the most money left */
  if((t = player2->cash - player1->cash))
    return t;

  /* A draw */
  return 0;
}

/* Clean up, and start the next round */
int finish_round(Round *round)
{
  Game *game;

  if(!round)
    return FALSE;

  game = round->game;

  game->round = NULL;
  free_round(round);

  Nprintf("Round %d ended.\n", game->round_num);

  /* Or end the game */
  if(game->round_num >= game->max_rounds)
    end_game(game);
  else /* (Move onto next round) */
    set_display_mode(game->world, NATOW_DISPLAY_PLAYER_SHOP);

  return TRUE;
}

/* Manage the input during a round */
int manage_round_input(Round *round, int key)
{
  Player *player;
  World *world;

  world = round->world;

  /* *** All non-player related input *** */
  if(key == KL(KC_menu))
    push_display_mode(world, NATOW_DISPLAY_ROUND_MENU);

  else if(key == KL(KC_conf))
    push_display_mode(world, NATOW_DISPLAY_KEY_CONFIG);

  else if(key == KL(KC_pause))
    push_display_mode(world, NATOW_DISPLAY_ROUND_PAUSED);

  else if(key == KL(KC_video))
    push_display_mode(world, NATOW_DISPLAY_VIDEO_CONFIG);

  else if(key == KL(KC_rp))
  {
    if(round->projectile_list)
    {
      remove_projectile(round->projectile_list);
      Nprintf("Removing projectile (forced)\n");
    }
  }

  /* Shortcuts to controlling video config */
  /* (May become obsolete) */
  else if((key == 'l') || (key == 'L'))
  {
    world->is_light = !world->is_light;
    if(world->is_light)
      glEnable(GL_LIGHTING);
    else
      glDisable(GL_LIGHTING);
  }

  else if((key == 'h')||(key == 'H'))
  {
    world->is_smooth = !world->is_smooth;
    if(world->is_smooth)
      glShadeModel(GL_SMOOTH);
    else
      glShadeModel(GL_FLAT);
  }

  else if((key == 'j')||(key == 'J'))
    world->is_fog = !world->is_fog;

  else if((key == 'k')||(key == 'K'))
    world->is_water = !world->is_water;

  else if(key == '*')
    world->is_sky = !world->is_sky;

  else if((key == 'o')||(key == 'O'))
  {
    world->is_texture = !world->is_texture;
    if(world->is_texture)
      glEnable(GL_TEXTURE_2D);
    else
      glDisable(GL_TEXTURE_2D);
  }

  else if((key == 'p')||(key == 'P'))
    world->is_pin = !world->is_pin;

  /* *** All player-related input below here *** */
  if(!(player = round->current_player))
    return FALSE;

  /* In single player mode, 1st player is always active */
  if(round->game->mode == SINGLE_PLAYER)
    player = world->player_list;

  /* If player is managed by an AI, don't let user
     modify the tank state */
  if(player->player_type != PLAYER_HUMAN)
    return TRUE;

  /* Human player-related stuff */
  if(key == KL(KC_eng))
    start_player_engine(player);

  else if(key == KL(KC_vp1))
    inc_player_accel(player, 0.5f);

  else if(key == KL(KC_vp2))
    inc_player_accel(player, 1.0f);

  else if(key == KL(KC_vm1))
    inc_player_accel(player, -0.5f);

  else if(key == KL(KC_vm2))
    inc_player_accel(player, -1.0f);

  else if(key == KL(KC_br))
    toggle_player_brakes(player);

  else if(key == KL(KC_hp2))
    inc_player_heading(player, 4.0f);

  else if(key == KL(KC_hp1))
    inc_player_heading(player, 2.0f);

  else if(key == KL(KC_hm1))
    inc_player_heading(player, -2.0f);

  else if(key == KL(KC_hm2))
    inc_player_heading(player, -4.0f);

  else if(key == KL(KC_mp))
    player->muzzle_vel += 5.0;

  else if(key == KL(KC_mm))
    player->muzzle_vel += -5.0;

  else if(key == KL(KC_tp))
    player->thrust += 5.0;

  else if(key == KL(KC_tm))
    player->thrust -= 5.0;

  else if(key == KL(KC_thp1))
    inc_player_turret_heading(player, 2.0f);

  else if(key == KL(KC_thp2))
    inc_player_turret_heading(player, 4.0f);

  else if(key == KL(KC_thm1))
    inc_player_turret_heading(player, -2.0f);

  else if(key == KL(KC_thm2))
    inc_player_turret_heading(player, -4.0f);

  else if(key == KL(KC_tpm1))
    inc_player_turret_pitch(player, -2.0f);

  else if(key == KL(KC_tpm2))
    inc_player_turret_pitch(player, -4.0f);

  else if(key == KL(KC_tpp1))
    inc_player_turret_pitch(player, 2.0f);

  else if(key == KL(KC_tpp2))
    inc_player_turret_pitch(player, 4.0f);

  else if(key == KL(KC_cyc))
    player_end_turn(player);

  else if(key == KL(KC_weapon))
    cycle_weapon_fowards(player);

  return TRUE;
}

/* Destroy a round */
int free_round(Round *round)
{
  if(!round)
    return FALSE;

  remove_camera(round->camera);

  free_player_list(round->players);
  clear_projectiles(round);
  clear_trails(round);
  clear_blasts(round);
  clear_birds(round);
  clear_smoke(round);

  Nfree(round);

  return TRUE;
}
