/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* satellite.c
   Manage satellites creation, drawing and updating */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <GL/gl.h>

#include "satellite.h"
#include "world.h"
#include "game.h"
#include "map.h"
#include "text.h"
#include "tools.h"

void rotate_x(float angle, float *x, float *y, float *z);
void rotate_y(float angle, float *x, float *y, float *z);
void rotate_z(float angle, float *x, float *y, float *z);
void init_satellite_state(World *world, Satellite *sat);

/* create a new satellite */
Satellite *_create_satellite(World *world,
                             float radius, GLfloat color[4],
			     float ax, float bx,
			     float ay, float by,
			     float az, float bz)
{
  Satellite *tmp=NULL;

  Nprintf("  Adding satellite %g %g,%g-%g,%g-%g,%g %g,%g,%g\n",
          radius, ax, bx, ay, by, az, bz,
	  color[0], color[1], color[2]);

  if ((tmp = malloc(sizeof(Satellite))) == NULL)
    {
    Neprintf("    Cannot allocate satellite. Satellite is discarded.\n");
    return(NULL);
    }

  tmp->world = world;
  
  tmp->radius   = radius;
  tmp->color[0] = color[0];
  tmp->color[1] = color[1];
  tmp->color[2] = color[2];
  tmp->color[3] = color[3];
  tmp->ax   = ax;
  tmp->bx   = bx;
  tmp->ay   = ay;
  tmp->by   = by;
  tmp->az   = az;
  tmp->bz   = bz;
  tmp->current = 0.;
  tmp->rotx = 0.;
  tmp->roty = 0.;
  tmp->rotz = 0.;
  tmp->px = 0.;
  tmp->py = 0.;
  tmp->pz = 0.;
  
  /* init position */
  init_satellite_state(world, tmp);

  Nprintf("  Satellite added.\n");
  return(tmp);
}

/* add a satellite */
Satellite *create_satellite(World *world, char *param)
{
  float radius;
  GLfloat color[4];
  float ax, bx, ay, by, az, bz;

  sscanf(param, "%f%f%f%f%f%f%f%f%f%f",
         &radius, &ax, &bx, &ay, &by, &az, &bz,
	 &(color[0]), &(color[1]), &(color[2]));
  color[3] = 1.;
  
  return(_create_satellite(world, radius, color,
                           ax, bx, ay, by, az, bz));
}


/* delete a satellite */
void delete_satellite(Satellite *sat)
{
  Nfree(sat);
}

/* draw a given satellite */
#define nSTEP (3.14159/16)
#define nMAX(a,b) ((a)>(b)?(a):(b))
void draw_satellite(Satellite *sat)
{
  World *world = sat->world;
  float angle;

  /* draw the satellite */
  glPushAttrib(GL_ENABLE_BIT | GL_LIGHTING_BIT);
  glDisable(GL_TEXTURE_2D);
  glDisable(GL_BLEND);
  glDisable(GL_DEPTH_TEST);
  glDisable(GL_LIGHTING);
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, sat->color);
  glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 128.0f);
  glColor4fv(sat->color);

  glPushMatrix();
  glRotatef(sat->rotx, 1, 0, 0);
  glRotatef(sat->roty, 0, 1, 0);
  glRotatef(sat->rotz, 0, 0, 1);
  glTranslatef(world->game->map->width/2., 0, world->game->map->height/2.);
  glBegin(GL_POLYGON);
  for(angle=0.; angle<2.*3.14; angle+=nSTEP)
    {
    glNormal3f(0,-1,0);
    glVertex3f(sat->radius*cos(angle),
               nMAX(world->game->map->width,world->game->map->height) * 20,
	       sat->radius*sin(angle));
    }
  glEnd();
  glPopMatrix();

  glPopAttrib();
}
#undef nSTEP

/* update position of a given satellite */
void update_satellite(Satellite *sat, float dt)
{
  Map *map = sat->world->game->map;
  float x, y, z;

  /* nothing to do */
  if ((sat->ax == 0.)&&(sat->ay == 0.)&&(sat->az == 0.))
    return;

  /* compute position */
  x = map->width/2.;
  z = map->height/2.;
  y = nMAX(map->width, map->height) * 3.9;

  sat->rotx = sat->ax*sat->current + sat->bx;
  sat->roty = sat->ay*sat->current + sat->by;
  sat->rotz = sat->az*sat->current + sat->bz;
  rotate_x(sat->rotx, &x, &y, &z);
  rotate_y(sat->roty, &x, &y, &z);
  rotate_z(sat->rotz, &x, &y, &z);

  sat->px = x;
  sat->py = y;
  sat->pz = z;

  sat->current += dt;
}


/* this is perhaps still defined in vector.c... I will check */
#define mP2(v) ((v)*(v))
float v_norm(float x, float y, float z)
{
  return(sqrt(mP2(x)+mP2(y)+mP2(z)));
}

/* update initial position of a given satellite */
void init_satellite_state(World *world, Satellite *sat)
{
  float x, y, z;

  /* compute position */
  x = world->game->map->width/2.;
  z = world->game->map->height/2.;
  y = nMAX(world->game->map->width,world->game->map->height) * 3.9;

  sat->current = 0.;

  sat->rotx = sat->ax*sat->current + sat->bx;
  sat->roty = sat->ay*sat->current + sat->by;
  sat->rotz = sat->az*sat->current + sat->bz;

  rotate_x(sat->rotx, &x, &y, &z);
  rotate_y(sat->roty, &x, &y, &z);
  rotate_z(sat->rotz, &x, &y, &z);

  sat->px = x;
  sat->py = y;
  sat->pz = z;
}
#undef nMAX



/* various tools */
void rotate_x(float angle, float *x, float *y, float *z)
{
  float tx, ty, tz;

  tx = *x;
  ty = *y * cos(angle) - *z * sin(angle);
  tz = *y * sin(angle) + *z * cos(angle);

  *x = tx;
  *y = ty;
  *z = tz;
}
void rotate_y(float angle, float *x, float *y, float *z)
{
  float tx, ty, tz;

  tx = *x * cos(angle) + *z * sin(angle);
  ty = *y;
  tz = - *x * sin(angle) + *z * cos(angle);

  *x = tx;
  *y = ty;
  *z = tz;
}
void rotate_z(float angle, float *x, float *y, float *z)
{
  float tx, ty, tz;

  tx = *x * cos(angle) - *y * sin(angle);
  ty = *x * sin(angle) + *z * cos(angle);
  tz = *z;

  *x = tx;
  *y = ty;
  *z = tz;
}

/* Free a sattelite */
int free_satellite(Satellite *satellite)
{
  if(!satellite)
    return FALSE;

  Nfree(satellite);

  return TRUE;
}
