/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yanncik Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* smoke.c
   This file contains all the smoke related code */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <GL/gl.h>

#include "smoke.h"
#include "world.h"
#include "game.h"
#include "text.h"
#include "vector.h"
#include "map.h"
#include "round.h"
#include "tools.h"

static int free_smoke(Smoke *smoke);

/* Adds some smokes into the world */
void add_smoke(Round *round, float *pos, float radius, float force)
{
  Smoke *new_smoke=NULL;
  int density;
  int i;

  /* number of smoke entities */
  density = (force/9.);
  /* multiplied by the smoke % (0-200%) */
  density = (int)((float)density*round->world->smoke_percent/100.);

  for(i=0; i<density; i++)
  {
    new_smoke = malloc(sizeof(Smoke));

    if(new_smoke == NULL)
    {
      Neprintf("Warning: Unable to allocate space for smoke");
      return;
    }

    new_smoke->world = round->game->world;
    new_smoke->game = round->game;
    new_smoke->round = round;
 
    new_smoke->pos[0] = pos[0];
    new_smoke->pos[1] = pos[1];
    new_smoke->pos[2] = pos[2];

    new_smoke->pos[0] += random_float(-radius, radius);
    new_smoke->pos[1] += random_float(-radius, radius);
    new_smoke->pos[2] += random_float(-radius, radius);
    new_smoke->angle = random_float(-180., 180.);
    if (new_smoke->pos[1] < round->game->sea_level)
      new_smoke->pos[1] = round->game->sea_level + 1.;
    new_smoke->pos[1] += random_float(0., 1.5);

    new_smoke->radius = (radius/density)*(1. + random_float(-0.05, 0.2));

    new_smoke->speed = 0.05+random_float(-0.01,0.04);
    new_smoke->age = 0.0;
    new_smoke->last = NULL;
    new_smoke->next = NULL;

    /* Attach to the front of the list */
    new_smoke->next = round->smoke_list;
    if(round->smoke_list != NULL)
      round->smoke_list->last = new_smoke;
    round->smoke_list = new_smoke;
  }
}

/* Remove a smoke from the list */
void remove_smoke(Smoke *smoke)
{
  if(smoke->next != NULL)
    smoke->next->last = smoke->last;
  if(smoke->last != NULL)
    smoke->last->next = smoke->next;
  else
    smoke->round->smoke_list = smoke->next;

  Nfree(smoke);
}

/* Remove all smokes from the list */
void remove_smokes(Round *round)
{
  Smoke *smoke_ptr = round->smoke_list, *temp_smoke;

  while(smoke_ptr != NULL)
  {
    temp_smoke = smoke_ptr->next;

    remove_smoke(temp_smoke);

    smoke_ptr = temp_smoke;
  }
}

/* Updates all the smokes in the world */
void update_smokes(Round *round, float dt)
{
  Smoke *smoke_ptr = round->smoke_list, *temp_smoke;

  while(smoke_ptr != NULL)
  {
    /* Smokes may be removed so update now */
    temp_smoke = smoke_ptr->next;

    update_smoke(smoke_ptr, dt);

    smoke_ptr = temp_smoke;
  }
}

/* Updates a smoke */
void update_smoke(Smoke *smoke, float dt)
{
  smoke->age += dt;
  smoke->pos[1] += dt*3.;
  smoke->angle += dt*(45.+random_float(-1.,15.));

  if (! compute_smoke_color(NULL, smoke))
    remove_smoke(smoke);
}

/* Draws all the smokes in the world */
void draw_smokes(Round *round)
{
  Smoke *smoke_ptr = round->smoke_list;

  glPushAttrib(GL_DEPTH_BUFFER_BIT | GL_COLOR_BUFFER_BIT | GL_LIGHTING_BIT);
  glDepthMask(GL_TRUE);
  glDisable(GL_LIGHTING);
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

  while(smoke_ptr)
  {
    draw_smoke(smoke_ptr);
    smoke_ptr = smoke_ptr->next;
  }

  glPopAttrib();
}

/* compute smoke color */
int compute_smoke_color(GLfloat *color, Smoke *smoke)
{
  GLfloat col[4] = {1., 1., 1., .5};

  /* special: just say if smoke is full transparent */
  if (color == NULL)
    {
    if (.5 - smoke->age*smoke->speed <= 0.)
      return(FALSE);
    return(TRUE);
    }
  /* compute color */
  color[0] = col[0];
  color[1] = col[1];
  color[2] = col[2];
  color[3] = .5 - smoke->age*smoke->speed;
  if (color[3] <= 0.)
    color[3] = 0.; /* may not occur */

  return(TRUE);
}


/* Draws a smoke */
void draw_smoke(Smoke *smoke)
{
  GLfloat smoke_color[4];

  glPushAttrib(GL_LIGHTING_BIT);
  glDisable(GL_LIGHTING);

  compute_smoke_color(smoke_color, smoke);
  glPushMatrix();
  
  glColor4fv(smoke_color);
  glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, smoke_color);
  glTranslatef(smoke->pos[0], smoke->pos[1], smoke->pos[2]);
  glRotatef(smoke->angle, 0., 1., 0.);

  glScalef(1. + smoke->radius*smoke->age*smoke->speed/8.,
           1. + smoke->radius*smoke->age*smoke->speed/9.,
	   1. + smoke->radius*smoke->age*smoke->speed/8.);
  draw_rsmoke();

  glPopMatrix();

  glPopAttrib();
}

/* Draw the smoke entity */
void draw_rsmoke(void)
{
  glBegin(GL_TRIANGLES);

  glVertex3f(1.,0.,0.);
  glVertex3f(0.,0.,1.);
  glVertex3f(0.,.2,0.);
  
  glVertex3f(0.,0.,-1.);
  glVertex3f(1.,0.,0.);
  glVertex3f(0.,.2,0.);
  
  glVertex3f(-1,0.,0.);
  glVertex3f(0.,0.,-1.);
  glVertex3f(0.,.2,0.);
  
  glVertex3f(0.,0.,1.);
  glVertex3f(-1.,0.,0.);
  glVertex3f(0.,.2,0.);

  glVertex3f(1.,0.,0.);
  glVertex3f(0.,-.2,0.);
  glVertex3f(0.,0.,1.);
  
  glVertex3f(0.,0.,-1.);
  glVertex3f(0.,-.2,0.);
  glVertex3f(1.,0.,0.);
  
  glVertex3f(-1,0.,0.);
  glVertex3f(0.,-.2,0.);
  glVertex3f(0.,0.,-1.);
  
  glVertex3f(0.,0.,1.);
  glVertex3f(0.,-.2,0.);
  glVertex3f(-1.,0.,0.);

  glEnd();
}

/* Remove all smoke from a round */
int clear_smoke(Round *round)
{
  Smoke *smoke, *tmp;

  if(!round)
    return FALSE;

  smoke = round->smoke_list;
  while(smoke)
  {
    tmp = smoke->next;
    free_smoke(smoke);
    smoke = tmp;
  }

  round->smoke_list = NULL;

  return TRUE;
}

/* Free a piece of smoke */
static int free_smoke(Smoke *smoke)
{
  if(!smoke)
    return FALSE;

  Nfree(smoke);

  return TRUE;
}
