/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* tank.c
   This file contains the functions relating to creation, manipulation,
   and display of tanks */

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>

#ifdef USE_GLASS
#include <glass.h>
#endif

#include "tank.h"
#include "world.h"
#include "tanks.h"
#include "map.h"
#include "projectile.h"
#include "texture.h"
#include "vector.h"
#include "text.h"
#include "tools.h"
#include "file.h"

/* Allocate space for a new tank */
Tank *allocate_tank(void)
{
  Tank *tank = malloc(sizeof(Tank));

  if(tank == NULL)
  {
    fprintf(stderr, "Error: Unable to allocate space for tank\n");
    return NULL;
  }

  tank->world = NULL;
  tank->name = NULL;
  tank->comment = NULL;

#ifdef USE_GLASS
  /* The GLASS model */
  tank->library = NULL;
  tank->texture_model = NULL;
  tank->model = NULL;

  tank->heading = -1;
  tank->rot = -1;
  tank->rot_x = -1;
  tank->rot_y = -1;
  tank->rot_z = -1;
  tank->turret_heading = -1;
  tank->turret_pitch = -1;

  tank->barrel = -1;
#endif

  tank->has_engine = FALSE;
  tank->mass = 0.0;
  tank->speed = 0.0;
  tank->acc = 0.0;
  tank->health = 0.0;

  tank->texture = 0;

  tank->underwater_ability = 15.0;
  tank->underwater_damage = 5.0;

  tank->scalex = 1.;
  tank->scaley = 1.;
  tank->scalez = 1.;

  tank->last = NULL;
  tank->next = NULL;

  return tank;
}

/* Add a tank to the world */
void add_tank(World *world, Tank *tank)
{
  Tank **tank_ptr = &world->tank_list, *last_tank = NULL;

  /* Add to the end of the list */
  while(*tank_ptr != NULL)
  {
    last_tank = *tank_ptr;
    tank_ptr = &(*tank_ptr)->next;
  }

  *tank_ptr = tank;

  tank->world = world;

  tank->next = NULL;
  tank->last = last_tank;

  world->ntanks++;
}

/* Set a tanks name */
void set_tank_name(Tank *tank, const char *name)
{
  Nfree(tank->name);
  if(name == NULL)
    tank->name = strdup("");
  else
    tank->name = strdup(name);
}

/* Get a tanks name */
const char *get_tank_name(Tank *tank)
{
  return tank->name;
}

/* Set a tanks comment */
void set_tank_comment(Tank *tank, const char *comment)
{
  Nfree(tank->comment);
  if(comment == NULL)
    tank->comment = strdup("");
  else
    tank->comment = strdup(comment);
}

/* Get a tanks comment */
const char *get_tank_comment(Tank *tank)
{
  return tank->comment;
}

#ifdef USE_GLASS
/* Set the GLASS texture object used by the tank */
void set_tank_tex_model(Tank *tank, const char *fname)
{
  if(!tank || !fname)
    return;

  if(!tank->library)
    tank->library = glassLibraryNew("tank lib");

  /* If already has a texture model */
  if(tank->texture_model)
  {
    _glassLibraryRemoveObject(tank->library, tank->texture_model);
    glassFreeObject(tank->texture_model);
  }

  tank->texture_model = glassLibraryLoadObject(tank->library, fname);

  if(!tank->texture_model)
    Neprintf("Error loading tank texture model: %s\n", fname);
}

/* Set the GLASS object used by the tank */
void set_tank_model(Tank *tank, const char *fname)
{
  GlassObject *model;
  int valid_model = TRUE;

  if(!tank || !fname)
    return;

  if(!tank->library)
    tank->library = glassLibraryNew("tank lib");

  /* If already has a model */
  if(tank->model)
  {
    _glassLibraryRemoveObject(tank->library, tank->model);
    glassFreeObject(tank->model);
  }

  tank->model = model = glassLibraryLoadObject(tank->library, fname);

  if(!model)
  {
    Neprintf("Error loading tank model: %s\n", fname);
    return;
  }

  /* The variables */
  if((tank->heading = glassFindVariable(model, "tank heading")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no heading variable!\n");
  }
  if((tank->rot = glassFindVariable(model, "tank rot")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no rotation value variable!\n");
  }
  if((tank->rot_x = glassFindVariable(model, "tank rot x")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no x rotation vector variable!\n");
  }
  if((tank->rot_y = glassFindVariable(model, "tank rot y")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no y rotation vector variable!\n");
  }
  if((tank->rot_z = glassFindVariable(model, "tank rot z")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no z rotation vector variable!\n");
  }
  if((tank->turret_heading = glassFindVariable(model, "gun heading")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no gun heading variable!\n");
  }
  if((tank->turret_pitch = glassFindVariable(model, "gun pitch")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no gun pitch variable!\n");
  }
  
  if((tank->barrel = glassFindActivePoint(model, "gun tip")) == -1)
  {
    valid_model = FALSE;
    Neprintf("Tank has no gun barrel active point!\n");
  }
  
  if(!valid_model)
  {
    Neprintf("Tank model file %s (%s), is not a valid tank model, removing.\n"
	     "(see above for errors)\n",
	     fname, _glassObjectGetName(tank->model));
    _glassLibraryRemoveObject(tank->library, tank->model);
    glassFreeObject(tank->model);
    tank->model = NULL;
    tank->heading = -1;
    tank->rot = -1;
    tank->rot_x = -1;
    tank->rot_y = -1;
    tank->rot_z = -1;
    tank->turret_heading = -1;
    tank->turret_pitch = -1;
    tank->barrel = -1;
  }
}

/* Get the GLASS object used by the tank */
GlassObject *get_tank_model(Tank *tank)
{
  return tank->model;
}
#endif

/* Set a tanks mass */
void set_tank_mass(Tank *tank, float mass)
{
  tank->mass = mass;
}

/* Set a tanks maximum speed */
void set_tank_speed(Tank *tank, float speed)
{
  tank->speed = speed;
}

/* Set a tanks maximum acceleration */
void set_tank_acc(Tank *tank, float acc)
{
  tank->acc = acc;
}

/* Set a tanks health */
void set_tank_health(Tank *tank, float health)
{
  tank->health = health;
}

/* Set tank texture */
void set_tank_texture(Tank *tank, char *texture)
{
  tank->texture = load_gl_texture(texture, FALSE);

  if (tank->texture == 0)
    Neprintf("  Warning: cannot load tank texture.\n");
}

/* Set tank scales */
void set_tank_scales(Tank *tank, char *scales)
{
  float sx, sy, sz;

  sscanf(scales, "%f%f%f", &sx, &sy, &sz);
  /* check bounds */
  if (sx < 0.1)
    sx = 0.1;
  if (sy < 0.1)
    sy = 0.1;
  if (sz < 0.1)
    sz = 0.1;
  if (sx > 2.)
    sx = 2.;
  if (sy > 2.)
    sy = 2.;
  if (sz > 2.)
    sz = 2.;
  tank->scalex = sx;
  tank->scaley = sy;
  tank->scalez = sz;
}

/* Load tanks from a tank file, return FALSE if unable to open file */
int load_tanks(World *world, const char *fname)
{
  FILE *tank_file = fopen(fname, "r");
  Tank *tank;
  char *option;
  int ntanks = 0;

  Nprintf("Loading tank list...\n");

  if(tank_file == NULL)
  {
    Nprintf("  Unable to open tank file \"%s\"\n", fname);
    return FALSE;
  }

  while(1)
  {
    /* Get the next available option */
    option = get_file_word(tank_file);

    /* If no more options */
    if(option == NULL)
    {
      fclose(tank_file);
      Nprintf("  %d tank(s) found\n", ntanks);
      Nprintf("Tank list loaded\n");
      return TRUE;
    }

    if(!strcmp(option, "tank"))
    {
      tank = load_tank(tank_file, fname);
      add_tank(world, tank);
      ntanks++;
    }
    else
      Neprintf("  Warning: Unknown option \"%s\" in tank file \"%s\".\n", option, fname);
  }

  /* Shouldn't get here */
  fclose(tank_file);
  return FALSE;
}

/* Load a tank from a file */
Tank *load_tank(FILE *tank_file, const char *fname)
{
  Tank *tank = allocate_tank();
  char *option = NULL, *value = NULL;
#ifdef USE_GLASS
  char *model_file = NULL, *texture_file = NULL;
#endif

  if(!tank || !fname)
    return NULL;

  while(1)
  {
    Nfree(option);
    Nfree(value);
    option = get_file_word(tank_file);
    /* Reached the end of the tank definition */
    if(!strcmp(option, "end"))
    {
      Nfree(option);

#ifdef USE_GLASS
      set_tank_tex_model(tank, texture_file);
      set_tank_model(tank, model_file);

      Nfree(texture_file);
      Nfree(model_file);
#endif

      /* If cannot move, must have no engine */
      if(tank->speed || tank->acc)
	tank->has_engine = TRUE;

      return tank;
    }
    value = get_file_word(tank_file);

    if(!option || !value)
    {
      Nfree(option);
      Nfree(value);
#ifdef USE_GLASS
      Nfree(texture_file);
      Nfree(model_file);
#endif

      Neprintf("  Warning: Unexpected EOF in tank\n");
      return tank;
    }

    /* The name of the tank */
    if(!strcmp(option, "name"))
      set_tank_name(tank, value);

#ifdef USE_GLASS
    /* A GLASS model containing the texture used by the model */
    else if(!strcmp(option, "texture_model"))
    {
      Nfree(texture_file);
      texture_file = make_absolute_fname(fname, value);
    }

    /* The tanks GLASS model */
    else if(!strcmp(option, "model"))
    {
      Nfree(model_file);
      model_file = make_absolute_fname(fname, value);
    }
#else
    /* Ignore them, you're not using GLASS */
    else if(!strcmp(option, "texture_model"))
      Nprintf("  Option \"texture_model\" in tank file ignored, as not using GLASS\n");

    else if(!strcmp(option, "model"))
      Nprintf("  Option \"model\" in tank file ignored, as not using GLASS\n");
#endif

    /* A comment about the tank */
    else if(!strcmp(option, "comment"))
      set_tank_comment(tank, value);

    /* texture associated to the tank */
    else if(!strcmp(option, "texture"))
      set_tank_texture(tank, value);

    /* scales for the tank */
    else if(!strcmp(option, "scales"))
      set_tank_scales(tank, value);

    /* The mass of the tank, in kg */
    else if(!strcmp(option, "mass"))
      set_tank_mass(tank, atof(value));

    /* The top speed of the tank, in ms^-1 */
    else if(!strcmp(option, "speed"))
      set_tank_speed(tank, atof(value));

    /* The maximum acceleration of the tank, in ms^-2 */
    else if(!strcmp(option, "accel"))
      set_tank_acc(tank, atof(value));

    /* The health of the tank, in *magic* units... :) */
    else if(!strcmp(option, "health"))
      set_tank_health(tank, atof(value));

    /* The time the tank can stay under water level w/o damages */
    else if(!strcmp(option, "water_resist"))
      tank->underwater_ability = atof(value);

    /* Damages taken by the tank if stayed too long under water level */
    else if(!strcmp(option, "water_damage"))
      tank->underwater_damage = atof(value);

    else
      Neprintf("  Warning: Unknown option \"%s\" in tank file\n", option);
  }

  return NULL;
}

/* Find a tank by name, return NULL if not found  */
Tank *find_tank(World *world, const char *name)
{
  Tank *tank = world->tank_list;

  while(tank != NULL)
  {
    if(!strcmp(tank->name, name))
      return tank;
    tank = tank->next;
  }

  return NULL;
}

/* Draw a tank */
void draw_tank(Tank *tank)
{
#ifdef USE_GLASS
  GlassObject *model = tank->model;

  if(!model)
    return;

  glassSetVariable(model, tank->heading, 0.0);
  glassSetVariable(model, tank->rot, 0.0);
  glassSetVariable(model, tank->rot_x, 0.0);
  glassSetVariable(model, tank->rot_y, 1.0);
  glassSetVariable(model, tank->rot_z, 0.0);
  glassSetVariable(model, tank->turret_pitch, -10.0);
  glassSetVariable(model, tank->turret_heading, 0.0);

  glassDrawObject(tank->model);
#else

  GLfloat white[4] = {1.0, 1.0, 1.0, 1.0};
  GLfloat c[4] = {0.6, 0.5, 0.3, 1.0};
  DisplayLists *lists = &tank->world->lists;

  if(tank->world->is_texture)
  {
    glEnable(GL_TEXTURE_2D);
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, white);
    glColor3fv(white);
  }
  else
  {
    glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, c);
    glColor3fv(white);
  }

  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
  glBindTexture(GL_TEXTURE_2D, tank->texture);

  /* Change tank size */
  glScalef(tank->scalex, tank->scaley, tank->scalez);

  /* Body */
  glCallList(lists->body);

  /* Turret */
  glTranslatef(0.0f, 0.0f, 0.25);
  glCallList(lists->turret);

  /* Gun */
  glTranslatef(0.0f, 1.75, 0.0f);
  glCallList(lists->gun);
  glDisable(GL_TEXTURE_2D);
#endif
}

/* Free all the tanks in the world */
void free_tanks(World *world)
{
  Tank *tank = world->tank_list, *tmp_tank;

  while(tank != NULL)
  {
    tmp_tank = tank->next;
    free_tank(tank);
    tank = tmp_tank;
  }
}

/* Free a tank */
void free_tank(Tank *tank)
{
  free(tank->name);
  if (tank->texture != 0)
    glDeleteTextures(1, &(tank->texture));
  
#ifdef USE_GLASS
  glassFreeLibrary(tank->library);
#endif
  Nfree(tank);
}
