/* 
 * NATOW - Tank to tank warfare in a 3D fully destructable world
 *
 * Copyright (C) 2000-2001
 *                     Robert Ancell  <bob27@users.sourceforge.net>
 *                     Yannick Perret <hexasoft@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* vector.c
   This file contains basic vector functions */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <math.h>

#include "vector.h"


/* macro to replace the costy 'fabs' from math.h */
#ifndef nABS
#define nABS(a) ((a)<0?-(a):(a))
#endif

/* Make a 3d vector out of three floats */
float *make_vector(float x, float y, float z)
{
  float *v = malloc(sizeof(float) * 3);

  v[0] = x;
  v[1] = y;
  v[2] = z;

  return v;
}
/* Make a 3d vector out of three floats */
void _make_vector(float x, float y, float z, float *v)
{
  v[0] = x;
  v[1] = y;
  v[2] = z;
}

/* Duplicate a 3d vector */
float *vector_dup(float *v)
{
  return make_vector(v[0], v[1], v[2]);
}

/* Duplicate a 3d vector */
float *_vector_dup(float *v0, float *v)
{
  v[0] = v0[0];
  v[1] = v0[1];
  v[2] = v0[2];

  return v;
}

/* Normalize a 3d vector (using the 2-norm) */
float *normalize_vector(float *v)
{
  float l = vector_2norm(v);

  if(l != 0.0)
  {
    v[0] /= l;
    v[1] /= l;
    v[2] /= l;
  }

  return v;
}

/* Return the 1-norm of a 3d vector */
float vector_1norm(float *v)
{
  return nABS(v[0]) + nABS(v[1]) + nABS(v[2]);
}

/* Return the 2-norm of a 3d vector (cartesian length) */
float vector_2norm(float *v)
{
  return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

/* Return the norm of a 3d vector without the sqrt */
float vector_norm(float *v)
{
  return(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

/* Return the infinity-norm of a 3d vector */
float vector_infnorm(float *v)
{
  return nABS(v[0]) + nABS(v[1]) + nABS(v[2]);
}

/* Add two 3d vectors */
float *vector_add(float *v1, float *v2)
{
  float *v = malloc(sizeof(float) * 3);

  v[0] = v1[0] + v2[0];
  v[1] = v1[1] + v2[1];
  v[2] = v1[2] + v2[2];

  return v;
}

/* Add two 3d vectors */
void _vector_add(float *v1, float *v2, float *v)
{
  v[0] = v1[0] + v2[0];
  v[1] = v1[1] + v2[1];
  v[2] = v1[2] + v2[2];
}

/* Subtract two 3d vectors (v1 - v2) */
float *vector_sub(float *v1, float *v2)
{
  float *v = malloc(sizeof(float) * 3);

  v[0] = v1[0] - v2[0];
  v[1] = v1[1] - v2[1];
  v[2] = v1[2] - v2[2];

  return v;
}

/* Subtract two 3d vectors (v1 - v2) */
float *_vector_sub(float *v1, float *v2, float *v)
{
  v[0] = v1[0] - v2[0];
  v[1] = v1[1] - v2[1];
  v[2] = v1[2] - v2[2];

  return v;
}

/* Scale a 3d vector by a constant */
float *vector_scale(float scalar, float *v0)
{
  float *v = malloc(sizeof(float) * 3);

  v[0] = v0[0] * scalar;
  v[1] = v0[1] * scalar;
  v[2] = v0[2] * scalar;

  return v;
}

/* Scale a 3d vector by a constant */
void _vector_scale(float scalar, float *v0, float *v)
{
  v[0] = v0[0] * scalar;
  v[1] = v0[1] * scalar;
  v[2] = v0[2] * scalar;
}

/* Dot product for two 3d vectors */
float vector_dot(float *v1, float *v2)
{
  return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

/* Cross product */
float *vector_cross(float *v1, float *v2)
{
  float *v = malloc(sizeof(float) * 3);

  v[0] = v1[1]*v2[2] - v1[2]*v2[1];
  v[1] = v1[2]*v2[0] - v1[0]*v2[2];
  v[2] = v1[0]*v2[1] - v1[1]*v2[0];

  return v;
}

/* Cross product */
void _vector_cross(float *v1, float *v2, float *v)
{
  v[0] = v1[1]*v2[2] - v1[2]*v2[1];
  v[1] = v1[2]*v2[0] - v1[0]*v2[2];
  v[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

/* Makes a normalized vector that points in a random (spherical) direction */
float *make_random_vector(void)
{
  float *v = malloc(sizeof(float) * 3);
  float theta, phi, sin_theta, cos_theta, sin_phi, cos_phi;

  /* Get random spherical angles */
  theta = 2.0 * M_PI * rand()/(RAND_MAX+1.0);
  phi = asin(2.0 * rand()/(RAND_MAX+1.0) - 1.0);

  sin_theta = sin(theta);
  cos_theta = cos(theta);
  sin_phi = sin(phi);
  cos_phi = cos(phi);

  v[0] = cos_phi * cos_theta;
  v[1] = sin_phi;
  v[2] = cos_phi * sin_theta;

  return v;
}

/* Makes a normalized vector that points in a random (spherical) direction */
void _make_random_vector(float *v)
{
  float theta, phi, sin_theta, cos_theta, sin_phi, cos_phi;

  /* Get random spherical angles */
  theta = 2.0 * M_PI * rand()/(RAND_MAX+1.0);
  phi = asin(2.0 * rand()/(RAND_MAX+1.0) - 1.0);

  sin_theta = sin(theta);
  cos_theta = cos(theta);
  sin_phi = sin(phi);
  cos_phi = cos(phi);

  v[0] = cos_phi * cos_theta;
  v[1] = sin_phi;
  v[2] = cos_phi * sin_theta;
}

/* Make a random vector that deviates a maximum of 'angle' from the direction */
float *make_random_spread(float *dir, float dangle)
{
  float angle = 0.5 * dangle;
  float max_sin_theta = sin(angle);
  float *v = malloc(sizeof(float) * 3);
  float theta, phi, sin_theta, cos_theta, sin_phi, cos_phi;
  /* The direction angles */
  float cos_phi_d;
  /* removed to eliminate warnings
  float theta_d, phi_d, sin_theta_d, cos_theta_d, sin_phi_d, cos_phi_d;
  float k;
  */

  cos_phi_d = dir[0]*dir[0] + dir[2]*dir[2];
  /*min_sin_phi = asin(*/

  /* Get random spherical angles */
  theta = 2.0 * M_PI * rand()/(RAND_MAX+1.0);
  phi = 0.5 * M_PI - asin(max_sin_theta * rand()/(RAND_MAX+1.0));

  sin_theta = sin(theta);
  cos_theta = cos(theta);
  sin_phi = sin(phi);
  cos_phi = cos(phi);

  v[0] = cos_phi * cos_theta;
  v[1] = sin_phi;
  v[2] = cos_phi * sin_theta;

  return v;
}

/* Make a random vector that deviates a maximum of 'angle' from the direction */
void _make_random_spread(float *dir, float dangle, float *v)
{
  float angle = 0.5 * dangle;
  float max_sin_theta = sin(angle);
  float theta, phi, sin_theta, cos_theta, sin_phi, cos_phi;
  /* The direction angles */
  float cos_phi_d;
  /* removed to eliminate warnings
  float theta_d, phi_d, sin_theta_d, cos_theta_d, sin_phi_d, cos_phi_d;
  float k;
  */

  cos_phi_d = dir[0]*dir[0] + dir[2]*dir[2];
  /*min_sin_phi = asin(*/

  /* Get random spherical angles */
  theta = 2.0 * M_PI * rand()/(RAND_MAX+1.0);
  phi = 0.5 * M_PI - asin(max_sin_theta * rand()/(RAND_MAX+1.0));

  sin_theta = sin(theta);
  cos_theta = cos(theta);
  sin_phi = sin(phi);
  cos_phi = cos(phi);

  v[0] = cos_phi * cos_theta;
  v[1] = sin_phi;
  v[2] = cos_phi * sin_theta;
}

/* Formats a 3d vector an prints it to stdout */
void print_vector(float *v)
{
  printf("(%f, %f, %f)\n", v[0], v[1], v[2]);
}

/* distance between two points */
float vector_distance(float *v1, float *v2)
{
  float tmp[3];

  _vector_sub(v1, v2, tmp);
  return(vector_2norm(tmp));
}
